Package: dist.structure
Title: Structured Random Variables for Reliability System Distributions
Version: 0.5.0
Authors@R: person("Alexander", "Towell",
    email = "lex@metafunctor.com",
    role = c("aut", "cre"),
    comment = c(ORCID = "0000-0001-6443-9897"))
Description: Extends the 'algebraic.dist' distribution algebra to
    random variables with internal structure: coherent reliability
    systems decomposed into components arranged by a structure function
    (series, parallel, k-out-of-n, bridge, and arbitrary topologies via
    minimal path sets). Every 'dist_structure' object is a 'dist', so
    the full distribution algebra (mean, vcov, sampler, surv, cdf) works
    automatically via default methods that compose component-level
    distributions through the topology. Adds structural queries:
    structure function evaluation, minimal path and cut sets, system
    signature, critical states, dual, Birnbaum structural importance,
    and system reliability. Topology shortcut constructors (series_dist,
    parallel_dist, kofn_dist, bridge_dist) produce ready-to-use dists
    from component dists and a chosen structure.
License: MIT + file LICENSE
URL: https://github.com/queelius/dist.structure
BugReports: https://github.com/queelius/dist.structure/issues
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: algebraic.dist, stats, utils
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, spelling, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-07 07:57:13 UTC; spinoza
Author: Alexander Towell [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6443-9897>)
Maintainer: Alexander Towell <lex@metafunctor.com>
Repository: CRAN
Date/Publication: 2026-05-12 18:10:02 UTC
Built: R 4.6.0; ; 2026-05-12 18:20:37 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_dist_structure,dist)
S3method(as_dist_structure,dist_structure)
S3method(birnbaum_importance,dist_structure)
S3method(cdf,cold_standby_dist)
S3method(cdf,dist_structure)
S3method(component,coherent_dist)
S3method(component,cold_standby_dist)
S3method(component,dual_of_system)
S3method(compose_systems,dist_structure)
S3method(critical_states,dist_structure)
S3method(criticality_importance,dist_structure)
S3method(density,exp_kofn)
S3method(density,exp_series)
S3method(density,wei_kofn)
S3method(dual,coherent_dist)
S3method(dual,dist_structure)
S3method(dual,dual_of_system)
S3method(format,dist_structure)
S3method(hazard,exp_kofn)
S3method(hazard,exp_series)
S3method(hazard,wei_kofn)
S3method(hazard,wei_series)
S3method(is_coherent,dist_structure)
S3method(mean,cold_standby_dist)
S3method(mean,exp_parallel)
S3method(mean,exp_series)
S3method(mean,wei_homogeneous_series)
S3method(min_cuts,dist_structure)
S3method(min_paths,coherent_dist)
S3method(min_paths,dist_structure)
S3method(min_paths,parallel_dist)
S3method(min_paths,series_dist)
S3method(ncomponents,coherent_dist)
S3method(ncomponents,cold_standby_dist)
S3method(ncomponents,dual_of_system)
S3method(phi,dist_structure)
S3method(phi,dual_of_system)
S3method(phi,kofn_dist)
S3method(phi,parallel_dist)
S3method(phi,series_dist)
S3method(print,dist_structure)
S3method(reliability,dist_structure)
S3method(sampler,cold_standby_dist)
S3method(sampler,dist_structure)
S3method(sampler,exp_kofn)
S3method(sampler,exp_parallel)
S3method(sampler,exp_series)
S3method(sampler,gamma_series)
S3method(sampler,lognormal_series)
S3method(sampler,wei_homogeneous_series)
S3method(sampler,wei_kofn)
S3method(sampler,wei_series)
S3method(structural_importance,dist_structure)
S3method(substitute_component,dist_structure)
S3method(surv,cold_standby_dist)
S3method(surv,dist_structure)
S3method(surv,exp_kofn)
S3method(surv,exp_parallel)
S3method(surv,exp_series)
S3method(surv,gamma_series)
S3method(surv,lognormal_series)
S3method(surv,parallel_dist)
S3method(surv,series_dist)
S3method(surv,wei_homogeneous_series)
S3method(surv,wei_kofn)
S3method(surv,wei_series)
S3method(system_censoring,dist_structure)
S3method(system_lifetime,dist_structure)
S3method(system_signature,dist_structure)
S3method(vesely_fussell_importance,dist_structure)
export(as_dist_structure)
export(birnbaum_importance)
export(bridge_dist)
export(coherent_dist)
export(cold_standby_dist)
export(component)
export(compose_systems)
export(consecutive_k_dist)
export(critical_states)
export(criticality_importance)
export(dual)
export(exp_kofn)
export(exp_parallel)
export(exp_series)
export(gamma_series)
export(is_coherent)
export(is_dist_structure)
export(kofn_dist)
export(lognormal_series)
export(max_iid)
export(min_cuts)
export(min_iid)
export(min_paths)
export(ncomponents)
export(order_statistic)
export(parallel_dist)
export(phi)
export(reliability)
export(series_dist)
export(structural_importance)
export(substitute_component)
export(system_censoring)
export(system_lifetime)
export(system_signature)
export(validate_dist_structure)
export(vesely_fussell_importance)
export(wei_homogeneous_series)
export(wei_kofn)
export(wei_series)
importFrom(algebraic.dist,cdf)
importFrom(algebraic.dist,hazard)
importFrom(algebraic.dist,sampler)
importFrom(algebraic.dist,surv)
importFrom(stats,density)
importFrom(stats,dexp)
importFrom(stats,dweibull)

# dist.structure

<!-- badges: start -->
[![R-universe status](https://queelius.r-universe.dev/badges/dist.structure)](https://queelius.r-universe.dev/dist.structure)
<!-- badges: end -->

Structured random variables for reliability system distributions.
`dist.structure` extends the `algebraic.dist` distribution algebra to
random variables with internal structure: coherent reliability systems
decomposed into components arranged by a structure function (series,
parallel, k-out-of-n, bridge, consecutive-k, and arbitrary topologies
via minimal path sets).

Every `dist_structure` object is a `dist`, so the full distribution
algebra (`mean`, `vcov`, `sampler`, `surv`, `cdf`) works automatically
via default methods that compose component-level distributions through
the topology. `dist.structure` adds structural queries (`phi`,
`min_paths`, `min_cuts`, `system_signature`, `critical_states`, `dual`,
`is_coherent`), importance measures (Birnbaum structural, Birnbaum
reliability, criticality, Vesely-Fussell), and compositional operations
(`substitute_component`, `compose_systems`).

## Installation

From r-universe:

```r
install.packages("dist.structure",
  repos = c("https://queelius.r-universe.dev", "https://cran.r-project.org"))
```

From GitHub:

```r
# install.packages("remotes")
remotes::install_github("queelius/dist.structure")
```

## Quick tour

```r
library(dist.structure)

# A 2-out-of-3 system of independent exponentials.
sys <- exp_kofn(k = 2, rates = c(1, 2, 3))

# Distribution queries (inherited from algebraic.dist).
algebraic.dist::surv(sys)(1)           # system survival at t = 1
algebraic.dist::sampler(sys)(10)       # 10 system-lifetime samples

# Topology queries.
phi(sys, state = c(1, 1, 0))           # system functions? -> 1
min_paths(sys)                         # minimal path sets
system_signature(sys)                  # Samaniego signature

# Importance.
structural_importance(sys, j = 1)      # fraction of pivotal states
birnbaum_importance(sys, j = 1, p = 0.8)  # dR/dp_j at p = 0.8
reliability(sys, p = 0.9)              # system reliability

# Composition.
sys2 <- substitute_component(sys, j = 2,
  new_component = algebraic.dist::weibull_dist(shape = 2, scale = 1))
```

See `vignette("getting-started", package = "dist.structure")` for a
longer tour, and the other vignettes (`coherent-systems`,
`distribution-interface`, `importance-measures`, `composition`,
`non-coherent`) for deeper topics.

## Ecosystem

`dist.structure` sits at the bottom of the reliability ecosystem. It
depends only on `algebraic.dist`. Downstream packages implement the
protocol on their own subclasses:

- **serieshaz** registers dist.structure methods on `dfr_dist_series`
  so series systems built from arbitrary dynamic failure rate
  components get the protocol for free.
- **kofn** provides maximum likelihood estimation for k-out-of-n
  systems, using dist.structure as its DGP and topology layer.
- **maskedcauses** wraps its exponential-series API over
  `dist.structure::exp_series`.

## License

MIT. See `LICENSE`.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----setup--------------------------------------------------------------------
library(dist.structure)
library(algebraic.dist)

## -----------------------------------------------------------------------------
sys <- parallel_dist(list(
  exponential(1),
  exponential(1),
  exponential(1)
))

# Survival at t = 1 for the all-exponential parallel:
algebraic.dist::surv(sys)(1)

sys2 <- substitute_component(sys, j = 1,
  new_component = weibull_dist(shape = 2, scale = 2))

# Survival improves because component 1 is now longer-lived:
algebraic.dist::surv(sys2)(1)

## -----------------------------------------------------------------------------
length(min_paths(sys2))                # still 3 (parallel)
ncomponents(sys2)                      # still 3

## -----------------------------------------------------------------------------
inner_a <- parallel_dist(list(exponential(1), exponential(1)))
inner_b <- parallel_dist(list(exponential(1), exponential(1)))

# Outer: series of 2 "slots" (the components are placeholder; they get
# replaced by the inners during compose_systems).
outer <- series_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_a, inner_b))
ncomponents(sys)                       # 4 total leaf components
length(min_paths(sys))                 # 4 composed paths

## -----------------------------------------------------------------------------
min_paths(sys)

## -----------------------------------------------------------------------------
inner_a <- series_dist(list(exponential(1), exponential(1)))
inner_b <- series_dist(list(exponential(1), exponential(1)))

outer <- parallel_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_a, inner_b))
ncomponents(sys)                       # 4 total
length(min_paths(sys))                 # 2 composed paths
min_paths(sys)                         # {1, 2} and {3, 4}

## -----------------------------------------------------------------------------
# Series of: (a 2-parallel block) + (a single exponential)
inner_par <- parallel_dist(list(exponential(1), exponential(1)))
outer <- series_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_par, exponential(1)))
ncomponents(sys)                       # 3: two from parallel + one single
min_paths(sys)                         # {1, 3} and {2, 3}

## -----------------------------------------------------------------------------
# A parallel of two different sub-systems:
sub_a <- series_dist(list(exponential(1), exponential(1)))     # simple series
sub_b <- bridge_dist(replicate(5, exponential(1), simplify = FALSE))  # bridge

outer <- parallel_dist(list(exponential(1), exponential(1)))
combined <- compose_systems(outer, list(sub_a, sub_b))

ncomponents(combined)                  # 2 + 5 = 7 total
length(min_paths(combined))            # = 1 + 4 (series has 1 path, bridge has 4)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(dist.structure)

## -----------------------------------------------------------------------------
alarm_dist <- function(k, sensor_components, master_component) {
  m_sensors <- length(sensor_components)
  stopifnot(k >= 1L, k <= m_sensors,
            inherits(master_component, "dist"),
            all(vapply(sensor_components,
                       function(d) inherits(d, "dist"),
                       logical(1L))))
  structure(
    list(
      k = as.integer(k),
      m_sensors = as.integer(m_sensors),
      m = as.integer(m_sensors + 1L),
      components = c(sensor_components, list(master_component))
    ),
    class = c("alarm_dist", "dist_structure",
              "univariate_dist", "continuous_dist", "dist")
  )
}

## -----------------------------------------------------------------------------
ncomponents.alarm_dist <- function(x) x$m

component.alarm_dist <- function(x, j, ...) {
  stopifnot(j >= 1L, j <= x$m)
  x$components[[j]]
}

phi.alarm_dist <- function(x, state) {
  stopifnot(length(state) == x$m)
  sensor_state <- state[seq_len(x$m_sensors)]
  master_state <- state[x$m]
  as.integer(sum(sensor_state) >= x$k && master_state == 1L)
}

## -----------------------------------------------------------------------------
sensors <- replicate(4, algebraic.dist::exponential(0.5), simplify = FALSE)
master <- algebraic.dist::exponential(0.1)  # master is most reliable
alarm <- alarm_dist(k = 2L, sensors, master)

validate_dist_structure(alarm)

## -----------------------------------------------------------------------------
# Number of components
ncomponents(alarm)

# Structure function evaluation
phi(alarm, c(1, 1, 0, 0, 1))   # 2 sensors + master alive: alarms
phi(alarm, c(1, 1, 1, 1, 0))   # all sensors but no master: silent

# Minimal cut sets (derived from phi via the Berge transversal)
min_cuts(alarm)

# Critical states for component j
critical_states(alarm, 5L)  # the master: critical in every state where >= k sensors are alive

# Reliability at uniform component reliability p
reliability(alarm, 0.9)

# Distribution algebra: system survival via component composition
S <- algebraic.dist::surv(alarm)
S(c(1, 5, 10))

# Sampling: m independent component lifetimes, combined via system_lifetime
withr::with_seed(1, {
  x <- algebraic.dist::sampler(alarm)(1000)
  mean(x)
})

## -----------------------------------------------------------------------------
parallel_of_master_and_sensors <- function(sensor_components, master_component) {
  components <- c(sensor_components, list(master_component))
  m <- length(components)
  obj <- coherent_dist(
    min_paths = list(seq_len(m - 1L), m),  # all sensors as one path; master as another
    components = components,
    m = m
  )
  class(obj) <- c("master_or_sensors_dist", class(obj))
  obj
}

# ==========================================================================
# Reliability importance measures
# ==========================================================================
#
# Three classical measures, all derivable from phi, min_cuts, and the
# component distributions:
#
#   - Birnbaum (reliability) importance:   dR/dp_j
#   - Criticality (Fussell):               component j failed AND critical | system failed
#   - Vesely-Fussell:                      some cut containing j is fully failed | system failed
#
# All live as default methods on `dist_structure`. Specialized subclasses
# may override for closed-form speed.
# ==========================================================================


#' @rdname birnbaum_importance
#' @export
birnbaum_importance.dist_structure <- function(x, j, p) {
  m <- ncomponents(x)
  if (length(p) == 1L) p <- rep(p, m)
  stopifnot(length(p) == m, all(p >= 0 & p <= 1),
            j >= 1L, j <= m, length(j) == 1L)
  p1 <- p; p1[j] <- 1
  p0 <- p; p0[j] <- 0
  reliability(x, p1) - reliability(x, p0)
}


# Helper: component survivals at time t, as a length-m numeric vector.
component_survivals_at <- function(x, t) {
  m <- ncomponents(x)
  vapply(seq_len(m), function(k) {
    algebraic.dist::surv(component(x, k))(t)
  }, numeric(1L))
}


#' @rdname criticality_importance
#' @export
criticality_importance.dist_structure <- function(x, j, t) {
  m <- ncomponents(x)
  stopifnot(j >= 1L, j <= m, length(j) == 1L)
  S <- component_survivals_at(x, t)
  F_j <- 1 - S[j]
  F_sys <- 1 - reliability(x, S)
  if (F_sys == 0) return(0)
  ib <- birnbaum_importance(x, j, S)
  as.numeric(ib * F_j / F_sys)
}


#' @rdname vesely_fussell_importance
#' @export
vesely_fussell_importance.dist_structure <- function(x, j, t) {
  m <- ncomponents(x)
  stopifnot(j >= 1L, j <= m, length(j) == 1L)
  cuts <- min_cuts(x)
  cuts_j <- Filter(function(K) j %in% K, cuts)
  if (length(cuts_j) == 0L) return(0)
  S <- component_survivals_at(x, t)
  F_vec <- 1 - S
  F_sys <- 1 - reliability(x, S)
  if (F_sys == 0) return(0)
  # Inclusion-exclusion: P(union of events "cut K_l is fully failed")
  # where each event's probability under independence is prod_{i in K_l} F_i.
  # For an intersection of several cut-failure events, all components in
  # the union of those cuts must have failed:
  # P(cut A failed AND cut B failed) = prod_{i in A union B} F_i.
  nC <- length(cuts_j)
  p_union <- 0
  for (sz in seq_len(nC)) {
    subsets <- utils::combn(nC, sz, simplify = FALSE)
    for (subset in subsets) {
      union_idx <- sort(unique(unlist(cuts_j[subset])))
      p_union <- p_union + (-1)^(sz + 1L) * prod(F_vec[union_idx])
    }
  }
  as.numeric(p_union / F_sys)
}

# ==========================================================================
# Cold standby system
# ==========================================================================
#
# A cold-standby system has m components arranged sequentially: one is
# active at a time, others wait with zero hazard. When the active one
# fails, a spare takes over (assumed instantaneous, perfect switching).
# System lifetime = T_1 + T_2 + ... + T_m, the sum of independent
# component lifetimes.
#
# Cold standby is NOT a coherent system in the structure-function sense
# (the topology is temporal succession, not order statistics). Therefore
# `cold_standby_dist` does NOT inherit `dist_structure`. It carries
# components and provides standard dist methods (surv, cdf, sampler,
# mean, expectation), but has no phi or min_paths.
#
# Special cases with closed-form aggregate distributions:
#   - iid Exp(rate)        ->  Gamma(shape = m, rate = rate)
#   - heterogeneous Exp    ->  hypoexponential (no simple closed form)
#   - independent normals  ->  Normal(sum mu, sqrt(sum var))
#
# The general default uses Monte Carlo for surv/cdf and exact (sum of
# component means) for `mean`.
# ==========================================================================


#' Cold-standby system distribution
#'
#' Constructs a distribution representing a cold-standby system: one
#' component active at a time with perfect, instantaneous switching to
#' the next spare upon failure. System lifetime equals the sum of
#' independent component lifetimes.
#'
#' Cold standby is not a coherent system in the structure-function sense
#' (its topology is temporal succession, not an order statistic), so the
#' returned object does not inherit `dist_structure`. It IS a `dist`
#' (with `surv`, `cdf`, `sampler`, `mean` available) and exposes
#' [ncomponents()] and [component()] for inspection.
#'
#' Defaults: `sampler` is exact (sample each component independently and
#' sum); `mean` is exact when every component implements `mean()`
#' (otherwise falls back to Monte Carlo via the sampler); `surv` and
#' `cdf` use Monte Carlo with a default of `1e5` simulated lifetimes
#' (override via the `mc` argument). The returned `surv` / `cdf`
#' closures cache their samples after the first call so subsequent
#' evaluations at different `t` values are deterministic given the same
#' `mc`.
#'
#' Methods inherited from `algebraic.dist::univariate_dist` that require
#' `density` or `sup` (notably `expectation`, `vcov`) are NOT supported
#' on cold-standby objects out of the box; specialized subclasses with
#' closed-form aggregate distributions (e.g., iid exponential collapses
#' to `Gamma(m, rate)`) should provide their own methods.
#'
#' @param components List of `dist` objects representing per-stage
#'   component lifetimes.
#' @return An object of class
#'   `c("cold_standby_dist", "univariate_dist", "continuous_dist", "dist")`.
#' @examples
#' # Cold standby of 3 iid Exp(1) components: aggregate is Gamma(3, 1)
#' sys <- cold_standby_dist(replicate(3,
#'   algebraic.dist::exponential(1), simplify = FALSE))
#' mean(sys)  # = 3
#' @export
cold_standby_dist <- function(components) {
  stopifnot(is.list(components), length(components) >= 1L)
  structure(
    list(components = components, m = length(components)),
    class = c("cold_standby_dist", "univariate_dist",
              "continuous_dist", "dist")
  )
}


# cold_standby_dist deliberately implements ncomponents() and component()
# without inheriting `dist_structure`. The two generics are shared across
# both class trees so that user code can iterate uniformly over
# component-bearing systems regardless of whether they are coherent.
# `is_dist_structure(cold_standby_dist(...))` returns FALSE.
#' @export
ncomponents.cold_standby_dist <- function(x) x$m


#' @export
component.cold_standby_dist <- function(x, j, ...) {
  stopifnot(j >= 1L, j <= x$m)
  x$components[[j]]
}


#' @rdname cold_standby_dist
#' @param x A `cold_standby_dist` object.
#' @param ... Ignored.
#' @export
sampler.cold_standby_dist <- function(x, ...) {
  m <- x$m
  comp_samplers <- lapply(x$components, algebraic.dist::sampler)
  function(n, ...) {
    rowSums(sample_component_matrix(comp_samplers, n))
  }
}


#' Mean of a cold-standby system: sum of component means
#'
#' Computes `sum_j E[T_j]` exactly when every component implements a
#' `mean()` method. Falls back to a Monte Carlo estimate from the
#' sampler when any component lacks an exact mean (e.g., a
#' `dist_structure` component whose `mean` would route through
#' `algebraic.dist::univariate_dist` and require `density` / `sup`,
#' which are not provided on `dist_structure` by default).
#'
#' @param x A `cold_standby_dist` object.
#' @param ... Passed to the Monte Carlo fallback as `mc` (default 1e5).
#' @return Numeric scalar.
#' @export
mean.cold_standby_dist <- function(x, ...) {
  exact <- tryCatch(
    sum(vapply(x$components, mean, numeric(1L))),
    error = function(e) NULL
  )
  # An exact NA propagation through `sum()` would silently return NA as
  # the system mean; treat it the same as a missing component method and
  # fall back to MC.
  if (!is.null(exact) && !is.na(exact)) return(exact)
  args <- list(...)
  mc <- if ("mc" %in% names(args)) args$mc else 1e5L
  mean(sampler.cold_standby_dist(x)(mc))
}


#' Cold-standby survival via Monte Carlo (with cached samples)
#'
#' Estimates `S(t) = P(T_1 + ... + T_m > t)` by simulating `mc` system
#' lifetimes and computing the empirical fraction exceeding `t`. The
#' returned closure caches its sample vector: the first call generates
#' `mc` samples, subsequent calls reuse them as long as `mc` is
#' unchanged (a different `mc` triggers a fresh draw). This makes
#' repeated `S(t)` calls deterministic given the same `mc`.
#'
#' For reproducibility across calls to `surv()` itself (i.e., between
#' separately constructed closures), set the RNG seed externally via
#' `set.seed()` before invoking `surv(x)`. Override the method on a
#' subclass with an exact aggregate distribution (e.g., iid exponential
#' collapses to `Gamma(m, rate)`) when an analytic form is available.
#'
#' @rdname cold_standby_dist
#' @export
surv.cold_standby_dist <- function(x, ...) {
  samp <- sampler.cold_standby_dist(x)
  cache <- new.env(parent = emptyenv())
  cache$samples <- NULL
  cache$mc <- NA_integer_
  function(t, mc = 1e5L, ...) {
    mc <- as.integer(mc)
    if (is.null(cache$samples) || cache$mc != mc) {
      cache$samples <- samp(mc)
      cache$mc <- mc
    }
    samples <- cache$samples
    vapply(t, function(ti) mean(samples > ti), numeric(1L))
  }
}


#' Cold-standby CDF: `1 - surv(x)(t, mc = mc)`
#'
#' Computes the CDF by deferring to a fresh `surv` closure. The CDF
#' closure has its own sample cache (independent of any external `surv`
#' closure), so `cdf(x)(t) + surv(x)(t)` is generally not exactly 1
#' unless callers reuse a single `surv` closure: prefer
#' `S <- surv(x); F_t <- 1 - S(t)` over computing both independently.
#'
#' @rdname cold_standby_dist
#' @export
cdf.cold_standby_dist <- function(x, ...) {
  S <- surv.cold_standby_dist(x)
  function(t, mc = 1e5L, ...) 1 - S(t, mc = mc)
}

# ==========================================================================
# Compositional operations
# ==========================================================================
#
# Build new dist_structure objects from existing ones without leaving the
# protocol:
#
#   - substitute_component(x, j, new_component): component-level edit
#   - compose_systems(outer, inner_list):        hierarchical nesting
# ==========================================================================


#' @rdname substitute_component
#' @export
substitute_component.dist_structure <- function(x, j, new_component) {
  m <- ncomponents(x)
  stopifnot(j >= 1L, j <= m, length(j) == 1L)
  # Collect existing components via the primitive, so this works for any
  # subclass providing component().
  components <- lapply(seq_len(m), function(k) component(x, k))
  components[[j]] <- new_component
  coherent_dist(
    min_paths = min_paths(x),
    components = components,
    m = m
  )
}


# Helper: return the component list of an inner argument (a dist_structure
# or a plain dist), along with its min_paths treated relative to a
# zero-based offset of its first component.
unpack_inner <- function(inner) {
  if (is_dist_structure(inner)) {
    n <- ncomponents(inner)
    components <- lapply(seq_len(n), function(k) component(inner, k))
    paths <- min_paths(inner)
  } else {
    # Plain dist: treated as a single-component series (its only path is
    # itself).
    components <- list(inner)
    paths <- list(1L)
  }
  list(components = components, paths = paths, size = length(components))
}


#' @rdname compose_systems
#' @details
#' Computational note: the composed minimal-path enumeration takes the
#' Cartesian product of inner-path choices over each outer path. For an
#' outer system with `p` paths each of length `q`, where each inner has
#' `r` paths, the candidate count grows as `O(p * r^q)` before
#' deduplication. Bridge-of-bridges and similar deeply nested
#' compositions can produce hundreds of candidates; if you find the call
#' slow, build the composed `coherent_dist` directly with a hand-curated
#' `min_paths` list.
#' @export
compose_systems.dist_structure <- function(outer, inner_list) {
  m_outer <- ncomponents(outer)
  stopifnot(length(inner_list) == m_outer)
  unpacked <- lapply(inner_list, unpack_inner)
  sizes <- vapply(unpacked, function(u) u$size, integer(1L))
  offsets <- c(0L, cumsum(sizes))[seq_len(m_outer)]
  # Flatten components in outer order.
  all_components <- do.call(c, lapply(unpacked, function(u) u$components))
  # For each inner, shift its min_paths by its offset to get global indices.
  shifted_paths <- lapply(seq_len(m_outer), function(k) {
    lapply(unpacked[[k]]$paths, function(P) as.integer(P + offsets[k]))
  })
  # Composed min_paths: for each outer path P_out, take the Cartesian
  # product of inner paths for k in P_out and union their (shifted) indices.
  # Build via lapply (no quadratic list-growth cost).
  composed_paths <- unlist(
    lapply(min_paths(outer), function(P_out) {
      inner_choices <- shifted_paths[P_out]
      grid <- expand.grid(lapply(inner_choices, seq_along),
                          KEEP.OUT.ATTRS = FALSE)
      lapply(seq_len(nrow(grid)), function(row_idx) {
        selected <- Map(`[[`, inner_choices, as.integer(grid[row_idx, ]))
        sort(unique(unlist(selected)))
      })
    }),
    recursive = FALSE
  )
  composed_paths <- minimize_sets(composed_paths)
  coherent_dist(
    min_paths = composed_paths,
    components = all_components,
    m = sum(sizes)
  )
}

# ==========================================================================
# Closed-form series of Gamma components
# ==========================================================================
#
# For a series of m independent Gammas with shapes k_j and rates r_j:
#   S_sys(t) = prod_j P(Gamma_j > t) = prod_j stats::pgamma(t, k_j, r_j,
#                                                            lower.tail = FALSE)
# No closed-form CDF inverse for the system in general; sampler uses min
# of independently sampled Gammas.
# ==========================================================================


#' Series of independent Gamma components (closed form)
#'
#' Constructs a `dist_structure` for a series system whose components are
#' independent Gammas. Closed-form `surv` is evaluated by the product of
#' per-component upper-tail probabilities; `cdf` is `1 - surv`; `sampler`
#' generates m independent Gammas and takes the min.
#'
#' @param shapes Positive numeric vector of length `m`: per-component
#'   Gamma shape parameters.
#' @param rates Positive numeric vector of length `m`: per-component
#'   Gamma rate parameters.
#' @return
#' `gamma_series()` returns an object of class
#'   `c("gamma_series", "series_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' @examples
#' sys <- gamma_series(shapes = c(2, 3), rates = c(1, 2))
#' algebraic.dist::surv(sys)(1)
#' @export
gamma_series <- function(shapes, rates) {
  stopifnot(length(shapes) == length(rates),
            all(shapes > 0), all(rates > 0))
  m <- length(shapes)
  components <- lapply(seq_len(m), function(j) {
    algebraic.dist::gamma_dist(shape = shapes[j], rate = rates[j])
  })
  obj <- series_dist(components)
  obj$shapes <- as.numeric(shapes)
  obj$rates <- as.numeric(rates)
  class(obj) <- c("gamma_series", class(obj))
  obj
}


#' @rdname gamma_series
#' @param x A `gamma_series` object.
#' @param ... Ignored.
#' @export
surv.gamma_series <- function(x, ...) {
  series_surv_product(stats::pgamma,
                      list(shape = x$shapes, rate = x$rates))
}


#' @rdname gamma_series
#' @export
sampler.gamma_series <- function(x, ...) {
  samplers <- make_component_samplers(stats::rgamma,
                                      shape = x$shapes, rate = x$rates)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L, min)
  }
}

# ==========================================================================
# Additional distribution-family series specializations
# ==========================================================================


test_that("gamma_series surv equals product of per-component upper tails", {
  shapes <- c(2, 3, 1.5)
  rates <- c(1, 0.5, 2)
  sys <- gamma_series(shapes, rates)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.3, 1, 3)) {
    expected <- prod(stats::pgamma(ti, shape = shapes, rate = rates,
                                   lower.tail = FALSE))
    expect_equal(S(ti), expected, tolerance = 1e-12)
  }
})


test_that("gamma_series with shape = 1 reduces to exp_series", {
  rates <- c(0.5, 1, 2)
  sys_gamma <- gamma_series(shapes = rep(1, 3), rates = rates)
  sys_exp <- exp_series(rates)
  S1 <- algebraic.dist::surv(sys_gamma)
  S2 <- algebraic.dist::surv(sys_exp)
  for (ti in c(0.3, 1, 2)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-10)
  }
})


test_that("gamma_series sampler empirical surv matches closed form", {
  shapes <- c(2, 3, 1.5)
  rates <- c(1, 0.5, 2)
  sys <- gamma_series(shapes, rates)
  withr::local_seed(42)
  x <- algebraic.dist::sampler(sys)(10000L)
  for (t0 in c(0.3, 1)) {
    expect_equal(mean(x > t0), algebraic.dist::surv(sys)(t0),
                 tolerance = 0.02)
  }
})


test_that("gamma_series class hierarchy is correct", {
  sys <- gamma_series(shapes = c(2, 3), rates = c(1, 1))
  expect_s3_class(sys, "gamma_series")
  expect_s3_class(sys, "series_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("lognormal_series surv equals product of per-component upper tails", {
  mu <- c(0, 1, -0.5)
  sd <- c(1, 0.5, 0.8)
  sys <- lognormal_series(mu, sd)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.5, 1, 3)) {
    expected <- prod(stats::plnorm(ti, meanlog = mu, sdlog = sd,
                                   lower.tail = FALSE))
    expect_equal(S(ti), expected, tolerance = 1e-12)
  }
})


test_that("lognormal_series sampler empirical surv matches closed form", {
  mu <- c(0, 1)
  sd <- c(1, 0.5)
  sys <- lognormal_series(mu, sd)
  withr::local_seed(42)
  x <- algebraic.dist::sampler(sys)(20000L)
  for (t0 in c(0.5, 1.5)) {
    expect_equal(mean(x > t0), algebraic.dist::surv(sys)(t0),
                 tolerance = 0.03)
  }
})


test_that("lognormal_series class hierarchy is correct", {
  sys <- lognormal_series(meanlogs = c(0, 1), sdlogs = c(1, 0.5))
  expect_s3_class(sys, "lognormal_series")
  expect_s3_class(sys, "series_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("gamma_series component(j) returns gamma_dist", {
  sys <- gamma_series(shapes = c(2, 3), rates = c(1, 2))
  c1 <- component(sys, 1L)
  expect_s3_class(c1, "gamma_dist")
  expect_equal(c1$shape, 2)
  expect_equal(c1$rate, 1)
})


test_that("lognormal_series component(j) returns lognormal", {
  sys <- lognormal_series(meanlogs = c(0, 1), sdlogs = c(1, 0.5))
  c2 <- component(sys, 2L)
  expect_s3_class(c2, "lognormal")
})

# ==========================================================================
# Closed-form series of Lognormal components
# ==========================================================================
#
# For a series of m independent Lognormals with parameters mu_j and sd_j:
#   S_sys(t) = prod_j stats::plnorm(t, meanlog = mu_j, sdlog = sd_j,
#                                    lower.tail = FALSE)
# Sampler generates m independent Lognormals and takes the min.
# ==========================================================================


#' Series of independent Lognormal components (closed form)
#'
#' Constructs a `dist_structure` for a series system whose components are
#' independent Lognormals. Closed-form `surv` is the product of
#' per-component upper-tail probabilities; `cdf` is `1 - surv`; `sampler`
#' generates m independent Lognormals and takes the min.
#'
#' @param meanlogs Numeric vector of length `m`: per-component meanlog
#'   parameters.
#' @param sdlogs Positive numeric vector of length `m`: per-component
#'   sdlog parameters.
#' @return
#' `lognormal_series()` returns an object of class
#'   `c("lognormal_series", "series_dist", "coherent_dist",
#'   "dist_structure", "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' @examples
#' sys <- lognormal_series(meanlogs = c(0, 1), sdlogs = c(1, 0.5))
#' algebraic.dist::surv(sys)(1)
#' @export
lognormal_series <- function(meanlogs, sdlogs) {
  stopifnot(length(meanlogs) == length(sdlogs), all(sdlogs > 0))
  m <- length(meanlogs)
  components <- lapply(seq_len(m), function(j) {
    algebraic.dist::lognormal(meanlog = meanlogs[j], sdlog = sdlogs[j])
  })
  obj <- series_dist(components)
  obj$meanlogs <- as.numeric(meanlogs)
  obj$sdlogs <- as.numeric(sdlogs)
  class(obj) <- c("lognormal_series", class(obj))
  obj
}


#' @rdname lognormal_series
#' @param x A `lognormal_series` object.
#' @param ... Ignored.
#' @export
surv.lognormal_series <- function(x, ...) {
  series_surv_product(stats::plnorm,
                      list(meanlog = x$meanlogs, sdlog = x$sdlogs))
}


#' @rdname lognormal_series
#' @export
sampler.lognormal_series <- function(x, ...) {
  samplers <- make_component_samplers(stats::rlnorm,
                                      meanlog = x$meanlogs,
                                      sdlog = x$sdlogs)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L, min)
  }
}

# ==========================================================================
# General coherent_dist + topology shortcut constructors
# ==========================================================================


#' Coherent system distribution from minimal path sets
#'
#' General-purpose constructor. Users supply a list of minimal path sets
#' (each a vector of component indices) and a list of component
#' distributions (each an `algebraic.dist::dist` object with parameters
#' baked in). The resulting object is a `dist_structure` and `dist`.
#'
#' @param min_paths List of integer vectors; each is a minimal path set.
#' @param components List of `dist` objects, length `m`. Each is a
#'   fully-parameterized component lifetime distribution.
#' @param m Optional integer. Inferred from `components` and `min_paths`
#'   if omitted.
#' @return An object of class
#'   `c("coherent_dist", "dist_structure", "univariate_dist", "dist")`.
#' @examples
#' # A bridge network with exponential components
#' sys <- coherent_dist(
#'   min_paths = list(c(1, 4), c(2, 5), c(1, 3, 5), c(2, 3, 4)),
#'   components = replicate(5, algebraic.dist::exponential(1), simplify = FALSE)
#' )
#' reliability(sys, 0.9)
#' @export
coherent_dist <- function(min_paths, components, m = NULL) {
  stopifnot(is.list(min_paths), is.list(components))
  is_dist <- vapply(components, function(d) inherits(d, "dist"),
                    logical(1L))
  if (!all(is_dist)) {
    bad <- which(!is_dist)
    stop(sprintf(
      "components must all inherit from 'dist'; offending positions: %s",
      paste(bad, collapse = ", ")
    ), call. = FALSE)
  }
  if (is.null(m)) {
    all_idx <- unlist(min_paths)
    m <- max(length(components),
             if (length(all_idx)) max(all_idx) else 0L)
  }
  stopifnot(length(components) == m)
  structure(
    list(
      min_paths = lapply(min_paths, as.integer),
      m = as.integer(m),
      components = components
    ),
    class = c("coherent_dist", "dist_structure",
              "univariate_dist", "continuous_dist", "dist")
  )
}


#' @export
ncomponents.coherent_dist <- function(x) x$m


#' @export
component.coherent_dist <- function(x, j, ...) {
  stopifnot(j >= 1L, j <= x$m)
  x$components[[j]]
}


#' @export
min_paths.coherent_dist <- function(x) x$min_paths


#' Dual of a coherent_dist: swap cuts and paths
#'
#' Overrides the lazy-wrapper default with a proper `coherent_dist`:
#' `min_paths(dual(x)) = min_cuts(x)`.
#'
#' @rdname dual
#' @export
dual.coherent_dist <- function(x) {
  coherent_dist(
    min_paths = min_cuts(x),
    components = x$components,
    m = x$m
  )
}


# ==========================================================================
# Topology shortcut constructors
# ==========================================================================


#' Series system distribution
#'
#' A series system fails if any single component fails. Equivalent to
#' `min(components)` as random variables; unlike `min` in the base algebra,
#' `series_dist` preserves the component decomposition so topology queries
#' work.
#'
#' @param components List of `dist` objects.
#' @return A `series_dist` inheriting from `coherent_dist`.
#' @examples
#' sys <- series_dist(replicate(3, algebraic.dist::exponential(1), simplify = FALSE))
#' algebraic.dist::surv(sys)(0.5)
#' @export
series_dist <- function(components) {
  stopifnot(is.list(components), length(components) >= 1L)
  m <- length(components)
  obj <- coherent_dist(
    min_paths = list(seq_len(m)),
    components = components,
    m = m
  )
  class(obj) <- c("series_dist", class(obj))
  obj
}


#' @export
phi.series_dist <- function(x, state) {
  stopifnot(length(state) == x$m)
  as.integer(all(state == 1L))
}

#' @export
min_paths.series_dist <- function(x) list(seq_len(x$m))


# Override the dist_structure default (which evaluates the 2^m-state
# reliability polynomial via min_path/min_cut enumeration) with the
# closed form S_sys(t) = prod_j S_j(t). This avoids the exponential blow-up
# for series of arbitrary components when no closed-form family-specific
# specialization (exp_series, wei_series, etc.) applies.
#' @export
surv.series_dist <- function(x, ...) {
  m <- ncomponents(x)
  comp_surv_fns <- lapply(seq_len(m), function(j) {
    algebraic.dist::surv(component(x, j))
  })
  function(t, ...) {
    vapply(t, function(ti) {
      prod(vapply(comp_surv_fns, function(S_j) S_j(ti), numeric(1L)))
    }, numeric(1L))
  }
}


#' Parallel system distribution
#'
#' A parallel system fails only when all components fail. Equivalent to
#' `max(components)` but preserves topology.
#'
#' @param components List of `dist` objects.
#' @return A `parallel_dist` inheriting from `coherent_dist`.
#' @export
parallel_dist <- function(components) {
  stopifnot(is.list(components), length(components) >= 1L)
  m <- length(components)
  obj <- coherent_dist(
    min_paths = as.list(seq_len(m)),
    components = components,
    m = m
  )
  class(obj) <- c("parallel_dist", class(obj))
  obj
}


#' @export
phi.parallel_dist <- function(x, state) {
  stopifnot(length(state) == x$m)
  as.integer(any(state == 1L))
}

#' @export
min_paths.parallel_dist <- function(x) as.list(seq_len(x$m))


# Closed-form override: S_sys(t) = 1 - prod_j F_j(t).  Avoids the 2^m
# reliability-polynomial enumeration for parallel systems whose component
# family is not covered by exp_parallel.
#' @export
surv.parallel_dist <- function(x, ...) {
  m <- ncomponents(x)
  comp_cdf_fns <- lapply(seq_len(m), function(j) {
    algebraic.dist::cdf(component(x, j))
  })
  function(t, ...) {
    vapply(t, function(ti) {
      1 - prod(vapply(comp_cdf_fns, function(F_j) F_j(ti), numeric(1L)))
    }, numeric(1L))
  }
}


#' k-out-of-n system distribution
#'
#' A k-out-of-n system functions if at least `k` of its `m` components
#' function. Equivalent to the `(m - k + 1)`-th order statistic of
#' component lifetimes.
#'
#' This constructor uses the **:G** convention: `k` is the number of
#' components that must remain **functioning** for the system to function.
#' `k = 1` is parallel; `k = m` is series. The companion `kofn` package
#' (which depends on `dist.structure`) uses the **:F** convention, where
#' `k` is the number of failures that trigger system failure; conversion
#' is `k_dist = m - k_kofn + 1`. When in doubt, draw a small example:
#' `kofn_dist(k = 2, ...)` for `m = 3` functions until two of the three
#' components have failed.
#'
#' @param k Minimum functioning components for system operation (:G).
#' @param components List of `dist` objects (length `m`).
#' @return A `kofn_dist` inheriting from `coherent_dist`.
#' @seealso [order_statistic()] for the closely-related order-statistic
#'   parameterization.
#' @export
kofn_dist <- function(k, components) {
  stopifnot(is.list(components), length(components) >= 1L)
  m <- length(components)
  stopifnot(k >= 1L, k <= m)
  paths <- utils::combn(m, k, simplify = FALSE)
  obj <- coherent_dist(
    min_paths = paths,
    components = components,
    m = m
  )
  obj$k <- as.integer(k)
  class(obj) <- c("kofn_dist", class(obj))
  obj
}


#' @export
phi.kofn_dist <- function(x, state) as.integer(sum(state) >= x$k)

# min_paths.kofn_dist is intentionally absent: kofn_dist() stores
# combn(m, k) as $min_paths at construction time, so the inherited
# min_paths.coherent_dist (which returns x$min_paths directly) is both
# correct and faster than recomputing combn on every call.


#' Bridge system distribution
#'
#' The classical 5-component bridge reliability network with minimal
#' path sets `{1,4}`, `{2,5}`, `{1,3,5}`, `{2,3,4}`. Components 1 and 2
#' are the input side, 4 and 5 the output side, and 3 the cross-link.
#' The bridge is a canonical non-series, non-parallel example used
#' throughout the reliability literature; see Barlow and Proschan (1975,
#' "Statistical Theory of Reliability and Life Testing") for the
#' standard treatment.
#'
#' @param components List of 5 `dist` objects.
#' @return A `bridge_dist` inheriting from `coherent_dist`.
#' @export
bridge_dist <- function(components) {
  stopifnot(is.list(components), length(components) == 5L)
  paths <- list(
    c(1L, 4L), c(2L, 5L),
    c(1L, 3L, 5L), c(2L, 3L, 4L)
  )
  obj <- coherent_dist(
    min_paths = paths,
    components = components,
    m = 5L
  )
  class(obj) <- c("bridge_dist", class(obj))
  obj
}


#' Consecutive-k-out-of-n system distribution (type G)
#'
#' The consecutive-k-out-of-n:G system functions when at least one block
#' of `k` consecutive components all function. Minimal path sets are the
#' `n - k + 1` consecutive blocks of size `k`. (Note: this is the :G
#' variant; the :F variant, "fails when any `k` consecutive fail", has
#' different minimal paths.)
#'
#' @param k Block size.
#' @param components List of `dist` objects (length `n`).
#' @return A `consecutive_k_dist` inheriting from `coherent_dist`.
#' @export
consecutive_k_dist <- function(k, components) {
  n <- length(components)
  stopifnot(k >= 1L, k <= n)
  paths <- lapply(seq_len(n - k + 1L), function(i) as.integer(i:(i + k - 1L)))
  obj <- coherent_dist(
    min_paths = paths,
    components = components,
    m = n
  )
  class(obj) <- c("consecutive_k_dist", class(obj))
  obj
}

# ==========================================================================
# system_signature default
# ==========================================================================


test_that("series signature is (1, 0, ..., 0)", {
  sys <- series_dist(iid_exp_components(4))
  sig <- system_signature(sys)
  expect_equal(sig, c(1, 0, 0, 0), tolerance = 1e-12)
})


test_that("parallel signature is (0, ..., 0, 1)", {
  sys <- parallel_dist(iid_exp_components(4))
  sig <- system_signature(sys)
  expect_equal(sig, c(0, 0, 0, 1), tolerance = 1e-12)
})


test_that("signature sums to 1 for k-of-n", {
  for (m in 3:5) for (k in 1:m) {
    sys <- kofn_dist(k, iid_exp_components(m))
    sig <- system_signature(sys)
    expect_length(sig, m)
    expect_equal(sum(sig), 1, tolerance = 1e-12)
    # k-of-n fails at the (m - k + 1)-th failure, so all mass is there
    expected <- integer(m)
    expected[m - k + 1L] <- 1L
    expect_equal(sig, expected, tolerance = 1e-12)
  }
})


test_that("bridge signature is the standard (0, 1/5, 3/5, 1/5, 0)", {
  # Classical result for the bridge network.
  sys <- bridge_dist(iid_exp_components(5))
  sig <- system_signature(sys)
  expect_equal(sig, c(0, 1/5, 3/5, 1/5, 0), tolerance = 1e-12)
})

# ==========================================================================
# Reliability importance measures
# ==========================================================================


test_that("birnbaum_importance for series equals prod of other reliabilities", {
  sys <- series_dist(iid_exp_components(4))
  p <- c(0.9, 0.8, 0.7, 0.6)
  for (j in 1:4) {
    expected <- prod(p[-j])
    expect_equal(birnbaum_importance(sys, j, p), expected, tolerance = 1e-10)
  }
})


test_that("birnbaum_importance for parallel equals prod of other unreliabilities", {
  sys <- parallel_dist(iid_exp_components(4))
  p <- c(0.9, 0.8, 0.7, 0.6)
  for (j in 1:4) {
    expected <- prod(1 - p[-j])
    expect_equal(birnbaum_importance(sys, j, p), expected, tolerance = 1e-10)
  }
})


test_that("birnbaum_importance for kofn equals binomial tail derivative at iid p", {
  # For k-of-m at iid p: R(p) = sum_{i=k}^{m} C(m, i) p^i (1-p)^(m-i).
  # dR/dp_j (same for each j by symmetry) = ...
  # Check: I_B for 2-of-3 at p = 0.5 equals
  #   P(sum of other two = k-1 = 1) = 2 * 0.5 * 0.5 = 0.5
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  expect_equal(birnbaum_importance(sys, 1L, 0.5), 0.5, tolerance = 1e-10)
  # 2-of-3 at p = 0.8: P(exactly 1 of other 2 alive) = 2 * 0.8 * 0.2 = 0.32
  expect_equal(birnbaum_importance(sys, 1L, 0.8), 0.32, tolerance = 1e-10)
})


test_that("birnbaum_importance scalar p recycles", {
  sys <- series_dist(iid_exp_components(3))
  expect_equal(
    birnbaum_importance(sys, 1L, 0.5),
    birnbaum_importance(sys, 1L, c(0.5, 0.5, 0.5)),
    tolerance = 1e-12
  )
})


test_that("criticality_importance equals Birnbaum * F_j / F_sys for series", {
  # Series of iid Exp(1): F_j(t) = 1 - exp(-t); F_sys(t) = 1 - exp(-m*t).
  # I_B(j; p) at iid p is p^(m-1).
  m <- 3L
  sys <- exp_series(rep(1, m))
  t0 <- 0.5
  p <- exp(-t0)  # common survival at t0
  ib <- p^(m - 1L)
  F_j <- 1 - p
  F_sys <- 1 - exp(-m * t0)
  expected <- ib * F_j / F_sys
  expect_equal(criticality_importance(sys, 1L, t0), expected,
               tolerance = 1e-10)
})


test_that("vesely_fussell_importance equals F_j / F_sys for series", {
  # For a series system, every component j is a single-component cut, so
  # the cuts containing j are just {{j}}. The union of events "cut {j}
  # failed" is just "component j failed", with probability F_j.
  # I_VF(j; t) = F_j / F_sys.
  m <- 3L
  sys <- exp_series(rep(1, m))
  t0 <- 0.5
  F_j <- 1 - exp(-t0)
  F_sys <- 1 - exp(-m * t0)
  expect_equal(vesely_fussell_importance(sys, 1L, t0), F_j / F_sys,
               tolerance = 1e-10)
})


test_that("vesely_fussell_importance returns 0 when component is in no cut", {
  # Pathological case: ensure correct 0 return when no cuts contain j.
  # This shouldn't happen for a coherent system unless j is irrelevant.
  # Test vacuously by choosing j for a parallel system of 1 component.
  # For 1-component parallel, min_cuts = {{1}}, so cuts containing 1 do exist.
  # Skip the vacuous check; instead, verify finiteness on realistic inputs.
  sys <- parallel_dist(iid_exp_components(3))
  t0 <- 1
  vf <- vesely_fussell_importance(sys, 1L, t0)
  expect_true(is.finite(vf))
  expect_true(vf >= 0 && vf <= 1)
})


test_that("importance measures handle bridge without error", {
  sys <- bridge_dist(iid_exp_components(5))
  t0 <- 1
  for (j in 1:5) {
    ib <- birnbaum_importance(sys, j, exp(-t0))
    ci <- criticality_importance(sys, j, t0)
    vf <- vesely_fussell_importance(sys, j, t0)
    expect_true(is.finite(ib) && ib >= 0 && ib <= 1)
    expect_true(is.finite(ci) && ci >= 0 && ci <= 1)
    expect_true(is.finite(vf) && vf >= 0 && vf <= 1)
  }
})

# ==========================================================================
# Distribution interface: dist_structure IS-A dist
# ==========================================================================


test_that("series_dist objects are dists", {
  sys <- series_dist(iid_exp_components(3))
  expect_true(algebraic.dist::is_dist(sys))
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("surv(series_dist) equals product of component survs", {
  # Series of 3 iid Exp(1): S_sys(t) = exp(-3t)
  sys <- series_dist(iid_exp_components(3, rate = 1))
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.1, 0.5, 1.0, 2.0)) {
    expect_equal(S(ti), exp(-3 * ti), tolerance = 1e-10)
  }
})


test_that("surv(parallel_dist) equals 1 - product of component cdfs", {
  # Parallel of 3 iid Exp(1): F_sys(t) = (1 - exp(-t))^3, S = 1 - F
  sys <- parallel_dist(iid_exp_components(3, rate = 1))
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.1, 0.5, 1.0, 2.0)) {
    expect_equal(S(ti), 1 - (1 - exp(-ti))^3, tolerance = 1e-10)
  }
})


test_that("cdf equals 1 - surv", {
  sys <- series_dist(iid_exp_components(2))
  S <- algebraic.dist::surv(sys)
  F_fn <- algebraic.dist::cdf(sys)
  for (ti in c(0.3, 0.7, 1.5)) {
    expect_equal(F_fn(ti), 1 - S(ti), tolerance = 1e-12)
  }
})


test_that("sampler returns the right number of samples", {
  sys <- series_dist(iid_exp_components(3, rate = 1))
  samp <- algebraic.dist::sampler(sys)
  withr::local_seed(1)
  x <- samp(100)
  expect_length(x, 100L)
  expect_true(all(x >= 0))
})


test_that("sampler of series iid Exp(rate) has approximately the right mean", {
  # Series of m iid Exp(rate) is Exp(m * rate). Mean = 1/(m*rate).
  m <- 3L
  rate <- 2
  sys <- series_dist(iid_exp_components(m, rate = rate))
  withr::local_seed(42)
  x <- algebraic.dist::sampler(sys)(10000)
  expected_mean <- 1 / (m * rate)
  expect_equal(mean(x), expected_mean, tolerance = 0.05)
})


test_that("sampler of parallel iid has mean close to harmonic-series formula", {
  # Max of m iid Exp(1): E[max] = 1 + 1/2 + ... + 1/m
  m <- 4L
  sys <- parallel_dist(iid_exp_components(m, rate = 1))
  withr::local_seed(42)
  x <- algebraic.dist::sampler(sys)(10000)
  expected_mean <- sum(1 / (1:m))
  expect_equal(mean(x), expected_mean, tolerance = 0.05)
})


test_that("surv consistent with reliability identity S_sys(t) = R(S(t))", {
  sys <- bridge_dist(iid_exp_components(5, rate = 1))
  S_sys <- algebraic.dist::surv(sys)
  for (ti in c(0.2, 0.8, 1.5)) {
    # Manual computation
    p <- exp(-ti)
    expected <- reliability(sys, p)
    expect_equal(S_sys(ti), expected, tolerance = 1e-10)
  }
})

# ==========================================================================
# Parallel and k-of-n closed-form specializations
# ==========================================================================


test_that("exp_parallel surv matches the classical product-complement formula", {
  rates <- c(1, 2, 3)
  sys <- exp_parallel(rates)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.2, 0.7, 1.5, 3)) {
    expected <- 1 - prod(1 - exp(-rates * ti))
    expect_equal(S(ti), expected, tolerance = 1e-12)
  }
})


test_that("exp_parallel mean matches harmonic sum for iid", {
  # iid case with m components and common rate lambda:
  # E[max] = (1/lambda) * (1 + 1/2 + ... + 1/m)
  for (m in 2:5) {
    lambda <- 2
    sys <- exp_parallel(rep(lambda, m))
    expected <- sum(1 / seq_len(m)) / lambda
    expect_equal(mean(sys), expected, tolerance = 1e-10)
  }
})


test_that("exp_parallel mean matches sampler empirical mean", {
  rates <- c(0.5, 1, 2)
  sys <- exp_parallel(rates)
  withr::local_seed(1)
  x <- algebraic.dist::sampler(sys)(10000)
  expect_equal(mean(x), mean(sys), tolerance = 0.05)
})


test_that("exp_parallel class hierarchy is correct", {
  sys <- exp_parallel(c(1, 2))
  expect_s3_class(sys, "exp_parallel")
  expect_s3_class(sys, "parallel_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("exp_kofn reduces to exp_series for k = m", {
  rates <- c(0.5, 0.3, 0.2)
  sys_kofn <- exp_kofn(k = 3L, rates)
  sys_series <- exp_series(rates)
  S1 <- algebraic.dist::surv(sys_kofn)
  S2 <- algebraic.dist::surv(sys_series)
  for (ti in c(0.3, 1, 2)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-10)
  }
})


test_that("exp_kofn reduces to exp_parallel for k = 1", {
  rates <- c(1, 2, 3)
  sys_kofn <- exp_kofn(k = 1L, rates)
  sys_par <- exp_parallel(rates)
  S1 <- algebraic.dist::surv(sys_kofn)
  S2 <- algebraic.dist::surv(sys_par)
  for (ti in c(0.2, 0.8, 2)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-10)
  }
})


test_that("exp_kofn sampler empirical mean matches k-th order statistic (iid)", {
  # For iid Exp(1), the i-th smallest value has E[T_(i)] = sum_{j=1}^{i} 1/(m-j+1).
  # A k-of-m system fails at the (m-k+1)-th smallest time (= the i=m-k+1 order stat).
  m <- 5L; k <- 3L
  sys <- exp_kofn(k, rep(1, m))
  withr::local_seed(1)
  x <- algebraic.dist::sampler(sys)(10000)
  i <- m - k + 1L
  expected_mean <- sum(1 / ((m - seq_len(i) + 1L)))
  expect_equal(mean(x), expected_mean, tolerance = 0.05)
})


test_that("exp_kofn class hierarchy is correct", {
  sys <- exp_kofn(k = 2L, rates = c(1, 2, 3))
  expect_s3_class(sys, "exp_kofn")
  expect_s3_class(sys, "kofn_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("wei_kofn reduces to wei_series for k = m", {
  shapes <- c(2, 2, 2); scales <- c(1, 2, 3)
  sys_kofn <- wei_kofn(3L, shapes, scales)
  sys_series <- wei_series(shapes, scales)
  S1 <- algebraic.dist::surv(sys_kofn)
  S2 <- algebraic.dist::surv(sys_series)
  for (ti in c(0.3, 1, 2)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-10)
  }
})


test_that("wei_kofn sampler survival probability matches closed-form surv", {
  shapes <- c(2, 2, 2); scales <- c(1, 2, 3)
  sys <- wei_kofn(2L, shapes, scales)
  withr::local_seed(1)
  x <- algebraic.dist::sampler(sys)(10000)
  for (t0 in c(0.3, 0.8, 1.5)) {
    emp_surv <- mean(x > t0)
    analytical_surv <- algebraic.dist::surv(sys)(t0)
    expect_equal(emp_surv, analytical_surv, tolerance = 0.02)
  }
})


test_that("wei_kofn class hierarchy is correct", {
  sys <- wei_kofn(k = 2L, shapes = c(1, 2), scales = c(1, 2))
  expect_s3_class(sys, "wei_kofn")
  expect_s3_class(sys, "kofn_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
})


test_that("exp_parallel and exp_kofn agree on surv at k = 1", {
  rates <- runif(4, 0.1, 2)
  sys_par <- exp_parallel(rates)
  sys_kofn <- exp_kofn(1L, rates)
  for (ti in c(0.2, 1, 3)) {
    expect_equal(
      algebraic.dist::surv(sys_par)(ti),
      algebraic.dist::surv(sys_kofn)(ti),
      tolerance = 1e-10
    )
  }
})


test_that("density.exp_kofn integrates to 1", {
  for (params in list(list(k = 2L, rates = c(1, 1, 1)),
                      list(k = 1L, rates = c(0.5, 1)),
                      list(k = 3L, rates = c(0.5, 1, 2, 0.7)))) {
    sys <- exp_kofn(params$k, params$rates)
    f <- density(sys)
    integral <- stats::integrate(f, 0, Inf, rel.tol = 1e-8)$value
    expect_equal(integral, 1, tolerance = 1e-4)
  }
})


test_that("density.exp_kofn equals -d/dt surv numerically", {
  sys <- exp_kofn(2L, c(0.5, 1, 1.5))
  f <- density(sys)
  S <- algebraic.dist::surv(sys)
  for (t0 in c(0.3, 0.8, 1.5)) {
    h <- 1e-5
    df_dt <- -(S(t0 + h) - S(t0 - h)) / (2 * h)
    expect_equal(f(t0), df_dt, tolerance = 1e-4)
  }
})


test_that("density.exp_kofn at k = m matches dexp at sum_rates", {
  rates <- c(0.5, 0.3, 0.2)
  sys_kofn <- exp_kofn(3L, rates)
  sys_series <- exp_series(rates)
  f1 <- density(sys_kofn)
  f2 <- density(sys_series)
  for (t0 in c(0.5, 1, 2)) {
    expect_equal(f1(t0), f2(t0), tolerance = 1e-10)
  }
})


test_that("density.exp_kofn at k = 1 matches max-of-iid density", {
  # k=1 is parallel: f_sys = d/dt prod F_j; for iid Exp(1), m components:
  # F(t) = 1 - exp(-t), F_sys = (1 - exp(-t))^m,
  # f_sys = m * (1 - exp(-t))^(m-1) * exp(-t)
  rates <- rep(1, 3)
  sys <- exp_kofn(1L, rates)
  f <- density(sys)
  for (t0 in c(0.5, 1, 2)) {
    expected <- 3 * (1 - exp(-t0))^2 * exp(-t0)
    expect_equal(f(t0), expected, tolerance = 1e-10)
  }
})


test_that("density.wei_kofn integrates to 1", {
  for (params in list(list(k = 2L, shapes = c(1.5, 2, 2.5),
                            scales = c(1, 2, 3)),
                       list(k = 1L, shapes = c(2, 2),
                            scales = c(1, 2)))) {
    sys <- wei_kofn(params$k, params$shapes, params$scales)
    f <- density(sys)
    integral <- stats::integrate(f, 0, Inf, rel.tol = 1e-8)$value
    expect_equal(integral, 1, tolerance = 1e-4)
  }
})


test_that("density.wei_kofn equals -d/dt surv numerically", {
  sys <- wei_kofn(2L, shapes = c(2, 2, 2), scales = c(1, 2, 3))
  f <- density(sys)
  S <- algebraic.dist::surv(sys)
  for (t0 in c(0.3, 0.8, 1.5)) {
    h <- 1e-5
    df_dt <- -(S(t0 + h) - S(t0 - h)) / (2 * h)
    expect_equal(f(t0), df_dt, tolerance = 1e-4)
  }
})


test_that("density methods support log = TRUE", {
  sys <- exp_kofn(2L, c(1, 1, 1))
  f <- density(sys)
  for (t0 in c(0.5, 1, 2)) {
    expect_equal(f(t0, log = TRUE), log(f(t0)), tolerance = 1e-12)
  }
})


test_that("exp_kofn surv matches the general default (small m)", {
  rates <- c(1, 2)
  sys_special <- exp_kofn(1L, rates)
  # Compare against the general kofn_dist default
  sys_general <- kofn_dist(1L,
    lapply(rates, function(r) algebraic.dist::exponential(r)))
  for (ti in c(0.3, 1, 2)) {
    expect_equal(algebraic.dist::surv(sys_special)(ti),
                 algebraic.dist::surv(sys_general)(ti),
                 tolerance = 1e-10)
  }
})

# ==========================================================================
# Closed-form parallel of exponential components
# ==========================================================================
#
# A parallel system of m independent exponentials with rates lambda_j has
#   F_sys(t) = prod_j (1 - exp(-lambda_j t))
#   S_sys(t) = 1 - F_sys(t)
# For iid components, E[max] = (1/lambda)(1 + 1/2 + ... + 1/m).
# For the heterogeneous case, E[max] = sum over nonempty subsets S of
# (-1)^(|S|+1) / sum(rates[S]) via inclusion-exclusion.
# ==========================================================================


#' Parallel of exponential components (closed form)
#'
#' Constructs a `dist_structure` representing a parallel system whose
#' components are independent exponentials. Closed-form methods are
#' provided for `surv`, `cdf`, `sampler`, and `mean` (the last via
#' inclusion-exclusion over the `2^m - 1` non-empty component subsets).
#'
#' @param rates Positive numeric vector of length `m`.
#' @return
#' `exp_parallel()` returns an object of class
#'   `c("exp_parallel", "parallel_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' - `mean()`: a numeric scalar (the mean system lifetime,
#'   computed in closed form via inclusion-exclusion).
#' @examples
#' sys <- exp_parallel(c(1, 2, 3))
#' algebraic.dist::surv(sys)(1)
#' @export
exp_parallel <- function(rates) {
  stopifnot(is.numeric(rates), length(rates) >= 1L, all(rates > 0))
  components <- lapply(rates, algebraic.dist::exponential)
  obj <- parallel_dist(components)
  obj$rates <- as.numeric(rates)
  class(obj) <- c("exp_parallel", class(obj))
  obj
}


#' @rdname exp_parallel
#' @param x An `exp_parallel` object.
#' @param ... Ignored.
#' @export
surv.exp_parallel <- function(x, ...) {
  rates <- x$rates
  function(t, ...) {
    vapply(t, function(ti) 1 - prod(1 - exp(-rates * ti)), numeric(1L))
  }
}


#' @rdname exp_parallel
#' @export
sampler.exp_parallel <- function(x, ...) {
  samplers <- make_component_samplers(stats::rexp, rate = x$rates)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L, max)
  }
}


#' @rdname exp_parallel
#' @export
mean.exp_parallel <- function(x, ...) {
  rates <- x$rates
  m <- length(rates)
  # Fast path: iid components. E[max of m iid Exp(lambda)] = H_m / lambda.
  if (length(unique(rates)) == 1L) {
    return(sum(1 / seq_len(m)) / rates[[1L]])
  }
  # General heterogeneous case: inclusion-exclusion over 2^m - 1 subsets.
  total <- 0
  for (sz in seq_len(m)) {
    subsets <- utils::combn(m, sz, simplify = FALSE)
    for (S in subsets) {
      total <- total + (-1)^(sz + 1L) / sum(rates[S])
    }
  }
  total
}

#' dist.structure: Structured Random Variables for Reliability System
#' Distributions
#'
#' Extends the algebraic.dist distribution algebra to random variables with
#' internal structure: coherent reliability systems decomposed into components
#' arranged by a structure function (series, parallel, k-out-of-n, bridge,
#' arbitrary topologies). Every `dist_structure` object is also a `dist`, so
#' the full distribution algebra works automatically via default methods that
#' compose component-level distributions through the topology.
#'
#' @section Two algebras, one hierarchy:
#'
#' `algebraic.dist` provides the base algebra of random variables:
#' `density`, `surv`, `hazard`, `sampler`, `mean`, `vcov`, `expectation`,
#' and arithmetic on RVs (transformations).
#'
#' `dist.structure` extends it with structure-preserving operations for
#' reliability systems: component-wise decomposition, structure function
#' evaluation, path and cut sets, system signature, critical states, dual
#' structures, and Birnbaum importance measures. Because every
#' `dist_structure` is a `dist`, the base algebra inherits automatically.
#'
#' @section Primitives implementors provide:
#'
#' - [ncomponents()]: number of components `m`
#' - [component()]: return the j-th component as a `dist`
#' - One of [phi()] or [min_paths()]: the other defaults automatically
#'
#' @section Topology queries (default methods on `dist_structure`):
#'
#' [phi()], [min_paths()], [min_cuts()], [critical_states()],
#' [system_signature()], [system_lifetime()], [system_censoring()],
#' [dual()], [is_coherent()], [structural_importance()], [reliability()]
#'
#' @section Distribution queries (default methods via component composition):
#'
#' [surv()], [cdf()], [sampler()]. Further base dist methods (`mean`,
#' `vcov`, `expectation`) inherit from `univariate_dist` via numerical
#' integration of `surv`. Implementors override any default for speed.
#'
#' @section Reference implementations:
#'
#' - [coherent_dist()]: general coherent system given minimal path sets +
#'   components
#' - Topology shortcuts: [series_dist()], [parallel_dist()], [kofn_dist()],
#'   [bridge_dist()], [consecutive_k_dist()]
#' - IID convenience: [min_iid()], [max_iid()], [order_statistic()]
#' - Coercions: [as_dist_structure()]
#'
#' @section Closed-form specializations:
#'
#' Fast analytical paths for common (component family, topology) pairs.
#' Override the generic defaults for speed without changing semantics:
#'
#' - Series: [exp_series()], [wei_series()], [wei_homogeneous_series()],
#'   [gamma_series()], [lognormal_series()]
#' - Parallel: [exp_parallel()]
#' - k-out-of-n: [exp_kofn()], [wei_kofn()]
#'
#' @section Importance measures:
#'
#' Both topology-only and reliability-aware:
#'
#' - [structural_importance()]: Birnbaum's structural measure (topology only)
#' - [birnbaum_importance()]: Birnbaum's reliability measure
#' - [criticality_importance()]: criticality importance
#' - [vesely_fussell_importance()]: Vesely-Fussell importance
#'
#' @section Compositional operations:
#'
#' Build new systems from existing ones while preserving the protocol:
#'
#' - [substitute_component()]: replace component j with a sub-system
#' - [compose_systems()]: compose two systems (cartesian product of paths)
#'
#' @section Non-coherent systems:
#'
#' [cold_standby_dist()] models a cold-standby spare arrangement (system
#' lifetime is the sum of component lifetimes). It is deliberately *not*
#' a `dist_structure` (no static structure function), but it shares the
#' [ncomponents()]/[component()] generics so user code can iterate
#' uniformly.
#'
#' @keywords internal
#' @importFrom algebraic.dist cdf sampler surv
"_PACKAGE"

# ==========================================================================
# Closed-form series of Weibull components
# ==========================================================================
#
# For a series of m independent Weibulls with shapes k_j and scales s_j:
#   Cumulative hazard of component j at time t: H_j(t) = (t / s_j)^{k_j}
#   System survival:  S_sys(t) = exp(-sum_j (t / s_j)^{k_j})
#   System hazard:    h_sys(t) = sum_j (k_j / s_j) * (t / s_j)^{k_j - 1}
#   System density:   f_sys(t) = h_sys(t) * S_sys(t)
#
# No closed form for the CDF inverse in general; sampler uses min of
# independently drawn Weibulls.
# ==========================================================================


#' Series of heterogeneous Weibull components (closed form)
#'
#' Constructs a `dist_structure` representing a series system whose
#' components are independent Weibull distributions with possibly
#' different shapes and scales. Closed-form methods are provided for
#' `surv`, `cdf`, `sampler`, and `algebraic.dist::hazard`.
#'
#' @param shapes Positive numeric vector of length `m`: Weibull shape
#'   parameters per component.
#' @param scales Positive numeric vector of length `m` (same length as
#'   `shapes`): Weibull scale parameters per component.
#' @return
#' `wei_series()` returns an object of class
#'   `c("wei_series", "series_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`, `hazard()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' @examples
#' sys <- wei_series(shapes = c(1, 2, 3), scales = c(1, 2, 3))
#' algebraic.dist::surv(sys)(1)
#' @export
wei_series <- function(shapes, scales) {
  stopifnot(length(shapes) == length(scales),
            all(shapes > 0), all(scales > 0))
  m <- length(shapes)
  components <- lapply(seq_len(m), function(j) {
    algebraic.dist::weibull_dist(shape = shapes[j], scale = scales[j])
  })
  obj <- series_dist(components)
  obj$shapes <- as.numeric(shapes)
  obj$scales <- as.numeric(scales)
  class(obj) <- c("wei_series", class(obj))
  obj
}


#' @rdname wei_series
#' @param x A `wei_series` object.
#' @param ... Ignored.
#' @export
surv.wei_series <- function(x, ...) {
  # exp(-sum((t/scale_j)^shape_j)) = prod_j exp(-(t/scale_j)^shape_j)
  # = prod_j pweibull(t, shape_j, scale_j, lower.tail = FALSE).
  # Using series_surv_product matches the gamma_series and lognormal_series
  # implementations.
  series_surv_product(stats::pweibull,
                      list(shape = x$shapes, scale = x$scales))
}


#' @rdname wei_series
#' @export
sampler.wei_series <- function(x, ...) {
  samplers <- make_component_samplers(stats::rweibull,
                                      shape = x$shapes, scale = x$scales)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L, min)
  }
}


# Closed-form Weibull hazard: h_sys(t) = sum_j (k_j / s_j) * (t / s_j)^(k_j - 1).
# Series hazards are additive, and each component's Weibull hazard has a
# direct algebraic form, so the system hazard avoids both numerical
# differentiation and the per-component dispatch cost of the algebraic.dist
# fallback.
#' @rdname wei_series
#' @method hazard wei_series
#' @importFrom algebraic.dist hazard
#' @export
hazard.wei_series <- function(x, ...) {
  shapes <- x$shapes
  scales <- x$scales
  function(t, ...) {
    vapply(t, function(ti) {
      sum((shapes / scales) * (ti / scales)^(shapes - 1))
    }, numeric(1L))
  }
}

# ==========================================================================
# Coercions
# ==========================================================================
#
# as_dist_structure(x): promote a plain dist (e.g., from algebraic.dist)
# into a 1-component dist_structure (a trivial series of one component).
# Useful for polymorphic functions that accept dist_structure inputs but
# are also handed plain dists.
# ==========================================================================


#' @rdname as_dist_structure
#' @export
as_dist_structure.dist_structure <- function(x, ...) x


#' @rdname as_dist_structure
#' @export
as_dist_structure.dist <- function(x, ...) {
  series_dist(list(x))
}

# ==========================================================================
# Closed-form series of Weibulls with common shape
# ==========================================================================
#
# When all components share shape k, the system cumulative hazard is:
#   sum_j (t / s_j)^k  =  t^k * sum_j (1 / s_j^k)
# Let s_agg = ( sum_j (1 / s_j^k) )^{-1/k}
# Then the series lifetime is itself Weibull(shape = k, scale = s_agg).
#
# This identity means wei_homogeneous_series reduces exactly to a single
# Weibull. The constructor preserves the original component information
# (so component(x, j) and topology queries still work) while exposing the
# aggregate Weibull for dist-level queries.
# ==========================================================================


#' Series of Weibull components with common shape (closed form as a Weibull)
#'
#' Constructs a `dist_structure` for a series of Weibull components sharing
#' a common shape parameter. By the standard identity, the system lifetime
#' is itself Weibull with the common shape and an aggregate scale
#' `(sum(1 / scale^shape))^(-1 / shape)`. Methods for `surv`, `cdf`,
#' `sampler`, and `mean` forward to this aggregate Weibull, giving exact
#' closed-form values.
#'
#' @param shape Positive scalar: common Weibull shape.
#' @param scales Positive numeric vector: per-component Weibull scales.
#' @return
#' `wei_homogeneous_series()` returns an object of class
#'   `c("wei_homogeneous_series", "wei_series", "series_dist",
#'   "coherent_dist", "dist_structure", "univariate_dist",
#'   "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' - `mean()`: a numeric scalar (the mean system lifetime,
#'   `aggregate_scale * gamma(1 + 1 / shape)`).
#' @examples
#' sys <- wei_homogeneous_series(shape = 2, scales = c(1, 2, 3))
#' # System lifetime is Weibull(shape = 2, scale = aggregate_scale)
#' algebraic.dist::surv(sys)(1)
#' @export
wei_homogeneous_series <- function(shape, scales) {
  stopifnot(length(shape) == 1L, shape > 0, all(scales > 0))
  shapes <- rep(shape, length(scales))
  obj <- wei_series(shapes = shapes, scales = scales)
  obj$shape <- as.numeric(shape)
  obj$aggregate_scale <- (sum(1 / scales^shape))^(-1 / shape)
  class(obj) <- c("wei_homogeneous_series", class(obj))
  obj
}


#' @rdname wei_homogeneous_series
#' @param x A `wei_homogeneous_series` object.
#' @param ... Ignored.
#' @export
surv.wei_homogeneous_series <- function(x, ...) {
  k <- x$shape
  s_agg <- x$aggregate_scale
  function(t, ...) exp(-(t / s_agg)^k)
}


#' @rdname wei_homogeneous_series
#' @export
sampler.wei_homogeneous_series <- function(x, ...) {
  k <- x$shape
  s_agg <- x$aggregate_scale
  function(n, ...) stats::rweibull(n, shape = k, scale = s_agg)
}


#' @rdname wei_homogeneous_series
#' @export
mean.wei_homogeneous_series <- function(x, ...) {
  x$aggregate_scale * gamma(1 + 1 / x$shape)
}

# ==========================================================================
# Topology defaults on `dist_structure`
# ==========================================================================
#
# Each default is a direct composition of other generics. Implementors
# override when a closed-form or specialized algorithm is faster.
# Internal helpers used here live in R/internal-utils.R.
# ==========================================================================


#' @rdname phi
#' @export
phi.dist_structure <- function(x, state) {
  for (P in min_paths(x)) {
    if (all(state[P] == 1L)) return(1L)
  }
  0L
}


# Default min_paths derived by enumeration from the structure function.
# Implementors who provide only phi.X (no min_paths.X) get this default;
# implementors who provide min_paths.X override it. The enumeration is
# O(2^m) and therefore most appropriate for small to moderate m or as
# a fallback during prototyping; specialized subclasses with closed-form
# minimal paths should provide their own min_paths method.
#' @rdname min_paths
#' @export
min_paths.dist_structure <- function(x) {
  m <- ncomponents(x)
  grid <- binary_grid(m)
  good <- vapply(seq_len(nrow(grid)),
                 function(i) phi(x, grid[i, ]), integer(1L))
  active <- which(good == 1L)
  if (length(active) == 0L) return(list())
  candidate_sets <- lapply(active, function(i) which(grid[i, ] == 1L))
  minimize_sets(candidate_sets)
}


#' @rdname min_cuts
#' @export
min_cuts.dist_structure <- function(x) {
  # Berge transversal: iteratively build minimal hitting sets of min_paths.
  paths <- min_paths(x)
  if (length(paths) == 0L) return(list())
  transversals <- list(integer(0))
  for (P in paths) {
    # Pre-size: every (transversal, path-element) pair generates one
    # candidate. Avoids the O(n^2) cost of list[[length(.) + 1L]] <- x.
    new_trans <- vector("list", length(transversals) * length(P))
    idx <- 1L
    for (tr in transversals) {
      for (p in P) {
        new_trans[[idx]] <- sort(unique(c(tr, p)))
        idx <- idx + 1L
      }
    }
    transversals <- minimize_sets(new_trans)
  }
  transversals
}


#' @rdname critical_states
#' @export
critical_states.dist_structure <- function(x, j) {
  m <- ncomponents(x)
  stopifnot(length(j) == 1L, j >= 1L, j <= m)
  other <- setdiff(seq_len(m), j)
  grid <- binary_grid(m - 1L)
  keep <- logical(nrow(grid))
  x1 <- x0 <- integer(m)
  for (i in seq_len(nrow(grid))) {
    s <- grid[i, ]
    x1[other] <- s; x1[j] <- 1L
    x0[other] <- s; x0[j] <- 0L
    keep[i] <- (phi(x, x1) == 1L) && (phi(x, x0) == 0L)
  }
  grid[keep, , drop = FALSE]
}


#' @rdname system_lifetime
#' @export
system_lifetime.dist_structure <- function(x, times) {
  m <- ncomponents(x)
  stopifnot(length(times) == m, all(times >= 0))
  ord <- order(times)
  sorted <- times[ord]
  state <- rep(1L, m)
  for (k in seq_len(m)) {
    state[ord[k]] <- 0L
    if (phi(x, state) == 0L) return(sorted[k])
  }
  sorted[m]
}


#' @rdname system_censoring
#' @export
system_censoring.dist_structure <- function(x, times) {
  t_sys <- system_lifetime(x, times)
  status <- rep("exact", length(times))
  status[times < t_sys] <- "left"
  status[times > t_sys] <- "right"
  list(system_time = t_sys, component_status = status)
}


#' @rdname is_coherent
#' @export
is_coherent.dist_structure <- function(x) {
  m <- ncomponents(x)
  grid <- binary_grid(m - 1L)
  rows <- seq_len(nrow(grid))
  for (j in seq_len(m)) {
    other <- setdiff(seq_len(m), j)
    any_differ <- FALSE
    x0 <- x1 <- integer(m)
    for (i in rows) {
      s <- grid[i, ]
      x0[other] <- s; x0[j] <- 0L
      x1[other] <- s; x1[j] <- 1L
      p0 <- phi(x, x0); p1 <- phi(x, x1)
      if (p1 < p0) return(FALSE)
      if (p1 != p0) any_differ <- TRUE
    }
    if (!any_differ) return(FALSE)
  }
  TRUE
}


#' @rdname structural_importance
#' @export
structural_importance.dist_structure <- function(x, j) {
  m <- ncomponents(x)
  crit <- critical_states(x, j)
  nrow(crit) / (2^(m - 1L))
}


#' @rdname system_signature
#' @export
system_signature.dist_structure <- function(x) {
  m <- ncomponents(x)
  # Signature is undefined for non-functioning systems (no min_paths
  # means phi is identically 0). Without this guard, the algorithm
  # below silently returns (1, 0, ..., 0) for any always-zero phi.
  if (length(min_paths(x)) == 0L || phi(x, rep(1L, m)) == 0L) {
    stop("system_signature is undefined for systems with no minimal ",
         "path sets (phi identically zero).", call. = FALSE)
  }
  if (m > 9L) {
    warning("system_signature via permutation enumeration is ",
            "expensive for m > 9 (m! grows quickly); override with a ",
            "specialized method when possible.")
  }
  sig_counts <- integer(m)
  for (perm in permutations(m)) {
    state <- rep(1L, m)
    for (i in seq_len(m)) {
      state[perm[[i]]] <- 0L
      if (phi(x, state) == 0L) {
        sig_counts[i] <- sig_counts[i] + 1L
        break
      }
    }
  }
  sig_counts / factorial(m)
}


#' @rdname reliability
#' @export
reliability.dist_structure <- function(x, p) {
  m <- ncomponents(x)
  if (length(p) == 1L) p <- rep(p, m)
  stopifnot(length(p) == m, all(p >= 0 & p <= 1))
  grid <- binary_grid(m)
  # Probability of each 2^m component state vector: state j contributes
  # p[j] if state[j] == 1 else (1 - p[j]). Vectorized across rows.
  prob_matrix <- grid * rep(p, each = nrow(grid)) +
    (1L - grid) * rep(1 - p, each = nrow(grid))
  state_probs <- apply(prob_matrix, 1L, prod)
  phi_values <- apply(grid, 1L, function(s) phi(x, s))
  sum(phi_values * state_probs)
}

# ==========================================================================
# Distribution defaults on `dist_structure`
# ==========================================================================
#
# System-level distribution methods composed from component distributions
# and topology. Implementors override for closed-form speed.
# ==========================================================================


#' System survival via reliability composition
#'
#' Returns a closure `function(t, ...)` where `surv(x)(t)` equals
#' `reliability(x, surv(component(x, j))(t) for each j)`. This is the
#' classical identity `S_sys(t) = R(S(t))` realized as a composition.
#'
#' @rdname dist_structure
#' @param x A [dist_structure] object.
#' @param ... Ignored.
#' @export
surv.dist_structure <- function(x, ...) {
  m <- ncomponents(x)
  comp_surv_fns <- lapply(seq_len(m), function(j) {
    algebraic.dist::surv(component(x, j))
  })
  function(t, ...) {
    vapply(t, function(ti) {
      ps <- vapply(comp_surv_fns, function(S_j) S_j(ti), numeric(1L))
      reliability(x, ps)
    }, numeric(1L))
  }
}


#' System CDF: `1 - surv(x)(t)`
#'
#' @rdname dist_structure
#' @export
cdf.dist_structure <- function(x, ...) {
  S <- algebraic.dist::surv(x)
  function(t, ...) 1 - S(t, ...)
}


#' System sampler via component samplers and system_lifetime
#'
#' Returns a closure `function(n, ...)` that draws `n` system lifetimes by
#' sampling each component independently and applying [system_lifetime()]
#' to combine.
#'
#' @rdname dist_structure
#' @export
sampler.dist_structure <- function(x, ...) {
  m <- ncomponents(x)
  comp_samplers <- lapply(seq_len(m), function(j) {
    algebraic.dist::sampler(component(x, j))
  })
  function(n, ...) {
    comp_matrix <- sample_component_matrix(comp_samplers, n)
    apply(comp_matrix, 1L, function(times) system_lifetime(x, times))
  }
}

# ==========================================================================
# Closed-form k-out-of-n of heterogeneous exponential components
# ==========================================================================
#
# For k-of-m of heterogeneous exponentials:
#   S_sys(t) = P(at least k of m exponentials are alive at t)
#            = sum_{A subset of [m], |A|>=k} prod_{j in A} exp(-rate_j t)
#              * prod_{j notin A} (1 - exp(-rate_j t))
# Sampler uses the (m - k + 1)-th order statistic of independently sampled
# component lifetimes.
# ==========================================================================


#' k-out-of-n system of independent exponential components (closed form)
#'
#' Constructs a `dist_structure` for a k-out-of-m system whose components
#' are independent exponentials. Closed-form methods are provided for
#' `surv`, `cdf`, `sampler`, `density`, and `hazard`. `mean` falls back
#' to numerical integration via the `dist_structure` default.
#'
#' @param k Minimum number of functioning components for system operation.
#' @param rates Positive numeric vector of length `m` with `m >= k`.
#' @return
#' `exp_kofn()` returns an object of class
#'   `c("exp_kofn", "kofn_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`, `density()`, `hazard()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' @examples
#' sys <- exp_kofn(k = 2, rates = c(1, 2, 3))
#' algebraic.dist::surv(sys)(1)
#' @export
exp_kofn <- function(k, rates) {
  stopifnot(is.numeric(rates), length(rates) >= 1L, all(rates > 0))
  m <- length(rates)
  stopifnot(k >= 1L, k <= m)
  components <- lapply(rates, algebraic.dist::exponential)
  obj <- kofn_dist(k, components)
  obj$rates <- as.numeric(rates)
  class(obj) <- c("exp_kofn", class(obj))
  obj
}


#' @rdname exp_kofn
#' @param x An `exp_kofn` object.
#' @param ... Ignored.
#' @export
surv.exp_kofn <- function(x, ...) {
  rates <- x$rates
  k <- x$k
  function(t, ...) {
    vapply(t, function(ti) {
      kofn_surv_probability(exp(-rates * ti), k)
    }, numeric(1L))
  }
}


#' @rdname exp_kofn
#' @export
sampler.exp_kofn <- function(x, ...) {
  order_idx <- length(x$rates) - x$k + 1L
  samplers <- make_component_samplers(stats::rexp, rate = x$rates)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L,
          function(row) sort(row)[order_idx])
  }
}


#' @rdname exp_kofn
#' @method hazard exp_kofn
#' @importFrom algebraic.dist hazard
#' @export
hazard.exp_kofn <- function(x, ...) {
  # h_sys(t) = f_sys(t) / S_sys(t); both factors have closed forms here.
  f_fn <- density.exp_kofn(x)
  S_fn <- surv.exp_kofn(x)
  function(t, ...) f_fn(t) / S_fn(t)
}


#' @rdname exp_kofn
#' @importFrom stats density dexp
#' @export
density.exp_kofn <- function(x, ...) {
  rates <- x$rates
  k <- x$k
  function(t, log = FALSE, ...) {
    vals <- vapply(t, function(ti) {
      surv <- exp(-rates * ti)
      dens <- rates * surv
      kofn_density_value(dens, surv, k)
    }, numeric(1L))
    if (isTRUE(log)) log(vals) else vals
  }
}

# ==========================================================================
# Internal helpers (not exported)
# ==========================================================================
#
# Shared utilities used by the topology defaults in defaults.R and the
# closed-form specializations in exp_*.R, wei_*.R, gamma_series.R, and
# lognormal_series.R. None of these are part of the public API.
# ==========================================================================


# Remove non-minimal sets from a list of integer vectors. Duplicates are
# collapsed first (the strict-subset minimality check would not remove
# equal-length duplicates on its own); this is important in Berge
# transversal iteration, where independent path extensions routinely
# produce repeated transversals.
minimize_sets <- function(sets) {
  sets <- unique(sets)
  n <- length(sets)
  if (n == 0L) return(sets)
  is_min <- rep(TRUE, n)
  for (i in seq_len(n)) {
    if (!is_min[i]) next
    for (j in seq_len(n)) {
      if (i == j || !is_min[j]) next
      if (all(sets[[i]] %in% sets[[j]]) &&
          length(sets[[i]]) < length(sets[[j]])) {
        is_min[j] <- FALSE
      }
    }
  }
  sets[is_min]
}


# Aggregate survival for a k-of-m system given the per-component survival
# probabilities at a single time point. Computes P(at least k of m
# components are alive), summing over all subsets A with |A| >= k of
# (prod S_j for j in A) * (prod (1 - S_j) for j not in A). Shared by
# exp_kofn and wei_kofn surv methods.
kofn_surv_probability <- function(comp_surv, k) {
  m <- length(comp_surv)
  comp_fail <- 1 - comp_surv
  all_idx <- seq_len(m)
  total <- 0
  for (sz in seq.int(k, m)) {
    for (A in utils::combn(m, sz, simplify = FALSE)) {
      alive <- prod(comp_surv[A])
      failed <- setdiff(all_idx, A)
      dead <- if (length(failed) == 0L) 1 else prod(comp_fail[failed])
      total <- total + alive * dead
    }
  }
  total
}


# Aggregate density for a k-of-m system at a single time point, given
# per-component densities f_j(t), survivals S_j(t), and CDFs F_j(t) =
# 1 - S_j(t). Uses the critical-state formula:
#   f_sys(t) = sum_{j=1}^m f_j(t) * P(component j is critical at t)
# where j is critical iff exactly (k - 1) of the other components are
# alive (so j's failure at t drops the alive count from k to k - 1,
# triggering system failure). Shared by exp_kofn and wei_kofn density
# methods.
kofn_density_value <- function(comp_dens, comp_surv, k) {
  m <- length(comp_surv)
  comp_fail <- 1 - comp_surv
  total <- 0
  for (j in seq_len(m)) {
    others <- setdiff(seq_len(m), j)
    p_crit <- 0
    if (k == 1L) {
      # Need 0 others alive.
      p_crit <- prod(comp_fail[others])
    } else if (k - 1L > length(others)) {
      p_crit <- 0
    } else {
      subsets <- utils::combn(others, k - 1L, simplify = FALSE)
      for (B in subsets) {
        alive_term <- if (length(B) == 0L) 1 else prod(comp_surv[B])
        dead_idx <- setdiff(others, B)
        dead_term <- if (length(dead_idx) == 0L) 1 else prod(comp_fail[dead_idx])
        p_crit <- p_crit + alive_term * dead_term
      }
    }
    total <- total + comp_dens[j] * p_crit
  }
  total
}


# Collect n samples from each of m component samplers into an n-by-m
# matrix. Handles the n == 1 / m == 1 edge cases where vapply would
# otherwise drop dimensions.
sample_component_matrix <- function(samplers, n) {
  m <- length(samplers)
  mat <- vapply(samplers, function(samp) samp(n), numeric(n))
  if (!is.matrix(mat)) dim(mat) <- c(n, m)
  mat
}


# Build a list of per-component sampler closures from a base stats
# rXxx function and parallel parameter vectors. The resulting samplers
# each take a single arg `n` and draw n independent values for that
# component. Used by the closed-form series/parallel/kofn constructors.
make_component_samplers <- function(rfun, ...) {
  args <- list(...)
  m <- length(args[[1L]])
  lapply(seq_len(m), function(j) {
    per_args <- lapply(args, `[[`, j)
    function(n) do.call(rfun, c(list(n = n), per_args))
  })
}


# Product-of-per-component-survivals closure for a series system where
# each component's survival at time ti has the form
# `pXxx(ti, ..., lower.tail = FALSE)`. `params` is a named list of
# parameter vectors (each of length m) to pass to `pfun`. Returns a
# function(t) that vectorises over `t`.
series_surv_product <- function(pfun, params) {
  function(t, ...) {
    vapply(t, function(ti) {
      args <- c(list(q = ti), params, list(lower.tail = FALSE))
      prod(do.call(pfun, args))
    }, numeric(1L))
  }
}


# 2^n binary grid as an integer matrix (no col names, no expand.grid
# attributes). Rows enumerate states in {0, 1}^n.
binary_grid <- function(n) {
  if (n == 0L) return(matrix(integer(0), nrow = 1L, ncol = 0L))
  grid <- expand.grid(rep(list(c(0L, 1L)), n), KEEP.OUT.ATTRS = FALSE)
  grid <- as.matrix(grid)
  colnames(grid) <- NULL
  grid
}


# All permutations of 1:n as a list of integer vectors. Base R has no
# such helper; this recursive implementation is adequate for the small m
# values at which enumeration-based signature computation is feasible.
permutations <- function(n) {
  if (n <= 1L) return(list(seq_len(n)))
  smaller <- permutations(n - 1L)
  result <- vector("list", n * length(smaller))
  idx <- 1L
  for (i in seq_len(n)) {
    rest <- seq_len(n)[-i]
    for (p in smaller) {
      result[[idx]] <- c(i, rest[p])
      idx <- idx + 1L
    }
  }
  result
}

# ==========================================================================
# Dual structures
# ==========================================================================


#' Default dual: lazy wrapper
#'
#' Returns a `dual_of_system` object carrying the original structure.
#' `phi_dual(state) = 1 - phi(original, 1 - state)` is evaluated on demand.
#' All other generics fall through to `dist_structure` defaults.
#'
#' @rdname dual
#' @export
dual.dist_structure <- function(x) {
  structure(
    list(original = x, m = ncomponents(x)),
    class = c("dual_of_system", "dist_structure",
              "univariate_dist", "continuous_dist", "dist")
  )
}


#' @export
ncomponents.dual_of_system <- function(x) x$m


#' @export
component.dual_of_system <- function(x, j, ...) {
  component(x$original, j, ...)
}


#' @export
phi.dual_of_system <- function(x, state) {
  state <- as.integer(state)
  1L - phi(x$original, 1L - state)
}


#' Dual of dual is the original (involution)
#'
#' @rdname dual
#' @export
dual.dual_of_system <- function(x) x$original

# ==========================================================================
# Virtual base class for structured distributions
# ==========================================================================


#' @title Virtual base class for structured distributions
#'
#' @description
#' `dist_structure` is the virtual S3 class for distributions whose random
#' variable has internal component structure: coherent reliability
#' systems decomposed into components arranged by a structure function.
#' Every concrete implementation (`coherent_dist`, `series_dist`,
#' `parallel_dist`, `kofn_dist`, `bridge_dist`, or user-defined
#' subclasses) should include `"dist_structure"`, the `algebraic.dist`
#' ancestor `"univariate_dist"`, and `"dist"` in its class vector.
#'
#' @details
#' Concrete implementations provide S3 methods for the generics in this
#' package. The minimum required methods are [ncomponents()], [component()],
#' and one of [phi()] or [min_paths()]; every other generic has a default
#' method on `dist_structure` that composes the primitives.
#'
#' If both `phi.<class>()` and `min_paths.<class>()` are provided, the
#' implementor is responsible for keeping them consistent: `phi` derives
#' the in-package generics `reliability`, `critical_states`, and
#' `is_coherent`; `min_paths` derives `min_cuts` and `system_signature`.
#' Inconsistent implementations produce silently inconsistent results.
#'
#' @return
#' This help topic documents the virtual base class together with the
#' three distribution-algebra default methods that compose component
#' distributions through the topology:
#'
#' - [surv.dist_structure()] returns a closure `function(t, ...)` that
#'   evaluates the system survival function via the reliability identity
#'   `S_sys(t) = R(S_1(t), ..., S_m(t))`.
#' - [cdf.dist_structure()] returns a closure `function(t, ...)` equal
#'   to `1 - surv(x)(t)`.
#' - [sampler.dist_structure()] returns a closure `function(n, ...)`
#'   that draws `n` independent system lifetimes by sampling each
#'   component and combining via [system_lifetime()].
#'
#' Concrete subclasses override any of these for closed-form speed; see
#' the closed-form specializations under the See Also entries.
#'
#' @seealso [validate_dist_structure()] for an implementor-side
#'   construction-time validator. [phi()] and [min_paths()] for the
#'   bidirectional protocol primitives. The closed-form families
#'   ([exp_series()], [wei_kofn()], etc.) for reference implementations.
#'
#' @name dist_structure
NULL


#' Predicate for dist_structure objects
#'
#' @param x Any object.
#' @return `TRUE` if `x` inherits from `"dist_structure"`.
#' @export
is_dist_structure <- function(x) inherits(x, "dist_structure")


#' Validate that an object satisfies the dist_structure protocol
#'
#' Checks that `x` declares `"dist_structure"` in its class chain and
#' provides the three required generics: [ncomponents()], [component()],
#' and at least one of [phi()] or [min_paths()].
#'
#' Useful in subclass constructors to fail fast with a clear error
#' message rather than at first method dispatch (where the user sees
#' opaque "no applicable method" errors). A typical pattern:
#'
#' ```r
#' my_subclass <- function(...) {
#'   obj <- structure(list(...), class = c("my_subclass", "dist_structure",
#'                                          "univariate_dist", "continuous_dist", "dist"))
#'   validate_dist_structure(obj)
#'   obj
#' }
#' ```
#'
#' @param x An object claiming to satisfy the `dist_structure` protocol.
#' @return `TRUE` (invisibly) if all checks pass. Stops with an
#'   informative error otherwise.
#' @export
#' @examples
#' # Success: a valid dist_structure (any built-in topology shortcut works).
#' validate_dist_structure(series_dist(replicate(3,
#'   algebraic.dist::exponential(1), simplify = FALSE)))
#'
#' # Failure: an object that declares dist_structure but provides no
#' # primitives. Wrapped in tryCatch for the example's success status.
#' bad <- structure(list(),
#'                  class = c("not_a_real_class", "dist_structure",
#'                            "univariate_dist", "continuous_dist", "dist"))
#' tryCatch(validate_dist_structure(bad),
#'          error = function(e) conditionMessage(e))
validate_dist_structure <- function(x) {
  if (!inherits(x, "dist_structure")) {
    stop("`x` must declare 'dist_structure' in its class chain; ",
         "got class = c(", paste(shQuote(class(x)), collapse = ", "), ").",
         call. = FALSE)
  }
  m <- tryCatch(ncomponents(x),
                error = function(e) {
                  stop("`x` does not implement `ncomponents()`: ", e$message,
                       call. = FALSE)
                })
  if (!is.numeric(m) || length(m) != 1L || m < 1L) {
    stop("`ncomponents(x)` must return a single positive integer; got ",
         paste(format(m), collapse = ", "), ".", call. = FALSE)
  }
  tryCatch(component(x, 1L),
           error = function(e) {
             stop("`x` does not implement `component()`: ", e$message,
                  call. = FALSE)
           })
  # Check via S3 method-table lookup (rather than calling phi/min_paths
  # and catching errors) because the default phi.dist_structure and
  # min_paths.dist_structure are mutually recursive: if a subclass
  # provides neither, calling the dispatched generic would loop. Looking
  # for a non-default registered method on any subclass-specific class is
  # the correct check.
  #
  # Note: `coherent_dist` is intentionally NOT stripped from user_classes.
  # `min_paths.coherent_dist` is registered (returns x$min_paths from the
  # constructor), so any subclass of `coherent_dist` legitimately
  # inherits a working min_paths implementation; validate should accept
  # it. Stripping `coherent_dist` would incorrectly reject those
  # subclasses.
  user_classes <- setdiff(
    class(x),
    c("dist_structure", "univariate_dist", "continuous_dist", "dist")
  )
  has_method <- function(generic) {
    for (cls in user_classes) {
      m <- utils::getS3method(generic, cls, optional = TRUE)
      if (!is.null(m)) return(TRUE)
    }
    FALSE
  }
  if (!has_method("phi") && !has_method("min_paths")) {
    stop("`x` must implement at least one of `phi.<class>()` or ",
         "`min_paths.<class>()`; neither was found on any subclass ",
         "(only the mutually-recursive `dist_structure` defaults).",
         call. = FALSE)
  }
  invisible(TRUE)
}


#' Format a dist_structure object
#'
#' @param x A [dist_structure] object.
#' @param ... Ignored.
#' @return Character vector suitable for [cat()].
#' @export
format.dist_structure <- function(x, ...) {
  cls <- setdiff(class(x),
                 c("dist_structure", "univariate_dist",
                   "continuous_dist", "dist"))
  head <- if (length(cls)) {
    paste0("<dist_structure: ", cls[[1L]], ">")
  } else {
    "<dist_structure>"
  }
  m <- tryCatch(ncomponents(x), error = function(e) NA_integer_)
  body <- if (is.na(m)) {
    "  components: (ncomponents method not defined)"
  } else {
    paste0("  components: ", m)
  }
  c(head, body)
}


#' Print a dist_structure object
#'
#' @param x A [dist_structure] object.
#' @param ... Passed to [format()].
#' @return `x`, invisibly.
#' @export
print.dist_structure <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

# ==========================================================================
# IID convenience constructors
# ==========================================================================
#
# Shortcuts for building dist_structure objects when all components share
# the same distribution.
# ==========================================================================


#' Series system of m iid components
#'
#' Equivalent to the minimum of `m` iid random variables from `d`. Unlike
#' `min` in the base distribution algebra, `min_iid` preserves the series
#' topology so structural queries (phi, min_paths, structural_importance)
#' remain available.
#'
#' @param d A `dist` object (the common component distribution).
#' @param m Number of components (positive integer).
#' @return A `series_dist`.
#' @examples
#' sys <- min_iid(algebraic.dist::exponential(1), m = 3)
#' # System survival at t=0.5 equals (exp(-t))^3 = exp(-1.5)
#' @export
min_iid <- function(d, m) {
  stopifnot(m >= 1L)
  series_dist(replicate(m, d, simplify = FALSE))
}


#' Parallel system of m iid components
#'
#' Equivalent to the maximum of `m` iid random variables from `d`.
#' Preserves the parallel topology for structural queries.
#'
#' @param d A `dist` object.
#' @param m Number of components.
#' @return A `parallel_dist`.
#' @export
max_iid <- function(d, m) {
  stopifnot(m >= 1L)
  parallel_dist(replicate(m, d, simplify = FALSE))
}


#' k-th order statistic of m iid components
#'
#' Constructs a `kofn_dist` whose system lifetime equals `T_(k)`, the
#' k-th order statistic of `m` iid draws from `d`. Under the k-of-m
#' parametrization, the system fails at the `(m - k + 1)`-th component
#' failure; setting the threshold to `m - k + 1` makes the system lifetime
#' equal `T_(k)`.
#'
#' The internal call is `kofn_dist(m - k + 1, ...)`: this maps the
#' order-statistic index `k` (where `k = 1` is the minimum, `k = m` is
#' the maximum) to the `:G` convention used by `kofn_dist` (where the
#' `k_dist` argument is the number of functioning components required).
#'
#' @param d A `dist` object.
#' @param k The order statistic index (1 = min, m = max).
#' @param m Number of iid components.
#' @return A `kofn_dist`.
#' @examples
#' # Median of 5 iid exponentials
#' sys <- order_statistic(algebraic.dist::exponential(1), k = 3, m = 5)
#' @export
order_statistic <- function(d, k, m) {
  stopifnot(k >= 1L, k <= m)
  kofn_dist(m - k + 1L, replicate(m, d, simplify = FALSE))
}

# ==========================================================================
# IID convenience constructors
# ==========================================================================


test_that("min_iid equals series_dist of m copies", {
  d <- algebraic.dist::exponential(2)
  sys <- min_iid(d, m = 4L)
  expect_s3_class(sys, "series_dist")
  expect_equal(ncomponents(sys), 4L)
  # All components are the same exponential
  for (j in 1:4) {
    expect_s3_class(component(sys, j), "exponential")
  }
})


test_that("max_iid equals parallel_dist of m copies", {
  d <- algebraic.dist::exponential(1)
  sys <- max_iid(d, m = 3L)
  expect_s3_class(sys, "parallel_dist")
  expect_equal(ncomponents(sys), 3L)
})


test_that("order_statistic constructs correct kofn", {
  d <- algebraic.dist::exponential(1)
  # Median of 5: T_(3). System fails at 3rd order statistic.
  sys <- order_statistic(d, k = 3L, m = 5L)
  expect_s3_class(sys, "kofn_dist")
  expect_equal(ncomponents(sys), 5L)
  # System lifetime = T_(k) with k = 3 means 3rd smallest time
  times <- c(0.5, 0.1, 0.3, 0.7, 0.9)
  # Sorted: 0.1, 0.3, 0.5, 0.7, 0.9. T_(3) = 0.5
  expect_equal(system_lifetime(sys, times), 0.5)
})


test_that("min_iid survival matches m-th power of component survival", {
  # min of m iid Exp(1) is Exp(m)
  d <- algebraic.dist::exponential(1)
  sys <- min_iid(d, m = 3L)
  S_sys <- algebraic.dist::surv(sys)
  for (ti in c(0.2, 0.5, 1.0)) {
    expect_equal(S_sys(ti), exp(-3 * ti), tolerance = 1e-10)
  }
})

# ==========================================================================
# Closed-form k-out-of-n of heterogeneous Weibull components
# ==========================================================================
#
# Same subset-enumeration formula as exp_kofn, with Weibull per-component
# survivals S_j(t) = exp(-(t / scale_j)^shape_j).
# ==========================================================================


#' k-out-of-n system of independent Weibull components (closed form)
#'
#' Constructs a `dist_structure` for a k-out-of-m system whose components
#' are independent (possibly heterogeneous) Weibulls. Closed-form `surv`,
#' `cdf`, `sampler`, `density`, and `hazard` via subset enumeration,
#' the critical-state density formula, and component order statistics.
#'
#' @param k Minimum functioning components for system operation.
#' @param shapes Positive numeric vector of length `m`.
#' @param scales Positive numeric vector of length `m`.
#' @return
#' `wei_kofn()` returns an object of class
#'   `c("wei_kofn", "kofn_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`, `density()`, `hazard()`: a closure `function(t, ...)`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' @examples
#' sys <- wei_kofn(k = 2, shapes = c(1, 2, 3), scales = c(1, 2, 3))
#' algebraic.dist::surv(sys)(1)
#' @export
wei_kofn <- function(k, shapes, scales) {
  stopifnot(length(shapes) == length(scales),
            all(shapes > 0), all(scales > 0))
  m <- length(shapes)
  stopifnot(k >= 1L, k <= m)
  components <- lapply(seq_len(m), function(j) {
    algebraic.dist::weibull_dist(shape = shapes[j], scale = scales[j])
  })
  obj <- kofn_dist(k, components)
  obj$shapes <- as.numeric(shapes)
  obj$scales <- as.numeric(scales)
  class(obj) <- c("wei_kofn", class(obj))
  obj
}


#' @rdname wei_kofn
#' @param x A `wei_kofn` object.
#' @param ... Ignored.
#' @export
surv.wei_kofn <- function(x, ...) {
  shapes <- x$shapes
  scales <- x$scales
  k <- x$k
  function(t, ...) {
    vapply(t, function(ti) {
      kofn_surv_probability(exp(-(ti / scales)^shapes), k)
    }, numeric(1L))
  }
}


#' @rdname wei_kofn
#' @export
sampler.wei_kofn <- function(x, ...) {
  order_idx <- length(x$shapes) - x$k + 1L
  samplers <- make_component_samplers(stats::rweibull,
                                      shape = x$shapes, scale = x$scales)
  function(n, ...) {
    apply(sample_component_matrix(samplers, n), 1L,
          function(row) sort(row)[order_idx])
  }
}


#' @rdname wei_kofn
#' @method hazard wei_kofn
#' @importFrom algebraic.dist hazard
#' @export
hazard.wei_kofn <- function(x, ...) {
  # h_sys(t) = f_sys(t) / S_sys(t); both factors have closed forms here.
  f_fn <- density.wei_kofn(x)
  S_fn <- surv.wei_kofn(x)
  function(t, ...) f_fn(t) / S_fn(t)
}


#' @rdname wei_kofn
#' @importFrom stats density dweibull
#' @export
density.wei_kofn <- function(x, ...) {
  shapes <- x$shapes
  scales <- x$scales
  k <- x$k
  m <- length(shapes)
  function(t, log = FALSE, ...) {
    vals <- vapply(t, function(ti) {
      surv <- exp(-(ti / scales)^shapes)
      dens <- vapply(seq_len(m), function(j) {
        stats::dweibull(ti, shape = shapes[j], scale = scales[j])
      }, numeric(1L))
      kofn_density_value(dens, surv, k)
    }, numeric(1L))
    if (isTRUE(log)) log(vals) else vals
  }
}

# ==========================================================================
# Closed-form series of exponential components
# ==========================================================================
#
# A series of m independent exponentials with rates lambda_j has itself
# an exponential system lifetime with rate = sum(lambda). This gives
# closed-form survival, sampler, hazard, density, and mean.
# ==========================================================================


#' Series of exponential components (closed form)
#'
#' Constructs a `dist_structure` representing a series system whose
#' components are independent exponentials. The system lifetime is itself
#' an `Exp(sum(rates))` distribution; all dist-level queries have
#' closed-form expressions that bypass the general default methods.
#'
#' @param rates Positive numeric vector of length `m`: per-component
#'   exponential rates.
#' @return
#' `exp_series()` returns an object of class
#'   `c("exp_series", "series_dist", "coherent_dist", "dist_structure",
#'   "univariate_dist", "continuous_dist", "dist")`.
#'
#' The associated S3 methods return:
#' - `surv()`, `density()`, `hazard()`: a closure `function(t, ...)`
#'   evaluating the named quantity at `t`.
#' - `cdf()` is derived via the `dist_structure` default and returns
#'   a closure `function(t, ...)` equal to `1 - surv(x)(t)`.
#' - `sampler()`: a closure `function(n, ...)` returning `n` random
#'   variates from the system lifetime distribution.
#' - `mean()`: a numeric scalar (the mean system lifetime).
#' @examples
#' sys <- exp_series(c(0.5, 0.3, 0.2))
#' algebraic.dist::surv(sys)(1)  # equals exp(-sum(rates) * 1)
#' mean(sys)                     # equals 1 / sum(rates)
#' @export
exp_series <- function(rates) {
  stopifnot(is.numeric(rates), length(rates) >= 1L, all(rates > 0))
  components <- lapply(rates, algebraic.dist::exponential)
  obj <- series_dist(components)
  obj$rates <- as.numeric(rates)
  obj$total_rate <- sum(rates)
  class(obj) <- c("exp_series", class(obj))
  obj
}


#' @rdname exp_series
#' @param x An `exp_series` object.
#' @param ... Ignored.
#' @export
surv.exp_series <- function(x, ...) {
  lam <- x$total_rate
  function(t, ...) exp(-lam * t)
}


#' @rdname exp_series
#' @export
sampler.exp_series <- function(x, ...) {
  lam <- x$total_rate
  function(n, ...) stats::rexp(n, rate = lam)
}


#' @rdname exp_series
#' @export
mean.exp_series <- function(x, ...) 1 / x$total_rate


#' @rdname exp_series
#' @importFrom stats density dexp
#' @export
density.exp_series <- function(x, ...) {
  lam <- x$total_rate
  function(t, log = FALSE, ...) stats::dexp(t, rate = lam, log = log)
}


#' @rdname exp_series
#' @method hazard exp_series
#' @importFrom algebraic.dist hazard
#' @export
hazard.exp_series <- function(x, ...) {
  lam <- x$total_rate
  function(t, log.p = FALSE, ...) {
    rep(if (isTRUE(log.p)) log(lam) else lam, length(t))
  }
}

# ==========================================================================
# S3 generic functions
# ==========================================================================


#' Number of components
#'
#' @param x A [dist_structure] object.
#' @return A positive integer `m`, the number of components.
#' @export
ncomponents <- function(x) UseMethod("ncomponents")


#' The j-th component as a dist
#'
#' Returns the j-th component as an `algebraic.dist::dist` object with
#' parameters baked in (so `sampler`, `surv`, etc. work directly on it).
#'
#' @param x A [dist_structure] object.
#' @param j Integer component index in `1:ncomponents(x)`.
#' @param ... Implementation-specific arguments (e.g., parameter overrides
#'   for lazily parameterized systems).
#' @return An object inheriting from `dist`.
#' @export
component <- function(x, j, ...) UseMethod("component")


#' Structure function
#'
#' Evaluate the coherent structure function `phi: {0, 1}^m -> {0, 1}` at a
#' component state vector `state`. By convention `state[j] = 1` means
#' component `j` is functioning; `phi(state) = 1` means the system is
#' functioning.
#'
#' `phi` and [min_paths()] are dual primitives: if a subclass provides
#' only `phi.<class>()`, the default [min_paths.dist_structure()]
#' enumerates minimal subsets via `phi`; if a subclass provides only
#' `min_paths.<class>()`, the default [phi.dist_structure()] checks
#' whether `state` covers any minimal path. Subclasses providing both
#' must keep them consistent.
#'
#' @param x A [dist_structure] object.
#' @param state Integer or logical vector of length `ncomponents(x)` in
#'   `{0, 1}`.
#' @return Integer scalar, `0` or `1`.
#' @seealso [min_paths()] for the dual primitive; [validate_dist_structure()]
#'   for construction-time checking.
#' @export
phi <- function(x, state) UseMethod("phi")


#' Minimal path sets
#'
#' Returns the minimal subsets of `1:m` whose joint functioning is
#' sufficient for the system to function. `min_paths` and [phi()] are
#' dual primitives in the `dist_structure` protocol: providing one is
#' enough, the other has an enumerative default.
#'
#' @param x A [dist_structure] object.
#' @return A list of integer vectors.
#' @seealso [phi()] for the dual primitive; [min_cuts()] for the dual
#'   topology query (minimal subsets whose joint failure causes system
#'   failure).
#' @export
min_paths <- function(x) UseMethod("min_paths")


#' Minimal cut sets
#'
#' @param x A [dist_structure] object.
#' @return A list of integer vectors.
#' @export
min_cuts <- function(x) UseMethod("min_cuts")


#' System signature
#'
#' Samaniego's signature: `s = (s_1, ..., s_m)` where `s_k` is the
#' probability the system fails at the k-th component failure under
#' i.i.d. absolutely-continuous component lifetimes. Only depends on the
#' structure (phi), not on the component distribution. A default method
#' on `dist_structure` enumerates the `m!` orderings; this is feasible
#' for `m` up to about 8 or 9. Specialized subclasses should override
#' with closed-form expressions.
#'
#' @param x A [dist_structure] object.
#' @return Numeric vector of length `ncomponents(x)` summing to 1.
#' @export
system_signature <- function(x) UseMethod("system_signature")


#' Critical states of a component
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @return Integer matrix with `m - 1` columns; rows are the states of the
#'   other components for which `j` is critical.
#' @export
critical_states <- function(x, j) UseMethod("critical_states")


#' System lifetime from component times
#'
#' @param x A [dist_structure] object.
#' @param times Non-negative numeric vector of length `ncomponents(x)`.
#' @return Scalar system lifetime.
#' @export
system_lifetime <- function(x, times) UseMethod("system_lifetime")


#' Per-component censoring from system observation
#'
#' @param x A [dist_structure] object.
#' @param times Non-negative numeric vector of length `ncomponents(x)`.
#' @return A list with `system_time` (scalar) and `component_status`
#'   (character vector of length `m`, values in `"exact"`, `"left"`,
#'   `"right"`).
#' @export
system_censoring <- function(x, times) UseMethod("system_censoring")


#' Dual structure
#'
#' The dual structure satisfies `phi_dual(state) = 1 - phi(1 - state)`. The
#' dual of a series system is parallel; the dual of k-out-of-n is
#' (n - k + 1)-out-of-n.
#'
#' @param x A [dist_structure] object.
#' @return A [dist_structure] object representing the dual.
#' @export
dual <- function(x) UseMethod("dual")


#' Coherence axiom check
#'
#' @param x A [dist_structure] object.
#' @return `TRUE` if monotone and every component relevant.
#' @export
is_coherent <- function(x) UseMethod("is_coherent")


#' Birnbaum structural importance
#'
#' Fraction of the `2^(m - 1)` states of the other components for which
#' component `j` is critical.
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @return Numeric scalar in `[0, 1]`.
#' @export
structural_importance <- function(x, j) UseMethod("structural_importance")


#' System reliability polynomial
#'
#' `R(p) = E[phi(X)]` where `X_j ~ Bernoulli(p_j)` independent. The
#' multilinear extension of `phi` to component reliabilities.
#'
#' @param x A [dist_structure] object.
#' @param p Numeric vector of length `ncomponents(x)` or a scalar recycled
#'   to all components; values in `[0, 1]`.
#' @return Numeric scalar in `[0, 1]`.
#' @export
reliability <- function(x, p) UseMethod("reliability")


#' Birnbaum reliability importance
#'
#' The Birnbaum importance of component `j` at component reliabilities
#' `p` is `dR/dp_j = R(p with p_j = 1) - R(p with p_j = 0)`. Measures
#' how much the system reliability changes if `j` moves between certain
#' failure and certain success (and, by monotonicity, the rate of
#' change at any intermediate `p_j`).
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @param p Numeric vector of length `ncomponents(x)` or a scalar in
#'   `[0, 1]`.
#' @return Numeric scalar in `[0, 1]`.
#' @export
birnbaum_importance <- function(x, j, p) UseMethod("birnbaum_importance")


#' Criticality (Fussell) importance at time t
#'
#' Probability that component `j` has failed AND is critical given
#' that the system has failed by time `t`. Equals
#' `I_B(j; S(t)) * F_j(t) / F_sys(t)`.
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @param t Scalar time.
#' @return Numeric scalar in `[0, 1]`.
#' @export
criticality_importance <- function(x, j, t) UseMethod("criticality_importance")


#' Vesely-Fussell importance at time t
#'
#' Probability that at least one minimal cut set containing `j` has all
#' its components failed, given the system has failed by time `t`.
#' Computed exactly via inclusion-exclusion over subsets of cuts that
#' contain `j`.
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @param t Scalar time.
#' @return Numeric scalar in `[0, 1]`.
#' @export
vesely_fussell_importance <- function(x, j, t) {
  UseMethod("vesely_fussell_importance")
}


#' Substitute a component
#'
#' Return a new [dist_structure] with the j-th component replaced by
#' `new_component`. Topology is preserved; the returned object is a
#' `coherent_dist` with the same min_paths and the modified component
#' list.
#'
#' @param x A [dist_structure] object.
#' @param j Component index.
#' @param new_component A dist-compatible object to install at position j.
#' @return A new [dist_structure] object.
#' @export
substitute_component <- function(x, j, new_component) {
  UseMethod("substitute_component")
}


#' Compose systems hierarchically
#'
#' Produce a new [dist_structure] by replacing each component of `outer`
#' with a sub-system (either a [dist_structure] or a plain dist). The
#' composed system's components are the flattened inner components; its
#' min_paths enumerate the Cartesian products of inner min_paths within
#' each outer min_path.
#'
#' @param outer A [dist_structure] object.
#' @param inner_list A list of length `ncomponents(outer)`. Each element
#'   is either a [dist_structure] or a dist (single-component sub-system).
#' @return A `coherent_dist` representing the composed system.
#' @export
compose_systems <- function(outer, inner_list) {
  UseMethod("compose_systems")
}


#' Coerce to dist_structure
#'
#' Wraps a plain `algebraic.dist` distribution in a 1-component [series_dist]
#' so that it satisfies the dist_structure protocol. Useful for
#' polymorphic code that operates on dist_structure inputs but is also
#' handed unwrapped distributions. A `dist_structure` is returned
#' unchanged.
#'
#' @param x A dist or dist_structure.
#' @param ... Reserved for methods.
#' @return A [dist_structure] object.
#' @export
as_dist_structure <- function(x, ...) UseMethod("as_dist_structure")

# ==========================================================================
# Topology generics and defaults
# ==========================================================================


test_that("series_dist has correct phi and min_paths", {
  sys <- series_dist(iid_exp_components(3))
  expect_equal(ncomponents(sys), 3L)
  expect_equal(phi(sys, c(1, 1, 1)), 1L)
  expect_equal(phi(sys, c(1, 1, 0)), 0L)
  expect_equal(min_paths(sys), list(1:3))
})


test_that("parallel_dist has correct phi and min_paths", {
  sys <- parallel_dist(iid_exp_components(3))
  expect_equal(phi(sys, c(0, 0, 0)), 0L)
  expect_equal(phi(sys, c(1, 0, 0)), 1L)
  expect_equal(length(min_paths(sys)), 3L)
})


test_that("kofn_dist has correct phi", {
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  expect_equal(phi(sys, c(1, 1, 0)), 1L)
  expect_equal(phi(sys, c(1, 0, 0)), 0L)
  expect_equal(phi(sys, c(1, 1, 1)), 1L)
})


test_that("bridge_dist has expected 4 minimal path sets", {
  sys <- bridge_dist(iid_exp_components(5))
  expect_equal(length(min_paths(sys)), 4L)
  expect_equal(ncomponents(sys), 5L)
})


test_that("min_cuts defaults via Berge transversal from min_paths", {
  # series of 3: min_paths = {1,2,3}, min_cuts = {1},{2},{3}
  series_sys <- series_dist(iid_exp_components(3))
  cuts <- min_cuts(series_sys)
  expect_equal(length(cuts), 3L)
  expect_setequal(unlist(cuts), c(1L, 2L, 3L))

  # parallel of 3: min_paths = {1},{2},{3}, min_cuts = {1,2,3}
  parallel_sys <- parallel_dist(iid_exp_components(3))
  cuts <- min_cuts(parallel_sys)
  expect_equal(length(cuts), 1L)
  expect_setequal(cuts[[1]], c(1L, 2L, 3L))
})


test_that("min_cuts returns distinct sets (no duplicates) for kofn", {
  # k-of-n where k > 1 and m > k + 1 used to produce duplicates because
  # the Berge-transversal extension generates equal-length transversals
  # through independent paths. 2-of-3 has exactly 3 min cuts: all pairs.
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  cuts <- min_cuts(sys)
  expect_equal(length(cuts), 3L)
  # Verify distinct
  expect_equal(length(unique(cuts)), length(cuts))
  # Each cut is a pair (size m - k + 1 = 2)
  expect_true(all(vapply(cuts, length, integer(1L)) == 2L))
})


test_that("min_cuts of bridge has exactly 4 distinct sets", {
  # Classical bridge: cuts = {1,2}, {4,5}, {1,3,5}, {2,3,4}
  sys <- bridge_dist(iid_exp_components(5))
  cuts <- min_cuts(sys)
  expect_equal(length(cuts), 4L)
  expect_equal(length(unique(cuts)), 4L)
})


test_that("critical_states default returns expected count for k-of-m", {
  # For k-of-m, component j is critical iff exactly (k-1) of the other
  # (m-1) components are functioning.
  for (m in 3:4) for (k in 1:m) {
    sys <- kofn_dist(k, iid_exp_components(m))
    expect_equal(nrow(critical_states(sys, 1L)),
                 choose(m - 1L, k - 1L))
  }
})


test_that("system_lifetime of k-of-m is the (m - k + 1)-th order statistic", {
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  # 2-of-3 fails when m - k + 1 = 2 components have failed.
  expect_equal(system_lifetime(sys, c(5, 1, 3)), 3)
  # Series (k = m): fails at min
  expect_equal(system_lifetime(series_dist(iid_exp_components(4)),
                               c(2, 5, 1, 3)), 1)
  # Parallel (k = 1): fails at max
  expect_equal(system_lifetime(parallel_dist(iid_exp_components(4)),
                               c(2, 5, 1, 3)), 5)
})


test_that("system_censoring classifies components correctly", {
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  cens <- system_censoring(sys, c(1, 3, 5))
  expect_equal(cens$system_time, 3)
  expect_equal(cens$component_status, c("left", "exact", "right"))
})


test_that("is_coherent returns TRUE for valid systems", {
  expect_true(is_coherent(series_dist(iid_exp_components(3))))
  expect_true(is_coherent(parallel_dist(iid_exp_components(3))))
  expect_true(is_coherent(kofn_dist(2L, iid_exp_components(4))))
  expect_true(is_coherent(bridge_dist(iid_exp_components(5))))
})


test_that("structural_importance is symmetric for k-of-m and matches formula", {
  sys <- kofn_dist(k = 2L, iid_exp_components(4))
  imps <- vapply(1:4, function(j) structural_importance(sys, j), 0)
  expect_equal(imps, rep(imps[[1]], 4L))
  # For k-of-m: choose(m - 1, k - 1) / 2^(m - 1)
  expect_equal(imps[[1]], choose(3L, 1L) / 2^3)
})


test_that("reliability at iid p = 0.5 matches expected for k-of-m", {
  # For k-of-m at iid p: R(p) = P(Binomial(m, p) >= k)
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  for (p in c(0.2, 0.5, 0.7)) {
    expect_equal(
      reliability(sys, p),
      sum(stats::dbinom(2:3, size = 3, prob = p)),
      tolerance = 1e-12
    )
  }
})


test_that("reliability of series at iid p is p^m", {
  sys <- series_dist(iid_exp_components(4))
  for (p in c(0.1, 0.5, 0.9)) {
    expect_equal(reliability(sys, p), p^4, tolerance = 1e-12)
  }
})


test_that("is_dist_structure recognizes subclasses", {
  expect_true(is_dist_structure(series_dist(iid_exp_components(2))))
  expect_true(is_dist_structure(parallel_dist(iid_exp_components(2))))
  expect_true(is_dist_structure(kofn_dist(1L, iid_exp_components(2))))
  expect_false(is_dist_structure(list()))
  expect_false(is_dist_structure(algebraic.dist::exponential(1)))
})

# ==========================================================================
# Closed-form specializations: exp_series, wei_series, wei_homogeneous_series
# ==========================================================================
#
# Each specialization is tested against:
#   (a) a known analytic identity
#   (b) the general dist_structure default (correctness check)
# ==========================================================================


test_that("exp_series has class hierarchy including series_dist", {
  sys <- exp_series(c(0.5, 0.3, 0.2))
  expect_s3_class(sys, "exp_series")
  expect_s3_class(sys, "series_dist")
  expect_s3_class(sys, "dist_structure")
  expect_s3_class(sys, "dist")
  expect_equal(ncomponents(sys), 3L)
})


test_that("exp_series surv matches Exp(sum(rates))", {
  rates <- c(0.5, 0.3, 0.2)
  sys <- exp_series(rates)
  S <- algebraic.dist::surv(sys)
  lam <- sum(rates)
  for (ti in c(0.1, 0.5, 1, 2, 5)) {
    expect_equal(S(ti), exp(-lam * ti), tolerance = 1e-12)
  }
})


test_that("exp_series sampler mean matches 1 / sum(rates)", {
  rates <- c(0.5, 0.3, 0.2)
  sys <- exp_series(rates)
  withr::local_seed(42)
  samples <- algebraic.dist::sampler(sys)(10000L)
  expect_equal(mean(samples), 1 / sum(rates), tolerance = 0.05)
})


test_that("exp_series mean is 1 / sum(rates) exactly", {
  rates <- c(1, 2, 3)
  sys <- exp_series(rates)
  expect_equal(mean(sys), 1 / 6, tolerance = 1e-12)
})


test_that("exp_series agrees with general series_dist on surv", {
  rates <- c(0.5, 0.3, 0.2)
  sys_special <- exp_series(rates)
  sys_general <- series_dist(lapply(rates, function(r)
    algebraic.dist::exponential(r)))
  S1 <- algebraic.dist::surv(sys_special)
  S2 <- algebraic.dist::surv(sys_general)
  for (ti in c(0.2, 1, 3)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-10)
  }
})


test_that("wei_series has correct class hierarchy", {
  sys <- wei_series(shapes = c(1, 2), scales = c(1, 2))
  expect_s3_class(sys, "wei_series")
  expect_s3_class(sys, "series_dist")
  expect_s3_class(sys, "dist_structure")
})


test_that("wei_series surv matches closed-form cumulative hazard", {
  shapes <- c(1, 2, 3)
  scales <- c(1, 2, 3)
  sys <- wei_series(shapes, scales)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.3, 1, 2)) {
    expected <- exp(-sum((ti / scales)^shapes))
    expect_equal(S(ti), expected, tolerance = 1e-12)
  }
})


test_that("wei_series sampler produces values consistent with Monte Carlo surv", {
  shapes <- c(2, 2, 2)
  scales <- c(1, 2, 3)
  sys <- wei_series(shapes, scales)
  withr::local_seed(1)
  samples <- algebraic.dist::sampler(sys)(10000L)
  t0 <- 0.5
  empirical_surv <- mean(samples > t0)
  expected_surv <- algebraic.dist::surv(sys)(t0)
  expect_equal(empirical_surv, expected_surv, tolerance = 0.02)
})


test_that("wei_series reduces to Exp when all shapes are 1", {
  # Series of Weibull(1, 1/rate) is Exp(rate); series of them is Exp(sum rates).
  scales <- c(1, 2, 4)  # corresponds to rates 1, 0.5, 0.25
  rates <- 1 / scales
  sys <- wei_series(shapes = rep(1, 3), scales = scales)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.3, 1, 2)) {
    expect_equal(S(ti), exp(-sum(rates) * ti), tolerance = 1e-12)
  }
})


test_that("wei_homogeneous_series aggregates into a single Weibull", {
  shape <- 2
  scales <- c(1, 2, 3)
  sys <- wei_homogeneous_series(shape, scales)
  agg_scale <- (sum(1 / scales^shape))^(-1 / shape)
  S <- algebraic.dist::surv(sys)
  for (ti in c(0.2, 0.7, 1.5)) {
    expected <- exp(-(ti / agg_scale)^shape)
    expect_equal(S(ti), expected, tolerance = 1e-12)
  }
})


test_that("wei_homogeneous_series mean matches Weibull formula", {
  shape <- 3
  scales <- c(1, 1.5, 2)
  sys <- wei_homogeneous_series(shape, scales)
  agg_scale <- (sum(1 / scales^shape))^(-1 / shape)
  expected <- agg_scale * gamma(1 + 1 / shape)
  expect_equal(mean(sys), expected, tolerance = 1e-12)
})


test_that("wei_homogeneous_series is a wei_series subclass and inherits methods", {
  sys <- wei_homogeneous_series(shape = 2, scales = c(1, 2))
  expect_s3_class(sys, "wei_homogeneous_series")
  expect_s3_class(sys, "wei_series")
  expect_s3_class(sys, "series_dist")
  # component() still works
  c1 <- component(sys, 1L)
  expect_s3_class(c1, "weibull_dist")
  expect_equal(c1$scale, 1)
  expect_equal(c1$shape, 2)
})


test_that("wei_homogeneous_series agrees with wei_series on survival", {
  shape <- 2
  scales <- c(1, 2, 3)
  sys_hom <- wei_homogeneous_series(shape, scales)
  sys_het <- wei_series(rep(shape, 3), scales)
  S1 <- algebraic.dist::surv(sys_hom)
  S2 <- algebraic.dist::surv(sys_het)
  for (ti in c(0.2, 1, 2)) {
    expect_equal(S1(ti), S2(ti), tolerance = 1e-12)
  }
})


test_that("exp_series is faster than general series_dist on surv (smoke test)", {
  # Not a strict benchmark, but verifies both paths produce matching values.
  rates <- runif(5, 0.1, 2)
  sys_special <- exp_series(rates)
  sys_general <- series_dist(lapply(rates, algebraic.dist::exponential))
  ts <- c(0.1, 0.5, 1, 2)
  S1 <- algebraic.dist::surv(sys_special)(ts)
  S2 <- algebraic.dist::surv(sys_general)(ts)
  expect_equal(S1, S2, tolerance = 1e-10)
})

# ==========================================================================
# Dual system tests
# ==========================================================================


test_that("dual of series is parallel-like via phi", {
  sys <- series_dist(iid_exp_components(3))
  dsys <- dual(sys)
  # dual phi(state) = 1 - phi(1 - state); series.phi(1-state) = 1 iff all 1 flipped, i.e., state = 0
  expect_equal(phi(dsys, c(0, 0, 0)), 0L)
  expect_equal(phi(dsys, c(1, 0, 0)), 1L)
  expect_equal(phi(dsys, c(1, 1, 1)), 1L)
})


test_that("dual of k-of-m satisfies phi identity", {
  sys <- kofn_dist(k = 2L, iid_exp_components(4))
  dsys <- dual(sys)
  for (a in 0:1) for (b in 0:1) for (c in 0:1) for (d in 0:1) {
    state <- c(a, b, c, d)
    expect_equal(phi(dsys, state), 1L - phi(sys, 1L - state))
  }
})


test_that("dual of dual equals original (involution)", {
  sys <- kofn_dist(k = 2L, iid_exp_components(3))
  dd <- dual(dual(sys))
  # Either returns the same object (coherent_dist path via min_cuts) or a
  # double-wrapped equivalent. Check phi-equivalence.
  for (a in 0:1) for (b in 0:1) for (c in 0:1) {
    state <- c(a, b, c)
    expect_equal(phi(dd, state), phi(sys, state))
  }
})


test_that("dual_of_system inherits dist_structure defaults", {
  sys <- series_dist(iid_exp_components(3))
  dsys <- dual(sys)  # falls through to coherent_dist method producing a new coherent_dist
  expect_true(is_dist_structure(dsys))
  expect_equal(ncomponents(dsys), 3L)
})


test_that("coherent_dist dual swaps paths and cuts", {
  sys <- coherent_dist(
    min_paths = list(c(1L, 2L), c(3L)),
    components = iid_exp_components(3)
  )
  dsys <- dual(sys)
  # dual min_paths should equal original min_cuts
  expect_setequal(
    lapply(min_paths(dsys), sort),
    lapply(min_cuts(sys), sort)
  )
})


test_that("dual_of_system has component() (regression: lazy-wrapper dispatch)", {
  # The package-level closed-form classes route through dual.coherent_dist
  # and produce a fresh coherent_dist (not a dual_of_system), so they do
  # not exercise the lazy-wrapper path. This test directly constructs a
  # dual_of_system to verify that component() dispatches correctly --
  # before the C1 fix in v0.5.x, component(dual_of_system_obj, j) errored
  # with "no applicable method", breaking surv() / sampler() defaults
  # for any user-defined dist_structure subclass that did not also
  # inherit coherent_dist.
  sys <- series_dist(iid_exp_components(3))
  dsys <- structure(
    list(original = sys, m = ncomponents(sys)),
    class = c("dual_of_system", "dist_structure",
              "univariate_dist", "continuous_dist", "dist")
  )

  expect_no_error(component(dsys, 1L))
  expect_identical(component(dsys, 1L), component(sys, 1L))
  expect_identical(component(dsys, 2L), component(sys, 2L))
  expect_identical(component(dsys, 3L), component(sys, 3L))
})

# ==========================================================================
# Compositional operations: substitute_component, compose_systems
# ==========================================================================


test_that("substitute_component preserves topology", {
  sys <- series_dist(iid_exp_components(3, rate = 1))
  new_comp <- algebraic.dist::exponential(5)
  sys2 <- substitute_component(sys, 2L, new_comp)
  expect_equal(ncomponents(sys2), 3L)
  # phi unchanged
  expect_equal(phi(sys2, c(1, 1, 1)), 1L)
  expect_equal(phi(sys2, c(1, 0, 1)), 0L)
  # component 2 is the replacement
  c2 <- component(sys2, 2L)
  expect_s3_class(c2, "exponential")
  expect_equal(c2$rate, 5)
})


test_that("substitute_component changes system survival", {
  sys <- exp_series(c(1, 1, 1))
  # Original: Exp(3). Replace component 2 with Exp(2). New: Exp(1+2+1) = Exp(4).
  sys2 <- substitute_component(sys, 2L, algebraic.dist::exponential(2))
  S2 <- algebraic.dist::surv(sys2)
  t0 <- 0.5
  expected <- exp(-4 * t0)
  expect_equal(S2(t0), expected, tolerance = 1e-10)
})


test_that("compose_systems of series-of-series flattens to a bigger series", {
  # A series of two series sub-systems (each with 2 components) has 4
  # components in series, min_paths = {{1,2,3,4}}.
  inner_a <- series_dist(iid_exp_components(2, rate = 1))
  inner_b <- series_dist(iid_exp_components(2, rate = 1))
  outer <- series_dist(list(inner_a, inner_b))
  composed <- compose_systems(outer, list(inner_a, inner_b))
  expect_equal(ncomponents(composed), 4L)
  paths <- min_paths(composed)
  expect_equal(length(paths), 1L)
  expect_setequal(paths[[1]], 1:4)
})


test_that("compose_systems of series-of-parallel has the right min_paths", {
  # series_dist(list(parallel_of_2, parallel_of_2)):
  #   outer min_paths = {{1, 2}} (series of 2 outer components)
  #   inner_a min_paths = {{1}, {2}}
  #   inner_b min_paths = {{1}, {2}}
  # Composed min_paths: pick one inner path for each outer component in
  # the outer path, union. Four combinations:
  #   {1} + {3} = {1, 3}
  #   {1} + {4} = {1, 4}
  #   {2} + {3} = {2, 3}
  #   {2} + {4} = {2, 4}
  inner_a <- parallel_dist(iid_exp_components(2, rate = 1))
  inner_b <- parallel_dist(iid_exp_components(2, rate = 1))
  # outer structure: series of 2 things, each a "component" representing
  # one inner
  outer_comps <- lapply(1:2, function(i) algebraic.dist::exponential(1))
  outer <- series_dist(outer_comps)
  composed <- compose_systems(outer, list(inner_a, inner_b))
  expect_equal(ncomponents(composed), 4L)
  paths <- min_paths(composed)
  expect_equal(length(paths), 4L)
  path_sorted <- lapply(paths, sort)
  expect_true(list(c(1L, 3L)) %in% lapply(path_sorted, as.integer) ||
              any(vapply(path_sorted, function(P)
                setequal(P, c(1L, 3L)), logical(1L))))
  # Check each expected path appears
  expected_sets <- list(c(1L, 3L), c(1L, 4L), c(2L, 3L), c(2L, 4L))
  for (E in expected_sets) {
    expect_true(any(vapply(path_sorted, function(P)
      setequal(P, E), logical(1L))))
  }
})


test_that("compose_systems of parallel-of-series gives one path with all components", {
  # parallel of 2 series sub-systems (each with 2 components):
  #   outer min_paths = {{1}, {2}}
  #   inner_a min_paths = {{1, 2}}
  #   inner_b min_paths = {{1, 2}}  (shifted: {{3, 4}})
  # Composed paths: for each outer path {k}, union of inner paths for k.
  #   {1}: inner_a paths = {{1, 2}}, one choice, gives {1, 2}
  #   {2}: inner_b paths = {{3, 4}}, one choice, gives {3, 4}
  inner_a <- series_dist(iid_exp_components(2, rate = 1))
  inner_b <- series_dist(iid_exp_components(2, rate = 1))
  outer_comps <- lapply(1:2, function(i) algebraic.dist::exponential(1))
  outer <- parallel_dist(outer_comps)
  composed <- compose_systems(outer, list(inner_a, inner_b))
  expect_equal(ncomponents(composed), 4L)
  paths <- min_paths(composed)
  expect_equal(length(paths), 2L)
  path_sorted <- lapply(paths, sort)
  expected_sets <- list(c(1L, 2L), c(3L, 4L))
  for (E in expected_sets) {
    expect_true(any(vapply(path_sorted, function(P)
      setequal(P, E), logical(1L))))
  }
})


test_that("compose_systems with plain dist as inner treats it as single component", {
  outer <- series_dist(lapply(1:2, function(i) algebraic.dist::exponential(1)))
  composed <- compose_systems(outer,
                              list(algebraic.dist::exponential(1),
                                   algebraic.dist::exponential(2)))
  # Single-component inner: min_paths = {{1}}; combined with series outer
  # path {{1, 2}}, composed min_paths = {{1, 2}}.
  expect_equal(ncomponents(composed), 2L)
  paths <- min_paths(composed)
  expect_equal(length(paths), 1L)
  expect_equal(sort(paths[[1]]), 1:2)
})

# ==========================================================================
# Coercions
# ==========================================================================


test_that("as_dist_structure on a plain dist returns a 1-component series", {
  d <- algebraic.dist::exponential(2)
  ds <- as_dist_structure(d)
  expect_true(is_dist_structure(ds))
  expect_s3_class(ds, "series_dist")
  expect_equal(ncomponents(ds), 1L)
  c1 <- component(ds, 1L)
  expect_s3_class(c1, "exponential")
  expect_equal(c1$rate, 2)
})


test_that("as_dist_structure on a dist_structure returns it unchanged", {
  sys <- series_dist(iid_exp_components(3))
  result <- as_dist_structure(sys)
  expect_identical(result, sys)
})


test_that("as_dist_structure preserves dist methods", {
  d <- algebraic.dist::exponential(1)
  ds <- as_dist_structure(d)
  S_d <- algebraic.dist::surv(d)
  S_ds <- algebraic.dist::surv(ds)
  for (ti in c(0.5, 1, 2)) {
    expect_equal(S_d(ti), S_ds(ti), tolerance = 1e-12)
  }
})

# ==========================================================================
# Cold-standby system
# ==========================================================================


test_that("cold_standby_dist mean equals sum of component means", {
  comps <- list(
    algebraic.dist::exponential(1),     # mean 1
    algebraic.dist::exponential(2),     # mean 0.5
    algebraic.dist::gamma_dist(2, 1)    # mean 2
  )
  sys <- cold_standby_dist(comps)
  expect_equal(mean(sys), 1 + 0.5 + 2, tolerance = 1e-10)
})


test_that("cold_standby_dist of m iid Exp(rate) sampler ~ Gamma(m, rate)", {
  m <- 3L
  rate <- 2
  sys <- cold_standby_dist(iid_exp_components(m, rate))
  withr::local_seed(1)
  x <- algebraic.dist::sampler(sys)(10000L)
  # E[Gamma(m, rate)] = m / rate
  expect_equal(mean(x), m / rate, tolerance = 0.05)
  # Var = m / rate^2
  expect_equal(var(x), m / rate^2, tolerance = 0.1)
})


test_that("cold_standby_dist surv via Monte Carlo approximates Gamma surv", {
  m <- 4L
  rate <- 1
  sys <- cold_standby_dist(iid_exp_components(m, rate))
  withr::local_seed(1)
  S_mc <- algebraic.dist::surv(sys)
  for (ti in c(2, 4, 6)) {
    mc_estimate <- S_mc(ti, mc = 50000L)
    exact <- stats::pgamma(ti, shape = m, rate = rate, lower.tail = FALSE)
    expect_equal(mc_estimate, exact, tolerance = 0.02)
  }
})


test_that("cold_standby_dist surv closure caches samples (deterministic across calls)", {
  sys <- cold_standby_dist(iid_exp_components(3))
  withr::local_seed(42)
  S <- algebraic.dist::surv(sys)
  v1 <- S(2, mc = 1000L)
  v2 <- S(2, mc = 1000L)  # same mc -> uses cached samples
  expect_identical(v1, v2)
  v3 <- S(c(1, 2, 3), mc = 1000L)
  expect_identical(v3[2], v2)
})


test_that("cold_standby_dist surv regenerates when mc changes", {
  sys <- cold_standby_dist(iid_exp_components(3))
  withr::local_seed(42)
  S <- algebraic.dist::surv(sys)
  v_small <- S(2, mc = 1000L)
  v_large <- S(2, mc = 5000L)
  # different sample counts means different cache; values may differ
  expect_true(is.numeric(v_small) && is.numeric(v_large))
})


test_that("cold_standby_dist mean falls back to MC for components without exact mean", {
  # A series_dist of 2 exponentials does not provide a `mean()` method;
  # mean() routes through univariate_dist::mean -> expectation -> sup,
  # which is not implemented for dist_structure. cold_standby_dist's
  # mean.cold_standby_dist must catch this and fall back to MC.
  inner <- series_dist(iid_exp_components(2))
  sys <- cold_standby_dist(list(inner, algebraic.dist::exponential(1)))
  withr::local_seed(1)
  m <- mean(sys)
  # Inner is min of 2 iid Exp(1) = Exp(2), mean 0.5; outer Exp(1) mean 1.
  # Total expected ~ 1.5; MC tolerance ~5%.
  expect_equal(m, 1.5, tolerance = 0.1)
})


test_that("cold_standby_dist class does NOT include dist_structure", {
  sys <- cold_standby_dist(list(algebraic.dist::exponential(1)))
  expect_false(is_dist_structure(sys))
  expect_s3_class(sys, "cold_standby_dist")
  expect_s3_class(sys, "dist")
})


test_that("cold_standby_dist exposes ncomponents and component", {
  comps <- list(algebraic.dist::exponential(1),
                algebraic.dist::gamma_dist(2, 1))
  sys <- cold_standby_dist(comps)
  expect_equal(ncomponents(sys), 2L)
  expect_s3_class(component(sys, 1L), "exponential")
  expect_s3_class(component(sys, 2L), "gamma_dist")
})


test_that("cold_standby_dist sampler returns positive values of correct length", {
  sys <- cold_standby_dist(list(
    algebraic.dist::exponential(1),
    algebraic.dist::exponential(1)
  ))
  withr::local_seed(1)
  x <- algebraic.dist::sampler(sys)(50L)
  expect_length(x, 50L)
  expect_true(all(x > 0))
})


test_that("criticality_importance validates j is in range", {
  sys <- series_dist(iid_exp_components(3))
  expect_error(criticality_importance(sys, 0L, 0.5))
  expect_error(criticality_importance(sys, 4L, 0.5))
  expect_error(criticality_importance(sys, c(1L, 2L), 0.5))
})

---
title: "Importance measures"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Importance measures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## Four measures, four questions

Importance measures quantify how much a component matters to system
reliability. Different definitions answer different questions.
`dist.structure` provides four classical measures:

| Measure | Question |
|---|---|
| Structural | How often is j pivotal across all states of the other components? |
| Birnbaum (reliability) | At component reliabilities p, how much does dR/dp_j change R? |
| Criticality | Given the system failed by t, what fraction attributable to j failing and being critical? |
| Vesely-Fussell | Given the system failed by t, what is the probability that some cut set containing j is fully failed? |

Each reduces to a closed form in simple topologies but diverges on
complex ones. The bridge network is a canonical case where all four
matter, so we'll use it throughout.

## Setting up the bridge

```{r}
sys <- bridge_dist(replicate(5, exponential(1), simplify = FALSE))
min_paths(sys)
min_cuts(sys)
```

Components 1 and 2 appear in two minimal paths each; components 4 and
5 likewise. Component 3 (the cross-link) appears in two paths but not
the "primary" ones `{1,4}` and `{2,5}`.

## Structural importance

Structural importance is purely topological: it ignores component
distributions.

```{r}
sapply(1:5, function(j) structural_importance(sys, j))
```

For the bridge, components 1, 2, 4, 5 are structurally equivalent;
component 3 is less structurally important (smaller fraction of
pivotal states).

## Birnbaum reliability importance

Birnbaum importance is `dR/dp_j` at given component reliabilities `p`:

```{r}
p <- 0.9
sapply(1:5, function(j) birnbaum_importance(sys, j, p))
```

At iid p = 0.9, the ordering agrees with structural importance (same
symmetry). Where they diverge is at non-iid p or time-varying p.

## Time-dependent importance via criticality

Criticality importance turns Birnbaum into a time-varying quantity by
plugging in `p_j(t) = S_j(t)`:

```{r}
t_vals <- c(0.1, 0.5, 1, 2)
sapply(t_vals, function(t) {
  sapply(1:5, function(j) criticality_importance(sys, j, t))
})
```

Each column is a time point; each row is a component. Read it as: "at
time t, what fraction of system-failure probability is attributable to
component j having failed AND being critical".

At small t (high reliability), the breakdown looks like structural
importance (because all components have similar F_j(t)). At larger t,
the contributions of each component to system failure diverge.

## Vesely-Fussell importance

Vesely-Fussell uses minimal cut sets instead of direct pivotality:

```{r}
t0 <- 1
sapply(1:5, function(j) vesely_fussell_importance(sys, j, t0))
```

Component j's Vesely-Fussell importance at time t is the conditional
probability that some minimal cut set containing j is fully failed,
given that the system has failed by t. It tends to be larger than
criticality importance for components that appear in small cuts.

## Comparison

```{r}
# Build a comparison table at p = 0.9 and t s.t. S(t) = 0.9 for all components.
# For iid Exp(1), S(t) = 0.9 -> t = -log(0.9)
t_star <- -log(0.9)

tab <- data.frame(
  j = 1:5,
  structural = sapply(1:5, function(j) structural_importance(sys, j)),
  birnbaum = sapply(1:5, function(j) birnbaum_importance(sys, j, 0.9)),
  criticality = sapply(1:5, function(j) criticality_importance(sys, j, t_star)),
  vesely_fussell = sapply(1:5, function(j) vesely_fussell_importance(sys, j, t_star))
)
tab
```

For the bridge at iid p = 0.9 (equivalent to t = t_star), all four
measures agree on the ordering (components 1, 2, 4, 5 identical;
component 3 different). Under heterogeneous reliabilities the measures
can rank components differently.

## Closed-form checks

Some importance values have simple closed forms worth remembering:

```{r}
# Series: Birnbaum importance of j equals product of the OTHER p's.
series3 <- series_dist(replicate(3, exponential(1), simplify = FALSE))
p_vec <- c(0.9, 0.8, 0.7)
birnbaum_importance(series3, j = 1, p = p_vec)
prod(p_vec[-1])
```

```{r}
# Parallel: Birnbaum importance of j equals product of the OTHER (1 - p_i).
parallel3 <- parallel_dist(replicate(3, exponential(1), simplify = FALSE))
birnbaum_importance(parallel3, j = 1, p = p_vec)
prod(1 - p_vec[-1])
```

```{r}
# Series: Vesely-Fussell equals F_j / F_sys at time t, since each
# component forms its own min cut.
t0 <- 1
F_j_1 <- 1 - exp(-t0)
F_sys <- 1 - exp(-3 * t0)   # Exp(sum(rates) = 3)
vesely_fussell_importance(series3, j = 1, t = t0)
F_j_1 / F_sys
```

## When to use which

- **Structural**: system design, component selection, topology
  comparison. Independent of component reliability values; useful
  before you know the components.
- **Birnbaum**: design-point sensitivity. You pick a design point `p`
  and ask which component contributes most to dR/dp.
- **Criticality**: failure-attribution analysis. After the system has
  failed by time t, which component is "to blame" most often?
- **Vesely-Fussell**: same time-dependent question as criticality but
  through cut-set lens. Useful when you think about failure in terms of
  minimal failure combinations.

All four are computed from the same primitives (`phi`, `min_cuts`, the
component survivals and failures) and the differences are in which
event you condition on and which counting you do.

---
title: "Distribution interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Distribution interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## dist_structure IS-A dist

`dist_structure` is a virtual S3 class that inherits from
`univariate_dist` and `dist`. Every concrete constructor returns an
object that participates in the full `algebraic.dist` distribution
algebra via inheritance:

```{r}
sys <- series_dist(list(
  exponential(0.5),
  exponential(0.3),
  exponential(0.2)
))
class(sys)
algebraic.dist::is_dist(sys)
```

## Default dist methods

When you call `surv`, `cdf`, `sampler` on a `dist_structure` object,
dispatch goes through three paths in order of specificity:

1. **Specialized closed-form methods** (e.g., `surv.exp_series`,
   `surv.wei_homogeneous_series`). Fastest; used when available.
2. **Topology shortcut methods** (e.g., `surv.series_dist`). Rare; most
   specializations land on the specific parametric class.
3. **The dist_structure default**, which composes component
   distributions through the structure function via the reliability
   polynomial identity:

   ```
   S_sys(t) = R(S_1(t), S_2(t), ..., S_m(t))
   ```

   where `R` is the multilinear extension of `phi`. This is how
   `coherent_dist(min_paths = ...)` gets a surv function for any
   topology, even one you specify by hand.

Samplers follow the analogous composition: sample each component
independently, then apply `system_lifetime` to reduce to a scalar.

## Choosing specializations vs general constructors

dist.structure provides both general and specialized constructors for
common cases. You should prefer the specialization when available:

```{r}
# General: any components, arbitrary topology
sys_general <- series_dist(replicate(3, exponential(1), simplify = FALSE))

# Specialization: exploits Exp(sum(rates)) closed form
sys_special <- exp_series(c(1, 1, 1))
```

Both return identical distributions, but the specialization's `surv`
is a one-liner (`exp(-total_rate * t)`) while the general version
computes `prod(exp(-rate_j * t))` per component per t. Same math,
different cost.

Verify:

```{r}
for (ti in c(0.5, 1, 2)) {
  stopifnot(isTRUE(all.equal(
    algebraic.dist::surv(sys_general)(ti),
    algebraic.dist::surv(sys_special)(ti),
    tolerance = 1e-10
  )))
}
```

## Closed-form specializations in dist.structure

| Constructor | Closed form |
|---|---|
| `exp_series(rates)` | `Exp(sum(rates))` exactly |
| `wei_series(shapes, scales)` | `exp(-sum((t/scale)^shape))` for survival |
| `wei_homogeneous_series(shape, scales)` | single Weibull with aggregate scale |
| `gamma_series(shapes, rates)` | product of Gamma upper tails |
| `lognormal_series(meanlogs, sdlogs)` | product of Lognormal upper tails |
| `exp_parallel(rates)` | `1 - prod(1 - exp(-rate*t))` |
| `exp_kofn(k, rates)` | subset enumeration (`O(2^m)`) |
| `wei_kofn(k, shapes, scales)` | same for Weibull components |

Each has closed-form `surv`, `cdf`, `sampler`, and (where feasible)
`density`. The `surv` / `cdf` / `density` methods registered on these
subclasses dispatch before the `dist_structure` default.

## min, max, and order statistics as dist_structure

The base `algebraic.dist` has `min` and `max` operators on dists that
return plain dist objects. `dist.structure` provides structure-aware
counterparts:

```{r}
d <- exponential(1)

# Plain min: just a dist, no topology
d_min <- min(d, d, d)
class(d_min)
```

```{r}
# Structure-aware: a series_dist, topology preserved
sys_min <- min_iid(d, m = 3)
class(sys_min)
dist.structure::min_paths(sys_min)
```

Same survival function, but the structured version lets you ask `phi`,
`min_paths`, `structural_importance`, and so on.

Similarly:

```{r}
# Parallel: max of iid
sys_max <- max_iid(d, m = 3)
class(sys_max)

# k-th order statistic
sys_k <- order_statistic(d, k = 2, m = 5)
class(sys_k)
```

## Density and hazard

`density` and `hazard` have specialized implementations only where
efficient formulas exist:

```{r}
# Density of exp_series: dexp at the aggregate rate.
f <- density(exp_series(c(0.5, 0.3, 0.2)))
f(1)
dexp(1, rate = 1.0)
```

For general `coherent_dist` objects, `density` is typically computed
via `-d/dt surv(t)` numerically. If you need `density` on a system
where it matters for performance, register a specialized method or use
a specialization that ships one.

## The `dist.structure` + `algebraic.dist` stack

```
algebraic.dist                    dist, generics, arithmetic on RVs
    |
dist.structure                    dist with internal component structure
    |
specialized + general impls       exp_series, kofn_dist, coherent_dist, ...
    |
downstream packages               serieshaz, kofn, maskedcauses
```

Downstream packages (e.g., `serieshaz`'s `dfr_dist_series`) inherit
from `dist_structure` and get topology + importance + composition for
free. Users never need to know whether a given generic is provided by
the specialized class, the dist_structure default, or the
algebraic.dist base; S3 dispatch finds the most specific method and
that's the one that runs.

---
title: "Non-coherent systems: cold standby"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Non-coherent systems: cold standby}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## Why "non-coherent"

The structure-function framework assumes a monotone binary function
`phi: {0, 1}^m -> {0, 1}` that maps per-component states to a system
state. This captures a huge class of reliability topologies but not
everything practitioners build. Cold standby is a canonical example.

A **cold standby** system has m components but only one is active at a
time. When the active component fails, a dormant spare instantaneously
takes over. System lifetime is the sum of component lifetimes, not an
order statistic.

The topology is temporal succession, not a structure function on
states. So cold standby:

- is a valid distribution object (you can sample, compute survival, ask
  for mean),
- is NOT a `dist_structure`: it has no `phi`, no `min_paths`, no
  signature, no Birnbaum importance.

dist.structure provides `cold_standby_dist` for this case, with its
own class chain that deliberately does not inherit `dist_structure`.

## Constructing a cold standby

```{r}
sys <- cold_standby_dist(list(
  exponential(1),
  exponential(2),
  exponential(0.5)
))
class(sys)
dist.structure::is_dist_structure(sys)   # FALSE -- not a coherent system
```

## Mean and sampling: exact

`mean` is exact when every component has an exact mean method: it sums
the component means.

```{r}
mean(sys)                              # 1/1 + 1/2 + 1/0.5 = 3.5
```

The sampler is also exact: independent samples from each component,
summed.

```{r}
set.seed(1)
x <- algebraic.dist::sampler(sys)(5000)
mean(x)                                # empirical ~ 3.5
```

## Survival and CDF: Monte Carlo with caching

For general components, the density of a sum-of-independent-RVs has no
closed form (it's a convolution). dist.structure computes `surv` and
`cdf` via Monte Carlo sampling. The closure caches the samples on the
first call, so repeated evaluations at different `t` values are
deterministic given the same `mc`:

```{r}
set.seed(42)
S <- algebraic.dist::surv(sys)
S(c(1, 2, 3, 5), mc = 5000)           # survival at multiple t values
```

A different `mc` triggers a fresh draw:

```{r}
S(2, mc = 5000)                        # uses the same cache
S(2, mc = 10000)                       # new cache with 10000 samples
```

Because the cache lives inside the closure, using the same `S` closure
keeps results consistent. `cdf` has its own closure and its own
cache: `cdf(sys)(t) + surv(sys)(t)` computed independently is not
exactly 1. If you need that identity, use one `surv` closure and
compute the CDF from it:

```{r}
S <- algebraic.dist::surv(sys)
F_t <- function(t) 1 - S(t)
F_t(2) + S(2)                          # exactly 1
```

For reproducibility across separate closure constructions, seed before
each one:

```{r}
set.seed(1); S1 <- algebraic.dist::surv(sys); s1 <- S1(2)
set.seed(1); S2 <- algebraic.dist::surv(sys); s2 <- S2(2)
all.equal(s1, s2)
```

## Specialized closed-form case: iid exponentials

A cold standby of m iid `Exp(rate)` components has system lifetime
`Gamma(shape = m, rate = rate)`:

```{r}
m <- 4; rate <- 2
sys <- cold_standby_dist(replicate(m, exponential(rate), simplify = FALSE))
# Monte Carlo survival at t = 3
set.seed(1)
mc_est <- algebraic.dist::surv(sys)(3, mc = 10000)
# Exact via Gamma
exact <- pgamma(3, shape = m, rate = rate, lower.tail = FALSE)
cat(sprintf("MC: %.4f, Exact: %.4f\n", mc_est, exact))
```

If you hit this case often, you can avoid the Monte Carlo by
constructing an `algebraic.dist::gamma_dist` directly.

## What cold standby does NOT support

Generics requiring topology or component-level `density`/`sup` fail
cleanly:

```{r, error = TRUE}
dist.structure::phi(sys, c(1, 1, 1, 1))
```

```{r, error = TRUE}
dist.structure::min_paths(sys)
```

Methods inherited from `algebraic.dist::univariate_dist` that require
`density` or `sup` (notably `vcov`, `expectation`) are also not
supported by default. A specialized subclass with a closed-form
aggregate (e.g., iid exponential -> Gamma) can provide them by
overriding.

## When to reach for cold standby

- Sequential redundancy with dormant spares (perfect switching, zero
  standby hazard).
- Modelling sum-of-lifetimes directly as a distribution.
- Monte Carlo as a first-pass answer when no closed form is available.

When you need the full dist.structure protocol (topology, importance,
composition), cold standby is the wrong tool; use a coherent system
instead.

---
title: "Composition and substitution"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Composition and substitution}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## Building new systems from old

dist.structure provides two compositional operations:

- `substitute_component(sys, j, new_component)` swaps one component's
  distribution while preserving the topology.
- `compose_systems(outer, inner_list)` replaces each component of the
  outer system with an inner sub-system, flattening the components and
  enumerating the composed minimal paths.

These let you build hierarchical reliability networks without writing
min_paths by hand.

## Substituting a component

Start with a 3-component parallel system, then upgrade one component
to a longer-lived Weibull:

```{r}
sys <- parallel_dist(list(
  exponential(1),
  exponential(1),
  exponential(1)
))

# Survival at t = 1 for the all-exponential parallel:
algebraic.dist::surv(sys)(1)

sys2 <- substitute_component(sys, j = 1,
  new_component = weibull_dist(shape = 2, scale = 2))

# Survival improves because component 1 is now longer-lived:
algebraic.dist::surv(sys2)(1)
```

`substitute_component` returns a `coherent_dist` with the same
min_paths and the modified component list. Topology queries work
identically:

```{r}
length(min_paths(sys2))                # still 3 (parallel)
ncomponents(sys2)                      # still 3
```

## Hierarchical composition

`compose_systems` nests: give an outer topology and a list of inner
systems (one per outer component), and get back a flattened system
whose min_paths combine both levels of topology.

### Series of parallels

A common redundant design: two parallel blocks in series.

```{r}
inner_a <- parallel_dist(list(exponential(1), exponential(1)))
inner_b <- parallel_dist(list(exponential(1), exponential(1)))

# Outer: series of 2 "slots" (the components are placeholder; they get
# replaced by the inners during compose_systems).
outer <- series_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_a, inner_b))
ncomponents(sys)                       # 4 total leaf components
length(min_paths(sys))                 # 4 composed paths
```

The composed paths are the Cartesian product of inner paths within
each outer path. Outer path is `{1, 2}`; inner_a min_paths = `{1},
{2}`; inner_b shifted min_paths = `{3}, {4}`. Four combinations:
`{1, 3}, {1, 4}, {2, 3}, {2, 4}`.

```{r}
min_paths(sys)
```

### Parallel of series

The dual: two series blocks in parallel.

```{r}
inner_a <- series_dist(list(exponential(1), exponential(1)))
inner_b <- series_dist(list(exponential(1), exponential(1)))

outer <- parallel_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_a, inner_b))
ncomponents(sys)                       # 4 total
length(min_paths(sys))                 # 2 composed paths
min_paths(sys)                         # {1, 2} and {3, 4}
```

## Mixing a sub-system with a plain dist

`compose_systems` handles a mix of `dist_structure` inner systems and
plain `dist` inner components. A plain `dist` is treated as a single-
component atomic sub-system:

```{r}
# Series of: (a 2-parallel block) + (a single exponential)
inner_par <- parallel_dist(list(exponential(1), exponential(1)))
outer <- series_dist(list(exponential(1), exponential(1)))

sys <- compose_systems(outer, list(inner_par, exponential(1)))
ncomponents(sys)                       # 3: two from parallel + one single
min_paths(sys)                         # {1, 3} and {2, 3}
```

Component 3 is the atomic component from the second outer slot; it
must function jointly with either component 1 or component 2 for the
system to function.

## Arbitrary depth

Composition is arbitrarily nestable:

```{r}
# A parallel of two different sub-systems:
sub_a <- series_dist(list(exponential(1), exponential(1)))     # simple series
sub_b <- bridge_dist(replicate(5, exponential(1), simplify = FALSE))  # bridge

outer <- parallel_dist(list(exponential(1), exponential(1)))
combined <- compose_systems(outer, list(sub_a, sub_b))

ncomponents(combined)                  # 2 + 5 = 7 total
length(min_paths(combined))            # = 1 + 4 (series has 1 path, bridge has 4)
```

The structural richness compounds: you can analyze the whole system
via the full dist.structure protocol (sampler, surv, system_signature,
importance measures, dual, etc.).

## Substitution vs composition

- **`substitute_component`**: same topology, different component
  distribution. Use when exploring sensitivity to a single component's
  parameters or swapping one family for another.
- **`compose_systems`**: richer topology by nesting. Use when the
  system has hierarchical structure (redundant subsystems, modular
  designs, networks of networks).

Both produce objects that participate in the full dist.structure
protocol, so further substitutions and compositions compose cleanly.

---
title: "Implementing a dist_structure subclass"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Implementing a dist_structure subclass}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(dist.structure)
```

`dist.structure` is a protocol package: the closed-form constructors
(`exp_series`, `exp_kofn`, ...) and topology shortcuts (`series_dist`,
`kofn_dist`, `bridge_dist`, ...) are reference implementations of an S3
contract that anyone can extend. Downstream packages do exactly that:
`serieshaz` adds a `dfr_dist_series` class for arbitrary dynamic-failure-rate
series systems; `kofn` builds inference machinery on the closed-form
k-out-of-n classes; future packages will add their own.

This vignette walks through the protocol from the implementor's side: what
to provide, what you get for free, when to override defaults for
performance, and how to validate your subclass.

## The contract: four steps

Every `dist_structure` subclass provides:

1. **A class chain** that includes `"dist_structure"`, `"univariate_dist"`,
   and `"dist"`. Almost always also a sub-chain like `"coherent_dist"` if
   the system is coherent.
2. **`ncomponents(x)`**: number of components `m`.
3. **`component(x, j, ...)`**: the j-th component returned as an
   `algebraic.dist::dist` object with parameters baked in. The returned
   dist must be independently evaluable: `surv(component(x, j))(t)`,
   `sampler(component(x, j))(n)`, etc., must all work.
4. **At least one of**:
   - `phi(x, state)`: the structure function, returning 0 or 1 for any
     binary state vector of length `m`.
   - `min_paths(x)`: a list of integer vectors enumerating minimal
     functioning component subsets.

   `phi` and `min_paths` are dual: each has a default that derives from
   the other. If you provide *both*, they must be consistent
   (`phi(x, state) == 1` iff `state` covers some path in `min_paths(x)`).
   Inconsistent implementations produce silently inconsistent results,
   because `reliability` / `critical_states` / `is_coherent` route
   through `phi` while `min_cuts` / `system_signature` route through
   `min_paths`.

Everything else has a default that derives from those four primitives.
Defaults exist for `min_cuts`, `system_signature`, `critical_states`,
`reliability`, `is_coherent`, `structural_importance`, `system_lifetime`,
`system_censoring`, `dual`, `surv`, `cdf`, and `sampler`.

## A worked example: an alarm system

Consider a fire alarm with `m` redundant smoke sensors and one master
controller. The system **functions** if (i) at least `k` of the `m`
sensors function AND (ii) the master controller functions. This is not
exactly a k-out-of-n system (the master is special) and not exactly a
series-of-(parallel-of-sensors-with-master) (the master gates everything).
We can capture it directly as a custom `dist_structure`.

### Step 1: pick a representation

Components are indexed `1..m` for the sensors and `m+1` for the master.

### Step 2: write the constructor

```{r}
alarm_dist <- function(k, sensor_components, master_component) {
  m_sensors <- length(sensor_components)
  stopifnot(k >= 1L, k <= m_sensors,
            inherits(master_component, "dist"),
            all(vapply(sensor_components,
                       function(d) inherits(d, "dist"),
                       logical(1L))))
  structure(
    list(
      k = as.integer(k),
      m_sensors = as.integer(m_sensors),
      m = as.integer(m_sensors + 1L),
      components = c(sensor_components, list(master_component))
    ),
    class = c("alarm_dist", "dist_structure",
              "univariate_dist", "continuous_dist", "dist")
  )
}
```

The class chain declares membership in `dist_structure` and the
`algebraic.dist` ancestors. We bundle the master into the components
list so the `j`-th component is uniformly accessible.

### Step 3: implement the three required generics

```{r}
ncomponents.alarm_dist <- function(x) x$m

component.alarm_dist <- function(x, j, ...) {
  stopifnot(j >= 1L, j <= x$m)
  x$components[[j]]
}

phi.alarm_dist <- function(x, state) {
  stopifnot(length(state) == x$m)
  sensor_state <- state[seq_len(x$m_sensors)]
  master_state <- state[x$m]
  as.integer(sum(sensor_state) >= x$k && master_state == 1L)
}
```

That's all that's strictly required by the protocol.

### Step 4: validate

```{r}
sensors <- replicate(4, algebraic.dist::exponential(0.5), simplify = FALSE)
master <- algebraic.dist::exponential(0.1)  # master is most reliable
alarm <- alarm_dist(k = 2L, sensors, master)

validate_dist_structure(alarm)
```

If we had forgotten `phi.alarm_dist`, `validate_dist_structure()` would
have stopped with a clear error pointing at the missing primitive
instead of letting the failure surface much later in some default
method's dispatch.

## What you get for free

Every default below is now available on `alarm`:

```{r}
# Number of components
ncomponents(alarm)

# Structure function evaluation
phi(alarm, c(1, 1, 0, 0, 1))   # 2 sensors + master alive: alarms
phi(alarm, c(1, 1, 1, 1, 0))   # all sensors but no master: silent

# Minimal cut sets (derived from phi via the Berge transversal)
min_cuts(alarm)

# Critical states for component j
critical_states(alarm, 5L)  # the master: critical in every state where >= k sensors are alive

# Reliability at uniform component reliability p
reliability(alarm, 0.9)

# Distribution algebra: system survival via component composition
S <- algebraic.dist::surv(alarm)
S(c(1, 5, 10))

# Sampling: m independent component lifetimes, combined via system_lifetime
withr::with_seed(1, {
  x <- algebraic.dist::sampler(alarm)(1000)
  mean(x)
})
```

Notice that `min_cuts`, `critical_states`, `reliability`, `surv`, and
`sampler` all "just work" without being implemented for `alarm_dist`.
They composed from `ncomponents`, `component`, `phi`, and the
component-level distribution algebra.

## When to override a default

Defaults are designed for correctness across all topologies. They are
not always the fastest path. The two most common reasons to override:

**1. Closed-form distributional shortcuts.**

The default `surv.dist_structure` enumerates over component states via
the reliability polynomial, which is `O(2^m)`. If your subclass has an
analytical formula for `S_sys(t)`, override `surv` directly:

```r
surv.alarm_dist <- function(x, ...) {
  k <- x$k
  m_sens <- x$m_sensors
  sensor_surv <- lapply(seq_len(m_sens),
                        function(j) algebraic.dist::surv(component(x, j)))
  master_surv <- algebraic.dist::surv(component(x, x$m))
  function(t, ...) {
    vapply(t, function(ti) {
      # P(>= k sensors alive at ti) * P(master alive at ti).
      ps <- vapply(sensor_surv, function(S) S(ti), numeric(1L))
      sensor_p <- sum(vapply(seq.int(k, m_sens), function(sz) {
        sum(vapply(utils::combn(m_sens, sz, simplify = FALSE),
                   function(A) {
                     prod(ps[A]) * prod(1 - ps[setdiff(seq_len(m_sens), A)])
                   }, numeric(1L)))
      }, numeric(1L)))
      sensor_p * master_surv(ti)
    }, numeric(1L))
  }
}
```

**2. Topology-specific shortcuts.**

Closed-form `min_paths` saves the Berge transversal from rederiving them.
For our alarm system, every minimal path is a `(k-of-m-sensors)` subset
unioned with the master:

```r
min_paths.alarm_dist <- function(x) {
  k <- x$k
  m_sens <- x$m_sensors
  master_idx <- x$m
  lapply(utils::combn(m_sens, k, simplify = FALSE),
         function(P) sort(c(as.integer(P), master_idx)))
}
```

With `min_paths.alarm_dist` defined, `min_cuts(alarm)` becomes the
default Berge transversal but starts from a small explicit list rather
than re-deriving paths from `phi` (which it would do via the default
chain through `binary_grid`).

The general rule: **provide what you know in closed form; rely on
defaults for the rest**. Do not pre-emptively override a default unless
you have a reason; correctness across all dispatch chains is much easier
when only the primitives are subclass-specific.

## Inheriting from `coherent_dist`

If your subclass is a coherent system, you can inherit from
`coherent_dist` and skip step 3 entirely:

```{r}
parallel_of_master_and_sensors <- function(sensor_components, master_component) {
  components <- c(sensor_components, list(master_component))
  m <- length(components)
  obj <- coherent_dist(
    min_paths = list(seq_len(m - 1L), m),  # all sensors as one path; master as another
    components = components,
    m = m
  )
  class(obj) <- c("master_or_sensors_dist", class(obj))
  obj
}
```

`coherent_dist` already implements `ncomponents`, `component`, and
`min_paths` for you; you supply only the topology-specific data
(min_paths and components). The class chain automatically includes
`"coherent_dist"`, so all the usual dispatch (including
`dual.coherent_dist`, which produces a real dual `coherent_dist` instead
of a lazy wrapper) takes over.

## What the default `dual` does

`dual(x)` produces the dual structure `phi_dual(state) = 1 - phi(x, 1 - state)`.
When `x` inherits from `coherent_dist`, the `dual.coherent_dist` method
swaps minimal paths and cuts to produce a fresh `coherent_dist`. When
`x` is a generic `dist_structure` (your own subclass), the default
returns a lazy wrapper of class `"dual_of_system"`. The lazy wrapper
implements `ncomponents`, `component`, and `phi` by deferring to the
original; everything else (signatures, importance, surv) flows through
default methods exactly as for any other `dist_structure`.

If you want `dual()` on your subclass to return something more
specific, override:

```r
dual.alarm_dist <- function(x) {
  # ... return a transformed alarm_dist ...
}
```

## Cross-references

- The closed-form specializations in this package (`exp_series.R`,
  `wei_kofn.R`, etc.) are concrete examples of subclasses that override
  `surv`, `cdf`, `sampler`, `density`, and `hazard` for speed while
  inheriting the topology defaults.
- `serieshaz::dfr_dist_series` is an out-of-package implementor that
  adds a fully general dynamic-failure-rate component family while
  participating in the protocol.
- `kofn::kofn` is an inference layer built on top of `exp_kofn` /
  `wei_kofn`; it does not subclass `dist_structure` itself but uses
  the protocol heavily.

For more detail on any specific generic, see the `?phi`, `?min_paths`,
`?dual`, `?reliability`, and `?dist_structure` help pages.

---
title: "Getting started with dist.structure"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dist.structure}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## What `dist.structure` is

`dist.structure` is a small, principled extension of `algebraic.dist`.
It lets you build and query **random variables with internal
structure**: coherent reliability systems decomposed into components
via a structure function.

Every object produced by `dist.structure` is also a `dist` (from
`algebraic.dist`), so the full distribution algebra (`mean`, `vcov`,
`sampler`, `surv`, `cdf`, `density`) works on it. On top of that,
`dist.structure` exposes structural queries you cannot ask of a plain
distribution: minimal path sets, system signature, Birnbaum
importance, duality, and more.

This vignette is a five-minute tour. Other vignettes go deeper into
specific topics.

## Your first structured distribution

Start with a series system of three independent exponentials:

```{r}
sys <- series_dist(list(
  exponential(0.5),
  exponential(0.3),
  exponential(0.2)
))
sys
```

The object is a dist. Query it like any distribution:

```{r}
algebraic.dist::surv(sys)(1)           # P(system survives past t = 1)
algebraic.dist::cdf(sys)(1)            # P(system fails by t = 1)
set.seed(1)
algebraic.dist::sampler(sys)(5)        # five system-lifetime samples
```

The series-of-exponentials identity `sum(rates) = total rate` holds:
the system is `Exp(sum(rates))`. Confirm:

```{r}
algebraic.dist::surv(sys)(1)
exp(-(0.5 + 0.3 + 0.2) * 1)
```

## Structural queries

Now ask topology questions the base distribution algebra cannot
answer:

```{r}
ncomponents(sys)                       # 3 components
phi(sys, c(1, 1, 0))                   # system functions? No (series needs all)
phi(sys, c(1, 1, 1))                   # Yes
min_paths(sys)                         # the single path: all components
min_cuts(sys)                          # three singleton cuts
system_signature(sys)                  # (1, 0, 0): fails at first failure
```

The structure function `phi` takes a component state vector (0 =
failed, 1 = functioning) and returns 1 if the system is functioning.
For a series system, `phi = AND`: all components must function.

## A richer topology: k-out-of-n

Series and parallel are special cases of k-out-of-n (functions if at
least k of m components function). dist.structure has a shortcut:

```{r}
# 2-out-of-3 of heterogeneous exponentials; closed-form specialization
kofn <- exp_kofn(k = 2, rates = c(1, 2, 3))
ncomponents(kofn)
algebraic.dist::surv(kofn)(1)
system_signature(kofn)                 # (0, 1, 0) for 2-of-3
```

The k-of-n system in dist.structure uses the :G convention: `k`
functioning components are required for the system to function.
Consequently `kofn_dist(k = 1, ...)` is parallel and `kofn_dist(k = m,
...)` is series.

## Sampling, means, and composition

Every distribution generic composes naturally:

```{r}
set.seed(42)
samples <- algebraic.dist::sampler(kofn)(5000)
mean(samples)
```

You can also substitute or compose components:

```{r}
# Swap component 2 with a Weibull.
kofn2 <- substitute_component(kofn, j = 2,
  new_component = weibull_dist(shape = 2, scale = 1))
algebraic.dist::surv(kofn2)(1)
```

## Importance measures

How much does each component matter? dist.structure provides four
complementary importance measures:

```{r}
# Structural importance: fraction of pivotal states for component 1.
structural_importance(kofn, j = 1)

# Birnbaum (reliability) importance: dR/dp_j at given component
# reliabilities.
birnbaum_importance(kofn, j = 1, p = 0.9)

# Criticality importance at time t = 0.5.
criticality_importance(kofn, j = 1, t = 0.5)

# Vesely-Fussell importance at time t = 0.5 (via minimal cut sets).
vesely_fussell_importance(kofn, j = 1, t = 0.5)
```

All four are different answers to "how important is component j?" and
they agree for simple systems but diverge for complex ones. The
[importance measures vignette](importance-measures.html) works through
the distinctions.

## What to read next

- **[Coherent systems](coherent-systems.html)**: the structural
  concepts (phi, min_paths, min_cuts, signature, reliability
  polynomial) in depth, with series, parallel, k-of-n, bridge, and
  consecutive-k examples.
- **[Distribution interface](distribution-interface.html)**: how
  dist_structure inherits from dist, which generics default via
  composition, and when to reach for a specialization (`exp_series`,
  `wei_kofn`) vs the general `coherent_dist`.
- **[Importance measures](importance-measures.html)**: the four
  importance measures applied to a bridge network, including
  time-dependent versions.
- **[Composition](composition.html)**: hierarchical construction via
  `substitute_component` and `compose_systems`.
- **[Non-coherent systems](non-coherent.html)**: cold standby, where
  system lifetime is a sum (not an order statistic), and why this is
  explicitly NOT a dist_structure.

## Ecosystem

`dist.structure` is the shared topology and DGP layer for the
reliability packages in the queelius ecosystem:

- `serieshaz` implements the protocol on `dfr_dist_series` (arbitrary
  dynamic failure rate components).
- `kofn` provides MLE for k-out-of-n systems, delegating its DGP to
  dist.structure.
- `maskedcauses` wraps its exponential-series legacy API over
  `dist.structure::exp_series`.

Any object from these packages that declares `dist_structure` in its
class vector inherits the full protocol automatically: topology
queries, importance measures, composition operators, and the full
`dist` interface.

---
title: "Coherent systems"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Coherent systems}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(dist.structure)
library(algebraic.dist)
```

## What is a coherent system

A **coherent system** is a reliability structure defined by a monotone
binary-valued structure function `phi: {0, 1}^m -> {0, 1}` where every
component is relevant. Coherent systems capture the vast majority of
practical reliability topologies: series (all components required),
parallel (any one suffices), k-out-of-n, bridge networks, consecutive-k
systems, and arbitrary coherent structures specified by minimal path
sets.

This vignette works through the core structural concepts and the
dist.structure constructors that realize them.

## The structure function `phi`

`phi(x)` returns 1 when the component state vector `x` makes the
system function:

```{r}
sys <- series_dist(list(exponential(1), exponential(1), exponential(1)))
phi(sys, c(1, 1, 1))                   # all up: series functions
phi(sys, c(1, 0, 1))                   # one down: series fails
```

```{r}
par_sys <- parallel_dist(list(exponential(1), exponential(1)))
phi(par_sys, c(0, 1))                  # one up: parallel functions
phi(par_sys, c(0, 0))                  # all down: parallel fails
```

## Minimal path and cut sets

A **minimal path set** is a minimal collection of components whose
simultaneous functioning guarantees the system functions. A **minimal
cut set** is a minimal collection whose simultaneous failure causes
system failure. Paths and cuts are dual specifications of phi.

```{r}
min_paths(sys)                         # series: one path, all components
min_cuts(sys)                          # series: m singleton cuts
```

```{r}
min_paths(par_sys)                     # parallel: m singleton paths
min_cuts(par_sys)                      # parallel: one cut, all components
```

The duality extends to k-of-n:

```{r}
sys_kofn <- kofn_dist(k = 2,
  components = list(exponential(1), exponential(1), exponential(1)))
length(min_paths(sys_kofn))            # choose(3, 2) = 3 paths
length(min_cuts(sys_kofn))             # choose(3, 2) = 3 cuts (pairs)
```

dist.structure derives `min_cuts` from `min_paths` via the Berge
transversal algorithm, so you only need to provide one or the other
when building a custom topology.

## The bridge network

The classic bridge network is the smallest system that is neither
series nor parallel nor k-of-n. dist.structure provides a constructor:

```{r}
bridge <- bridge_dist(replicate(5, exponential(1), simplify = FALSE))
length(min_paths(bridge))              # 4 minimal paths
length(min_cuts(bridge))               # 4 minimal cuts
system_signature(bridge)               # (0, 1/5, 3/5, 1/5, 0)
```

The Samaniego signature `(0, 1/5, 3/5, 1/5, 0)` encodes the probability
the system fails at the k-th component failure under iid component
lifetimes. For the bridge, 3/5 of the time the system fails at the 3rd
component failure.

## Arbitrary coherent systems

Supply minimal paths directly via `coherent_dist`:

```{r}
# Two parallel sub-systems combined in series:
# Paths: any path picks one component from each block.
custom <- coherent_dist(
  min_paths = list(c(1, 3), c(1, 4), c(2, 3), c(2, 4)),
  components = replicate(4, exponential(1), simplify = FALSE)
)
ncomponents(custom)
length(min_paths(custom))
length(min_cuts(custom))
```

## Structural importance

The Birnbaum structural importance of component j is the fraction of
states of the other m-1 components in which j is pivotal:

```{r}
structural_importance(sys_kofn, j = 1)     # symmetric: same for all j
```

For a k-of-m system at iid p, structural importance equals
`choose(m - 1, k - 1) / 2^(m - 1)`.

## Reliability polynomial

`reliability(system, p)` is the multilinear extension of phi to the
unit cube: given component reliabilities `p`, it returns the system
reliability. For series systems at iid p this is `p^m`:

```{r}
for (p in c(0.1, 0.5, 0.9)) {
  cat(sprintf("p = %.1f: reliability = %.4f\n",
              p, reliability(sys, p)))
}
```

For k-of-m at iid p it matches the binomial tail probability:

```{r}
# 2-of-3 at p = 0.8: P(Binom(3, 0.8) >= 2)
reliability(sys_kofn, 0.8)
sum(dbinom(2:3, size = 3, prob = 0.8))
```

## Critical states

For each component j, `critical_states(system, j)` enumerates states
of the other m-1 components in which j is pivotal:

```{r}
crit <- critical_states(sys_kofn, j = 1)
nrow(crit)                             # 2 (= choose(2, 1) for 2-of-3)
crit                                   # each row: a state of components 2, 3
```

Component 1 is critical iff exactly one of components 2 and 3 is
functioning.

## Dual systems

The dual of a coherent system swaps paths and cuts: `phi_dual(x) = 1 -
phi(1 - x)`. The dual of a series is parallel; the dual of a k-of-m is
(m - k + 1)-of-m.

```{r}
dsys <- dual(sys)
phi(dsys, c(0, 0, 0))                  # like parallel: fails only if all down
phi(dsys, c(1, 0, 0))                  # functions if any up
```

For `coherent_dist` objects, `dual` returns a new coherent_dist whose
min_paths are the original's min_cuts. This is a proper involution:
`dual(dual(sys))` has the same phi as `sys`.

## Validating a custom topology

`is_coherent` checks the axioms: monotonicity and component relevance.

```{r}
is_coherent(sys)                       # TRUE
is_coherent(bridge)                    # TRUE
```

A hypothetical system where phi ignores some component (violating
relevance) would fail this check.

## Summary

| Concept | Generic | What it returns |
|---|---|---|
| Structure function | `phi(sys, state)` | 0 or 1 |
| Minimal paths | `min_paths(sys)` | list of integer vectors |
| Minimal cuts | `min_cuts(sys)` | list of integer vectors |
| Signature | `system_signature(sys)` | numeric m-vector summing to 1 |
| Critical states | `critical_states(sys, j)` | integer matrix of m-1 cols |
| Structural importance | `structural_importance(sys, j)` | scalar in [0, 1] |
| Reliability polynomial | `reliability(sys, p)` | scalar in [0, 1] |
| Dual | `dual(sys)` | a dist_structure |
| Coherence check | `is_coherent(sys)` | TRUE / FALSE |

Every constructor (`series_dist`, `parallel_dist`, `kofn_dist`,
`bridge_dist`, `consecutive_k_dist`, `coherent_dist`, and the iid
shortcuts) produces an object on which all of these work.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercions.R, R/generics.R
\name{as_dist_structure.dist_structure}
\alias{as_dist_structure.dist_structure}
\alias{as_dist_structure.dist}
\alias{as_dist_structure}
\title{Coerce to dist_structure}
\usage{
\method{as_dist_structure}{dist_structure}(x, ...)

\method{as_dist_structure}{dist}(x, ...)

as_dist_structure(x, ...)
}
\arguments{
\item{x}{A dist or dist_structure.}

\item{...}{Reserved for methods.}
}
\value{
A \link{dist_structure} object.
}
\description{
Wraps a plain \code{algebraic.dist} distribution in a 1-component \link{series_dist}
so that it satisfies the dist_structure protocol. Useful for
polymorphic code that operates on dist_structure inputs but is also
handed unwrapped distributions. A \code{dist_structure} is returned
unchanged.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iid_constructors.R
\name{min_iid}
\alias{min_iid}
\title{Series system of m iid components}
\usage{
min_iid(d, m)
}
\arguments{
\item{d}{A \code{dist} object (the common component distribution).}

\item{m}{Number of components (positive integer).}
}
\value{
A \code{series_dist}.
}
\description{
Equivalent to the minimum of \code{m} iid random variables from \code{d}. Unlike
\code{min} in the base distribution algebra, \code{min_iid} preserves the series
topology so structural queries (phi, min_paths, structural_importance)
remain available.
}
\examples{
sys <- min_iid(algebraic.dist::exponential(1), m = 3)
# System survival at t=0.5 equals (exp(-t))^3 = exp(-1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{system_signature.dist_structure}
\alias{system_signature.dist_structure}
\alias{system_signature}
\title{System signature}
\usage{
\method{system_signature}{dist_structure}(x)

system_signature(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
Numeric vector of length \code{ncomponents(x)} summing to 1.
}
\description{
Samaniego's signature: \verb{s = (s_1, ..., s_m)} where \code{s_k} is the
probability the system fails at the k-th component failure under
i.i.d. absolutely-continuous component lifetimes. Only depends on the
structure (phi), not on the component distribution. A default method
on \code{dist_structure} enumerates the \verb{m!} orderings; this is feasible
for \code{m} up to about 8 or 9. Specialized subclasses should override
with closed-form expressions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{consecutive_k_dist}
\alias{consecutive_k_dist}
\title{Consecutive-k-out-of-n system distribution (type G)}
\usage{
consecutive_k_dist(k, components)
}
\arguments{
\item{k}{Block size.}

\item{components}{List of \code{dist} objects (length \code{n}).}
}
\value{
A \code{consecutive_k_dist} inheriting from \code{coherent_dist}.
}
\description{
The consecutive-k-out-of-n:G system functions when at least one block
of \code{k} consecutive components all function. Minimal path sets are the
\code{n - k + 1} consecutive blocks of size \code{k}. (Note: this is the :G
variant; the :F variant, "fails when any \code{k} consecutive fail", has
different minimal paths.)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lognormal_series.R
\name{lognormal_series}
\alias{lognormal_series}
\alias{surv.lognormal_series}
\alias{sampler.lognormal_series}
\title{Series of independent Lognormal components (closed form)}
\usage{
lognormal_series(meanlogs, sdlogs)

\method{surv}{lognormal_series}(x, ...)

\method{sampler}{lognormal_series}(x, ...)
}
\arguments{
\item{meanlogs}{Numeric vector of length \code{m}: per-component meanlog
parameters.}

\item{sdlogs}{Positive numeric vector of length \code{m}: per-component
sdlog parameters.}

\item{x}{A \code{lognormal_series} object.}

\item{...}{Ignored.}
}
\value{
\code{lognormal_series()} returns an object of class
\code{c("lognormal_series", "series_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
}
}
\description{
Constructs a \code{dist_structure} for a series system whose components are
independent Lognormals. Closed-form \code{surv} is the product of
per-component upper-tail probabilities; \code{cdf} is \code{1 - surv}; \code{sampler}
generates m independent Lognormals and takes the min.
}
\examples{
sys <- lognormal_series(meanlogs = c(0, 1), sdlogs = c(1, 0.5))
algebraic.dist::surv(sys)(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R, R/dual.R, R/generics.R
\name{dual.coherent_dist}
\alias{dual.coherent_dist}
\alias{dual.dist_structure}
\alias{dual.dual_of_system}
\alias{dual}
\title{Dual of a coherent_dist: swap cuts and paths}
\usage{
\method{dual}{coherent_dist}(x)

\method{dual}{dist_structure}(x)

\method{dual}{dual_of_system}(x)

dual(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
A \link{dist_structure} object representing the dual.
}
\description{
Overrides the lazy-wrapper default with a proper \code{coherent_dist}:
\code{min_paths(dual(x)) = min_cuts(x)}.

Returns a \code{dual_of_system} object carrying the original structure.
\code{phi_dual(state) = 1 - phi(original, 1 - state)} is evaluated on demand.
All other generics fall through to \code{dist_structure} defaults.

The dual structure satisfies \code{phi_dual(state) = 1 - phi(1 - state)}. The
dual of a series system is parallel; the dual of k-out-of-n is
(n - k + 1)-out-of-n.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{coherent_dist}
\alias{coherent_dist}
\title{Coherent system distribution from minimal path sets}
\usage{
coherent_dist(min_paths, components, m = NULL)
}
\arguments{
\item{min_paths}{List of integer vectors; each is a minimal path set.}

\item{components}{List of \code{dist} objects, length \code{m}. Each is a
fully-parameterized component lifetime distribution.}

\item{m}{Optional integer. Inferred from \code{components} and \code{min_paths}
if omitted.}
}
\value{
An object of class
\code{c("coherent_dist", "dist_structure", "univariate_dist", "dist")}.
}
\description{
General-purpose constructor. Users supply a list of minimal path sets
(each a vector of component indices) and a list of component
distributions (each an \code{algebraic.dist::dist} object with parameters
baked in). The resulting object is a \code{dist_structure} and \code{dist}.
}
\examples{
# A bridge network with exponential components
sys <- coherent_dist(
  min_paths = list(c(1, 4), c(2, 5), c(1, 3, 5), c(2, 3, 4)),
  components = replicate(5, algebraic.dist::exponential(1), simplify = FALSE)
)
reliability(sys, 0.9)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{parallel_dist}
\alias{parallel_dist}
\title{Parallel system distribution}
\usage{
parallel_dist(components)
}
\arguments{
\item{components}{List of \code{dist} objects.}
}
\value{
A \code{parallel_dist} inheriting from \code{coherent_dist}.
}
\description{
A parallel system fails only when all components fail. Equivalent to
\code{max(components)} but preserves topology.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{critical_states.dist_structure}
\alias{critical_states.dist_structure}
\alias{critical_states}
\title{Critical states of a component}
\usage{
\method{critical_states}{dist_structure}(x, j)

critical_states(x, j)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}
}
\value{
Integer matrix with \code{m - 1} columns; rows are the states of the
other components for which \code{j} is critical.
}
\description{
Critical states of a component
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{phi.dist_structure}
\alias{phi.dist_structure}
\alias{phi}
\title{Structure function}
\usage{
\method{phi}{dist_structure}(x, state)

phi(x, state)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{state}{Integer or logical vector of length \code{ncomponents(x)} in
\verb{\{0, 1\}}.}
}
\value{
Integer scalar, \code{0} or \code{1}.
}
\description{
Evaluate the coherent structure function \verb{phi: \{0, 1\}^m -> \{0, 1\}} at a
component state vector \code{state}. By convention \code{state[j] = 1} means
component \code{j} is functioning; \code{phi(state) = 1} means the system is
functioning.
}
\details{
\code{phi} and \code{\link[=min_paths]{min_paths()}} are dual primitives: if a subclass provides
only \verb{phi.<class>()}, the default \code{\link[=min_paths.dist_structure]{min_paths.dist_structure()}}
enumerates minimal subsets via \code{phi}; if a subclass provides only
\verb{min_paths.<class>()}, the default \code{\link[=phi.dist_structure]{phi.dist_structure()}} checks
whether \code{state} covers any minimal path. Subclasses providing both
must keep them consistent.
}
\seealso{
\code{\link[=min_paths]{min_paths()}} for the dual primitive; \code{\link[=validate_dist_structure]{validate_dist_structure()}}
for construction-time checking.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma_series.R
\name{gamma_series}
\alias{gamma_series}
\alias{surv.gamma_series}
\alias{sampler.gamma_series}
\title{Series of independent Gamma components (closed form)}
\usage{
gamma_series(shapes, rates)

\method{surv}{gamma_series}(x, ...)

\method{sampler}{gamma_series}(x, ...)
}
\arguments{
\item{shapes}{Positive numeric vector of length \code{m}: per-component
Gamma shape parameters.}

\item{rates}{Positive numeric vector of length \code{m}: per-component
Gamma rate parameters.}

\item{x}{A \code{gamma_series} object.}

\item{...}{Ignored.}
}
\value{
\code{gamma_series()} returns an object of class
\code{c("gamma_series", "series_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
}
}
\description{
Constructs a \code{dist_structure} for a series system whose components are
independent Gammas. Closed-form \code{surv} is evaluated by the product of
per-component upper-tail probabilities; \code{cdf} is \code{1 - surv}; \code{sampler}
generates m independent Gammas and takes the min.
}
\examples{
sys <- gamma_series(shapes = c(2, 3), rates = c(1, 2))
algebraic.dist::surv(sys)(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compositional.R, R/generics.R
\name{substitute_component.dist_structure}
\alias{substitute_component.dist_structure}
\alias{substitute_component}
\title{Substitute a component}
\usage{
\method{substitute_component}{dist_structure}(x, j, new_component)

substitute_component(x, j, new_component)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}

\item{new_component}{A dist-compatible object to install at position j.}
}
\value{
A new \link{dist_structure} object.
}
\description{
Return a new \link{dist_structure} with the j-th component replaced by
\code{new_component}. Topology is preserved; the returned object is a
\code{coherent_dist} with the same min_paths and the modified component
list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/importance.R
\name{vesely_fussell_importance}
\alias{vesely_fussell_importance}
\alias{vesely_fussell_importance.dist_structure}
\title{Vesely-Fussell importance at time t}
\usage{
vesely_fussell_importance(x, j, t)

\method{vesely_fussell_importance}{dist_structure}(x, j, t)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}

\item{t}{Scalar time.}
}
\value{
Numeric scalar in \verb{[0, 1]}.
}
\description{
Probability that at least one minimal cut set containing \code{j} has all
its components failed, given the system has failed by time \code{t}.
Computed exactly via inclusion-exclusion over subsets of cuts that
contain \code{j}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{bridge_dist}
\alias{bridge_dist}
\title{Bridge system distribution}
\usage{
bridge_dist(components)
}
\arguments{
\item{components}{List of 5 \code{dist} objects.}
}
\value{
A \code{bridge_dist} inheriting from \code{coherent_dist}.
}
\description{
The classical 5-component bridge reliability network with minimal
path sets \verb{\{1,4\}}, \verb{\{2,5\}}, \verb{\{1,3,5\}}, \verb{\{2,3,4\}}. Components 1 and 2
are the input side, 4 and 5 the output side, and 3 the cross-link.
The bridge is a canonical non-series, non-parallel example used
throughout the reliability literature; see Barlow and Proschan (1975,
"Statistical Theory of Reliability and Life Testing") for the
standard treatment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_structure-class.R
\name{is_dist_structure}
\alias{is_dist_structure}
\title{Predicate for dist_structure objects}
\usage{
is_dist_structure(x)
}
\arguments{
\item{x}{Any object.}
}
\value{
\code{TRUE} if \code{x} inherits from \code{"dist_structure"}.
}
\description{
Predicate for dist_structure objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{component}
\alias{component}
\title{The j-th component as a dist}
\usage{
component(x, j, ...)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Integer component index in \code{1:ncomponents(x)}.}

\item{...}{Implementation-specific arguments (e.g., parameter overrides
for lazily parameterized systems).}
}
\value{
An object inheriting from \code{dist}.
}
\description{
Returns the j-th component as an \code{algebraic.dist::dist} object with
parameters baked in (so \code{sampler}, \code{surv}, etc. work directly on it).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{is_coherent.dist_structure}
\alias{is_coherent.dist_structure}
\alias{is_coherent}
\title{Coherence axiom check}
\usage{
\method{is_coherent}{dist_structure}(x)

is_coherent(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
\code{TRUE} if monotone and every component relevant.
}
\description{
Coherence axiom check
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_structure-class.R
\name{format.dist_structure}
\alias{format.dist_structure}
\title{Format a dist_structure object}
\usage{
\method{format}{dist_structure}(x, ...)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{...}{Ignored.}
}
\value{
Character vector suitable for \code{\link[=cat]{cat()}}.
}
\description{
Format a dist_structure object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compositional.R, R/generics.R
\name{compose_systems.dist_structure}
\alias{compose_systems.dist_structure}
\alias{compose_systems}
\title{Compose systems hierarchically}
\usage{
\method{compose_systems}{dist_structure}(outer, inner_list)

compose_systems(outer, inner_list)
}
\arguments{
\item{outer}{A \link{dist_structure} object.}

\item{inner_list}{A list of length \code{ncomponents(outer)}. Each element
is either a \link{dist_structure} or a dist (single-component sub-system).}
}
\value{
A \code{coherent_dist} representing the composed system.
}
\description{
Produce a new \link{dist_structure} by replacing each component of \code{outer}
with a sub-system (either a \link{dist_structure} or a plain dist). The
composed system's components are the flattened inner components; its
min_paths enumerate the Cartesian products of inner min_paths within
each outer min_path.
}
\details{
Computational note: the composed minimal-path enumeration takes the
Cartesian product of inner-path choices over each outer path. For an
outer system with \code{p} paths each of length \code{q}, where each inner has
\code{r} paths, the candidate count grows as \code{O(p * r^q)} before
deduplication. Bridge-of-bridges and similar deeply nested
compositions can produce hundreds of candidates; if you find the call
slow, build the composed \code{coherent_dist} directly with a hand-curated
\code{min_paths} list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exp_parallel.R
\name{exp_parallel}
\alias{exp_parallel}
\alias{surv.exp_parallel}
\alias{sampler.exp_parallel}
\alias{mean.exp_parallel}
\title{Parallel of exponential components (closed form)}
\usage{
exp_parallel(rates)

\method{surv}{exp_parallel}(x, ...)

\method{sampler}{exp_parallel}(x, ...)

\method{mean}{exp_parallel}(x, ...)
}
\arguments{
\item{rates}{Positive numeric vector of length \code{m}.}

\item{x}{An \code{exp_parallel} object.}

\item{...}{Ignored.}
}
\value{
\code{exp_parallel()} returns an object of class
\code{c("exp_parallel", "parallel_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
\item \code{mean()}: a numeric scalar (the mean system lifetime,
computed in closed form via inclusion-exclusion).
}
}
\description{
Constructs a \code{dist_structure} representing a parallel system whose
components are independent exponentials. Closed-form methods are
provided for \code{surv}, \code{cdf}, \code{sampler}, and \code{mean} (the last via
inclusion-exclusion over the \code{2^m - 1} non-empty component subsets).
}
\examples{
sys <- exp_parallel(c(1, 2, 3))
algebraic.dist::surv(sys)(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/importance.R
\name{birnbaum_importance}
\alias{birnbaum_importance}
\alias{birnbaum_importance.dist_structure}
\title{Birnbaum reliability importance}
\usage{
birnbaum_importance(x, j, p)

\method{birnbaum_importance}{dist_structure}(x, j, p)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}

\item{p}{Numeric vector of length \code{ncomponents(x)} or a scalar in
\verb{[0, 1]}.}
}
\value{
Numeric scalar in \verb{[0, 1]}.
}
\description{
The Birnbaum importance of component \code{j} at component reliabilities
\code{p} is \verb{dR/dp_j = R(p with p_j = 1) - R(p with p_j = 0)}. Measures
how much the system reliability changes if \code{j} moves between certain
failure and certain success (and, by monotonicity, the rate of
change at any intermediate \code{p_j}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cold_standby.R
\name{mean.cold_standby_dist}
\alias{mean.cold_standby_dist}
\title{Mean of a cold-standby system: sum of component means}
\usage{
\method{mean}{cold_standby_dist}(x, ...)
}
\arguments{
\item{x}{A \code{cold_standby_dist} object.}

\item{...}{Passed to the Monte Carlo fallback as \code{mc} (default 1e5).}
}
\value{
Numeric scalar.
}
\description{
Computes \verb{sum_j E[T_j]} exactly when every component implements a
\code{mean()} method. Falls back to a Monte Carlo estimate from the
sampler when any component lacks an exact mean (e.g., a
\code{dist_structure} component whose \code{mean} would route through
\code{algebraic.dist::univariate_dist} and require \code{density} / \code{sup},
which are not provided on \code{dist_structure} by default).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wei_kofn.R
\name{wei_kofn}
\alias{wei_kofn}
\alias{surv.wei_kofn}
\alias{sampler.wei_kofn}
\alias{hazard.wei_kofn}
\alias{density.wei_kofn}
\title{k-out-of-n system of independent Weibull components (closed form)}
\usage{
wei_kofn(k, shapes, scales)

\method{surv}{wei_kofn}(x, ...)

\method{sampler}{wei_kofn}(x, ...)

\method{hazard}{wei_kofn}(x, ...)

\method{density}{wei_kofn}(x, ...)
}
\arguments{
\item{k}{Minimum functioning components for system operation.}

\item{shapes}{Positive numeric vector of length \code{m}.}

\item{scales}{Positive numeric vector of length \code{m}.}

\item{x}{A \code{wei_kofn} object.}

\item{...}{Ignored.}
}
\value{
\code{wei_kofn()} returns an object of class
\code{c("wei_kofn", "kofn_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}, \code{density()}, \code{hazard()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
}
}
\description{
Constructs a \code{dist_structure} for a k-out-of-m system whose components
are independent (possibly heterogeneous) Weibulls. Closed-form \code{surv},
\code{cdf}, \code{sampler}, \code{density}, and \code{hazard} via subset enumeration,
the critical-state density formula, and component order statistics.
}
\examples{
sys <- wei_kofn(k = 2, shapes = c(1, 2, 3), scales = c(1, 2, 3))
algebraic.dist::surv(sys)(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{ncomponents}
\alias{ncomponents}
\title{Number of components}
\usage{
ncomponents(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
A positive integer \code{m}, the number of components.
}
\description{
Number of components
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_structure-class.R
\name{validate_dist_structure}
\alias{validate_dist_structure}
\title{Validate that an object satisfies the dist_structure protocol}
\usage{
validate_dist_structure(x)
}
\arguments{
\item{x}{An object claiming to satisfy the \code{dist_structure} protocol.}
}
\value{
\code{TRUE} (invisibly) if all checks pass. Stops with an
informative error otherwise.
}
\description{
Checks that \code{x} declares \code{"dist_structure"} in its class chain and
provides the three required generics: \code{\link[=ncomponents]{ncomponents()}}, \code{\link[=component]{component()}},
and at least one of \code{\link[=phi]{phi()}} or \code{\link[=min_paths]{min_paths()}}.
}
\details{
Useful in subclass constructors to fail fast with a clear error
message rather than at first method dispatch (where the user sees
opaque "no applicable method" errors). A typical pattern:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{my_subclass <- function(...) \{
  obj <- structure(list(...), class = c("my_subclass", "dist_structure",
                                         "univariate_dist", "continuous_dist", "dist"))
  validate_dist_structure(obj)
  obj
\}
}\if{html}{\out{</div>}}
}
\examples{
# Success: a valid dist_structure (any built-in topology shortcut works).
validate_dist_structure(series_dist(replicate(3,
  algebraic.dist::exponential(1), simplify = FALSE)))

# Failure: an object that declares dist_structure but provides no
# primitives. Wrapped in tryCatch for the example's success status.
bad <- structure(list(),
                 class = c("not_a_real_class", "dist_structure",
                           "univariate_dist", "continuous_dist", "dist"))
tryCatch(validate_dist_structure(bad),
         error = function(e) conditionMessage(e))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wei_homogeneous_series.R
\name{wei_homogeneous_series}
\alias{wei_homogeneous_series}
\alias{surv.wei_homogeneous_series}
\alias{sampler.wei_homogeneous_series}
\alias{mean.wei_homogeneous_series}
\title{Series of Weibull components with common shape (closed form as a Weibull)}
\usage{
wei_homogeneous_series(shape, scales)

\method{surv}{wei_homogeneous_series}(x, ...)

\method{sampler}{wei_homogeneous_series}(x, ...)

\method{mean}{wei_homogeneous_series}(x, ...)
}
\arguments{
\item{shape}{Positive scalar: common Weibull shape.}

\item{scales}{Positive numeric vector: per-component Weibull scales.}

\item{x}{A \code{wei_homogeneous_series} object.}

\item{...}{Ignored.}
}
\value{
\code{wei_homogeneous_series()} returns an object of class
\code{c("wei_homogeneous_series", "wei_series", "series_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
\item \code{mean()}: a numeric scalar (the mean system lifetime,
\code{aggregate_scale * gamma(1 + 1 / shape)}).
}
}
\description{
Constructs a \code{dist_structure} for a series of Weibull components sharing
a common shape parameter. By the standard identity, the system lifetime
is itself Weibull with the common shape and an aggregate scale
\code{(sum(1 / scale^shape))^(-1 / shape)}. Methods for \code{surv}, \code{cdf},
\code{sampler}, and \code{mean} forward to this aggregate Weibull, giving exact
closed-form values.
}
\examples{
sys <- wei_homogeneous_series(shape = 2, scales = c(1, 2, 3))
# System lifetime is Weibull(shape = 2, scale = aggregate_scale)
algebraic.dist::surv(sys)(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cold_standby.R
\name{cold_standby_dist}
\alias{cold_standby_dist}
\alias{sampler.cold_standby_dist}
\alias{surv.cold_standby_dist}
\alias{cdf.cold_standby_dist}
\title{Cold-standby system distribution}
\usage{
cold_standby_dist(components)

\method{sampler}{cold_standby_dist}(x, ...)

\method{surv}{cold_standby_dist}(x, ...)

\method{cdf}{cold_standby_dist}(x, ...)
}
\arguments{
\item{components}{List of \code{dist} objects representing per-stage
component lifetimes.}

\item{x}{A \code{cold_standby_dist} object.}

\item{...}{Ignored.}
}
\value{
An object of class
\code{c("cold_standby_dist", "univariate_dist", "continuous_dist", "dist")}.
}
\description{
Constructs a distribution representing a cold-standby system: one
component active at a time with perfect, instantaneous switching to
the next spare upon failure. System lifetime equals the sum of
independent component lifetimes.

Estimates \code{S(t) = P(T_1 + ... + T_m > t)} by simulating \code{mc} system
lifetimes and computing the empirical fraction exceeding \code{t}. The
returned closure caches its sample vector: the first call generates
\code{mc} samples, subsequent calls reuse them as long as \code{mc} is
unchanged (a different \code{mc} triggers a fresh draw). This makes
repeated \code{S(t)} calls deterministic given the same \code{mc}.

Computes the CDF by deferring to a fresh \code{surv} closure. The CDF
closure has its own sample cache (independent of any external \code{surv}
closure), so \code{cdf(x)(t) + surv(x)(t)} is generally not exactly 1
unless callers reuse a single \code{surv} closure: prefer
\verb{S <- surv(x); F_t <- 1 - S(t)} over computing both independently.
}
\details{
Cold standby is not a coherent system in the structure-function sense
(its topology is temporal succession, not an order statistic), so the
returned object does not inherit \code{dist_structure}. It IS a \code{dist}
(with \code{surv}, \code{cdf}, \code{sampler}, \code{mean} available) and exposes
\code{\link[=ncomponents]{ncomponents()}} and \code{\link[=component]{component()}} for inspection.

Defaults: \code{sampler} is exact (sample each component independently and
sum); \code{mean} is exact when every component implements \code{mean()}
(otherwise falls back to Monte Carlo via the sampler); \code{surv} and
\code{cdf} use Monte Carlo with a default of \code{1e5} simulated lifetimes
(override via the \code{mc} argument). The returned \code{surv} / \code{cdf}
closures cache their samples after the first call so subsequent
evaluations at different \code{t} values are deterministic given the same
\code{mc}.

Methods inherited from \code{algebraic.dist::univariate_dist} that require
\code{density} or \code{sup} (notably \code{expectation}, \code{vcov}) are NOT supported
on cold-standby objects out of the box; specialized subclasses with
closed-form aggregate distributions (e.g., iid exponential collapses
to \code{Gamma(m, rate)}) should provide their own methods.

For reproducibility across calls to \code{surv()} itself (i.e., between
separately constructed closures), set the RNG seed externally via
\code{set.seed()} before invoking \code{surv(x)}. Override the method on a
subclass with an exact aggregate distribution (e.g., iid exponential
collapses to \code{Gamma(m, rate)}) when an analytic form is available.
}
\examples{
# Cold standby of 3 iid Exp(1) components: aggregate is Gamma(3, 1)
sys <- cold_standby_dist(replicate(3,
  algebraic.dist::exponential(1), simplify = FALSE))
mean(sys)  # = 3
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{min_cuts.dist_structure}
\alias{min_cuts.dist_structure}
\alias{min_cuts}
\title{Minimal cut sets}
\usage{
\method{min_cuts}{dist_structure}(x)

min_cuts(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
A list of integer vectors.
}
\description{
Minimal cut sets
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R, R/importance.R
\name{criticality_importance}
\alias{criticality_importance}
\alias{criticality_importance.dist_structure}
\title{Criticality (Fussell) importance at time t}
\usage{
criticality_importance(x, j, t)

\method{criticality_importance}{dist_structure}(x, j, t)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}

\item{t}{Scalar time.}
}
\value{
Numeric scalar in \verb{[0, 1]}.
}
\description{
Probability that component \code{j} has failed AND is critical given
that the system has failed by time \code{t}. Equals
\verb{I_B(j; S(t)) * F_j(t) / F_sys(t)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{structural_importance.dist_structure}
\alias{structural_importance.dist_structure}
\alias{structural_importance}
\title{Birnbaum structural importance}
\usage{
\method{structural_importance}{dist_structure}(x, j)

structural_importance(x, j)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{j}{Component index.}
}
\value{
Numeric scalar in \verb{[0, 1]}.
}
\description{
Fraction of the \code{2^(m - 1)} states of the other components for which
component \code{j} is critical.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{series_dist}
\alias{series_dist}
\title{Series system distribution}
\usage{
series_dist(components)
}
\arguments{
\item{components}{List of \code{dist} objects.}
}
\value{
A \code{series_dist} inheriting from \code{coherent_dist}.
}
\description{
A series system fails if any single component fails. Equivalent to
\code{min(components)} as random variables; unlike \code{min} in the base algebra,
\code{series_dist} preserves the component decomposition so topology queries
work.
}
\examples{
sys <- series_dist(replicate(3, algebraic.dist::exponential(1), simplify = FALSE))
algebraic.dist::surv(sys)(0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iid_constructors.R
\name{max_iid}
\alias{max_iid}
\title{Parallel system of m iid components}
\usage{
max_iid(d, m)
}
\arguments{
\item{d}{A \code{dist} object.}

\item{m}{Number of components.}
}
\value{
A \code{parallel_dist}.
}
\description{
Equivalent to the maximum of \code{m} iid random variables from \code{d}.
Preserves the parallel topology for structural queries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{reliability.dist_structure}
\alias{reliability.dist_structure}
\alias{reliability}
\title{System reliability polynomial}
\usage{
\method{reliability}{dist_structure}(x, p)

reliability(x, p)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{p}{Numeric vector of length \code{ncomponents(x)} or a scalar recycled
to all components; values in \verb{[0, 1]}.}
}
\value{
Numeric scalar in \verb{[0, 1]}.
}
\description{
\code{R(p) = E[phi(X)]} where \code{X_j ~ Bernoulli(p_j)} independent. The
multilinear extension of \code{phi} to component reliabilities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist-defaults.R, R/dist_structure-class.R
\name{surv.dist_structure}
\alias{surv.dist_structure}
\alias{cdf.dist_structure}
\alias{sampler.dist_structure}
\alias{dist_structure}
\title{System survival via reliability composition}
\usage{
\method{surv}{dist_structure}(x, ...)

\method{cdf}{dist_structure}(x, ...)

\method{sampler}{dist_structure}(x, ...)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{...}{Ignored.}
}
\value{
This help topic documents the virtual base class together with the
three distribution-algebra default methods that compose component
distributions through the topology:
\itemize{
\item \code{\link[=surv.dist_structure]{surv.dist_structure()}} returns a closure \verb{function(t, ...)} that
evaluates the system survival function via the reliability identity
\code{S_sys(t) = R(S_1(t), ..., S_m(t))}.
\item \code{\link[=cdf.dist_structure]{cdf.dist_structure()}} returns a closure \verb{function(t, ...)} equal
to \code{1 - surv(x)(t)}.
\item \code{\link[=sampler.dist_structure]{sampler.dist_structure()}} returns a closure \verb{function(n, ...)}
that draws \code{n} independent system lifetimes by sampling each
component and combining via \code{\link[=system_lifetime]{system_lifetime()}}.
}

Concrete subclasses override any of these for closed-form speed; see
the closed-form specializations under the See Also entries.
}
\description{
Returns a closure \verb{function(t, ...)} where \code{surv(x)(t)} equals
\verb{reliability(x, surv(component(x, j))(t) for each j)}. This is the
classical identity \code{S_sys(t) = R(S(t))} realized as a composition.

Returns a closure \verb{function(n, ...)} that draws \code{n} system lifetimes by
sampling each component independently and applying \code{\link[=system_lifetime]{system_lifetime()}}
to combine.

\code{dist_structure} is the virtual S3 class for distributions whose random
variable has internal component structure: coherent reliability
systems decomposed into components arranged by a structure function.
Every concrete implementation (\code{coherent_dist}, \code{series_dist},
\code{parallel_dist}, \code{kofn_dist}, \code{bridge_dist}, or user-defined
subclasses) should include \code{"dist_structure"}, the \code{algebraic.dist}
ancestor \code{"univariate_dist"}, and \code{"dist"} in its class vector.
}
\details{
Concrete implementations provide S3 methods for the generics in this
package. The minimum required methods are \code{\link[=ncomponents]{ncomponents()}}, \code{\link[=component]{component()}},
and one of \code{\link[=phi]{phi()}} or \code{\link[=min_paths]{min_paths()}}; every other generic has a default
method on \code{dist_structure} that composes the primitives.

If both \verb{phi.<class>()} and \verb{min_paths.<class>()} are provided, the
implementor is responsible for keeping them consistent: \code{phi} derives
the in-package generics \code{reliability}, \code{critical_states}, and
\code{is_coherent}; \code{min_paths} derives \code{min_cuts} and \code{system_signature}.
Inconsistent implementations produce silently inconsistent results.
}
\seealso{
\code{\link[=validate_dist_structure]{validate_dist_structure()}} for an implementor-side
construction-time validator. \code{\link[=phi]{phi()}} and \code{\link[=min_paths]{min_paths()}} for the
bidirectional protocol primitives. The closed-form families
(\code{\link[=exp_series]{exp_series()}}, \code{\link[=wei_kofn]{wei_kofn()}}, etc.) for reference implementations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent_dist.R
\name{kofn_dist}
\alias{kofn_dist}
\title{k-out-of-n system distribution}
\usage{
kofn_dist(k, components)
}
\arguments{
\item{k}{Minimum functioning components for system operation (:G).}

\item{components}{List of \code{dist} objects (length \code{m}).}
}
\value{
A \code{kofn_dist} inheriting from \code{coherent_dist}.
}
\description{
A k-out-of-n system functions if at least \code{k} of its \code{m} components
function. Equivalent to the \code{(m - k + 1)}-th order statistic of
component lifetimes.
}
\details{
This constructor uses the \strong{:G} convention: \code{k} is the number of
components that must remain \strong{functioning} for the system to function.
\code{k = 1} is parallel; \code{k = m} is series. The companion \code{kofn} package
(which depends on \code{dist.structure}) uses the \strong{:F} convention, where
\code{k} is the number of failures that trigger system failure; conversion
is \code{k_dist = m - k_kofn + 1}. When in doubt, draw a small example:
\code{kofn_dist(k = 2, ...)} for \code{m = 3} functions until two of the three
components have failed.
}
\seealso{
\code{\link[=order_statistic]{order_statistic()}} for the closely-related order-statistic
parameterization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{min_paths.dist_structure}
\alias{min_paths.dist_structure}
\alias{min_paths}
\title{Minimal path sets}
\usage{
\method{min_paths}{dist_structure}(x)

min_paths(x)
}
\arguments{
\item{x}{A \link{dist_structure} object.}
}
\value{
A list of integer vectors.
}
\description{
Returns the minimal subsets of \code{1:m} whose joint functioning is
sufficient for the system to function. \code{min_paths} and \code{\link[=phi]{phi()}} are
dual primitives in the \code{dist_structure} protocol: providing one is
enough, the other has an enumerative default.
}
\seealso{
\code{\link[=phi]{phi()}} for the dual primitive; \code{\link[=min_cuts]{min_cuts()}} for the dual
topology query (minimal subsets whose joint failure causes system
failure).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exp_series.R
\name{exp_series}
\alias{exp_series}
\alias{surv.exp_series}
\alias{sampler.exp_series}
\alias{mean.exp_series}
\alias{density.exp_series}
\alias{hazard.exp_series}
\title{Series of exponential components (closed form)}
\usage{
exp_series(rates)

\method{surv}{exp_series}(x, ...)

\method{sampler}{exp_series}(x, ...)

\method{mean}{exp_series}(x, ...)

\method{density}{exp_series}(x, ...)

\method{hazard}{exp_series}(x, ...)
}
\arguments{
\item{rates}{Positive numeric vector of length \code{m}: per-component
exponential rates.}

\item{x}{An \code{exp_series} object.}

\item{...}{Ignored.}
}
\value{
\code{exp_series()} returns an object of class
\code{c("exp_series", "series_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}, \code{density()}, \code{hazard()}: a closure \verb{function(t, ...)}
evaluating the named quantity at \code{t}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
\item \code{mean()}: a numeric scalar (the mean system lifetime).
}
}
\description{
Constructs a \code{dist_structure} representing a series system whose
components are independent exponentials. The system lifetime is itself
an \code{Exp(sum(rates))} distribution; all dist-level queries have
closed-form expressions that bypass the general default methods.
}
\examples{
sys <- exp_series(c(0.5, 0.3, 0.2))
algebraic.dist::surv(sys)(1)  # equals exp(-sum(rates) * 1)
mean(sys)                     # equals 1 / sum(rates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_structure-class.R
\name{print.dist_structure}
\alias{print.dist_structure}
\title{Print a dist_structure object}
\usage{
\method{print}{dist_structure}(x, ...)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{...}{Passed to \code{\link[=format]{format()}}.}
}
\value{
\code{x}, invisibly.
}
\description{
Print a dist_structure object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{system_censoring.dist_structure}
\alias{system_censoring.dist_structure}
\alias{system_censoring}
\title{Per-component censoring from system observation}
\usage{
\method{system_censoring}{dist_structure}(x, times)

system_censoring(x, times)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{times}{Non-negative numeric vector of length \code{ncomponents(x)}.}
}
\value{
A list with \code{system_time} (scalar) and \code{component_status}
(character vector of length \code{m}, values in \code{"exact"}, \code{"left"},
\code{"right"}).
}
\description{
Per-component censoring from system observation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist.structure-package.R
\docType{package}
\name{dist.structure-package}
\alias{dist.structure}
\alias{dist.structure-package}
\title{dist.structure: Structured Random Variables for Reliability System
Distributions}
\description{
Extends the algebraic.dist distribution algebra to random variables with
internal structure: coherent reliability systems decomposed into components
arranged by a structure function (series, parallel, k-out-of-n, bridge,
arbitrary topologies). Every \code{dist_structure} object is also a \code{dist}, so
the full distribution algebra works automatically via default methods that
compose component-level distributions through the topology.
}
\section{Two algebras, one hierarchy}{


\code{algebraic.dist} provides the base algebra of random variables:
\code{density}, \code{surv}, \code{hazard}, \code{sampler}, \code{mean}, \code{vcov}, \code{expectation},
and arithmetic on RVs (transformations).

\code{dist.structure} extends it with structure-preserving operations for
reliability systems: component-wise decomposition, structure function
evaluation, path and cut sets, system signature, critical states, dual
structures, and Birnbaum importance measures. Because every
\code{dist_structure} is a \code{dist}, the base algebra inherits automatically.
}

\section{Primitives implementors provide}{

\itemize{
\item \code{\link[=ncomponents]{ncomponents()}}: number of components \code{m}
\item \code{\link[=component]{component()}}: return the j-th component as a \code{dist}
\item One of \code{\link[=phi]{phi()}} or \code{\link[=min_paths]{min_paths()}}: the other defaults automatically
}
}

\section{Topology queries (default methods on \code{dist_structure})}{


\code{\link[=phi]{phi()}}, \code{\link[=min_paths]{min_paths()}}, \code{\link[=min_cuts]{min_cuts()}}, \code{\link[=critical_states]{critical_states()}},
\code{\link[=system_signature]{system_signature()}}, \code{\link[=system_lifetime]{system_lifetime()}}, \code{\link[=system_censoring]{system_censoring()}},
\code{\link[=dual]{dual()}}, \code{\link[=is_coherent]{is_coherent()}}, \code{\link[=structural_importance]{structural_importance()}}, \code{\link[=reliability]{reliability()}}
}

\section{Distribution queries (default methods via component composition)}{


\code{\link[algebraic.dist:surv]{algebraic.dist::surv()}}, \code{\link[algebraic.dist:cdf]{algebraic.dist::cdf()}}, \code{\link[algebraic.dist:sampler]{algebraic.dist::sampler()}}. Further base dist methods (\code{mean},
\code{vcov}, \code{expectation}) inherit from \code{univariate_dist} via numerical
integration of \code{surv}. Implementors override any default for speed.
}

\section{Reference implementations}{

\itemize{
\item \code{\link[=coherent_dist]{coherent_dist()}}: general coherent system given minimal path sets +
components
\item Topology shortcuts: \code{\link[=series_dist]{series_dist()}}, \code{\link[=parallel_dist]{parallel_dist()}}, \code{\link[=kofn_dist]{kofn_dist()}},
\code{\link[=bridge_dist]{bridge_dist()}}, \code{\link[=consecutive_k_dist]{consecutive_k_dist()}}
\item IID convenience: \code{\link[=min_iid]{min_iid()}}, \code{\link[=max_iid]{max_iid()}}, \code{\link[=order_statistic]{order_statistic()}}
\item Coercions: \code{\link[=as_dist_structure]{as_dist_structure()}}
}
}

\section{Closed-form specializations}{


Fast analytical paths for common (component family, topology) pairs.
Override the generic defaults for speed without changing semantics:
\itemize{
\item Series: \code{\link[=exp_series]{exp_series()}}, \code{\link[=wei_series]{wei_series()}}, \code{\link[=wei_homogeneous_series]{wei_homogeneous_series()}},
\code{\link[=gamma_series]{gamma_series()}}, \code{\link[=lognormal_series]{lognormal_series()}}
\item Parallel: \code{\link[=exp_parallel]{exp_parallel()}}
\item k-out-of-n: \code{\link[=exp_kofn]{exp_kofn()}}, \code{\link[=wei_kofn]{wei_kofn()}}
}
}

\section{Importance measures}{


Both topology-only and reliability-aware:
\itemize{
\item \code{\link[=structural_importance]{structural_importance()}}: Birnbaum's structural measure (topology only)
\item \code{\link[=birnbaum_importance]{birnbaum_importance()}}: Birnbaum's reliability measure
\item \code{\link[=criticality_importance]{criticality_importance()}}: criticality importance
\item \code{\link[=vesely_fussell_importance]{vesely_fussell_importance()}}: Vesely-Fussell importance
}
}

\section{Compositional operations}{


Build new systems from existing ones while preserving the protocol:
\itemize{
\item \code{\link[=substitute_component]{substitute_component()}}: replace component j with a sub-system
\item \code{\link[=compose_systems]{compose_systems()}}: compose two systems (cartesian product of paths)
}
}

\section{Non-coherent systems}{


\code{\link[=cold_standby_dist]{cold_standby_dist()}} models a cold-standby spare arrangement (system
lifetime is the sum of component lifetimes). It is deliberately \emph{not}
a \code{dist_structure} (no static structure function), but it shares the
\code{\link[=ncomponents]{ncomponents()}}/\code{\link[=component]{component()}} generics so user code can iterate
uniformly.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/queelius/dist.structure}
  \item Report bugs at \url{https://github.com/queelius/dist.structure/issues}
}

}
\author{
\strong{Maintainer}: Alexander Towell \email{lex@metafunctor.com} (\href{https://orcid.org/0000-0001-6443-9897}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R, R/generics.R
\name{system_lifetime.dist_structure}
\alias{system_lifetime.dist_structure}
\alias{system_lifetime}
\title{System lifetime from component times}
\usage{
\method{system_lifetime}{dist_structure}(x, times)

system_lifetime(x, times)
}
\arguments{
\item{x}{A \link{dist_structure} object.}

\item{times}{Non-negative numeric vector of length \code{ncomponents(x)}.}
}
\value{
Scalar system lifetime.
}
\description{
System lifetime from component times
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iid_constructors.R
\name{order_statistic}
\alias{order_statistic}
\title{k-th order statistic of m iid components}
\usage{
order_statistic(d, k, m)
}
\arguments{
\item{d}{A \code{dist} object.}

\item{k}{The order statistic index (1 = min, m = max).}

\item{m}{Number of iid components.}
}
\value{
A \code{kofn_dist}.
}
\description{
Constructs a \code{kofn_dist} whose system lifetime equals \code{T_(k)}, the
k-th order statistic of \code{m} iid draws from \code{d}. Under the k-of-m
parametrization, the system fails at the \code{(m - k + 1)}-th component
failure; setting the threshold to \code{m - k + 1} makes the system lifetime
equal \code{T_(k)}.
}
\details{
The internal call is \code{kofn_dist(m - k + 1, ...)}: this maps the
order-statistic index \code{k} (where \code{k = 1} is the minimum, \code{k = m} is
the maximum) to the \verb{:G} convention used by \code{kofn_dist} (where the
\code{k_dist} argument is the number of functioning components required).
}
\examples{
# Median of 5 iid exponentials
sys <- order_statistic(algebraic.dist::exponential(1), k = 3, m = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exp_kofn.R
\name{exp_kofn}
\alias{exp_kofn}
\alias{surv.exp_kofn}
\alias{sampler.exp_kofn}
\alias{hazard.exp_kofn}
\alias{density.exp_kofn}
\title{k-out-of-n system of independent exponential components (closed form)}
\usage{
exp_kofn(k, rates)

\method{surv}{exp_kofn}(x, ...)

\method{sampler}{exp_kofn}(x, ...)

\method{hazard}{exp_kofn}(x, ...)

\method{density}{exp_kofn}(x, ...)
}
\arguments{
\item{k}{Minimum number of functioning components for system operation.}

\item{rates}{Positive numeric vector of length \code{m} with \code{m >= k}.}

\item{x}{An \code{exp_kofn} object.}

\item{...}{Ignored.}
}
\value{
\code{exp_kofn()} returns an object of class
\code{c("exp_kofn", "kofn_dist", "coherent_dist", "dist_structure", "univariate_dist", "continuous_dist", "dist")}.

The associated S3 methods return:
\itemize{
\item \code{surv()}, \code{density()}, \code{hazard()}: a closure \verb{function(t, ...)}.
\item \code{cdf()} is derived via the \code{dist_structure} default and returns
a closure \verb{function(t, ...)} equal to \code{1 - surv(x)(t)}.
\item \code{sampler()}: a closure \verb{function(n, ...)} returning \code{n} random
variates from the system lifetime distribution.
}
}
\description{
Constructs a \code{dist_structure} for a k-out-of-m system whose components
are independent exponentials. Closed-form methods are provided for
\code{surv}, \code{cdf}, \code{sampler}, \code{density}, and \code{hazard}. \code{mean} falls back
to numerical integration via the \code{dist_structure} default.
}
\examples{
sys <- exp_kofn(k = 2, rates = c(1, 2, 3))
algebraic.dist::surv(sys)(1)
}
